################################################################################
## IMPORTS
################################################################################

from extronlib import event
from extronlib.interface import SerialInterface, EthernetClientInterface
from extronlib.system import Wait

import re
import math

################################################################################
## CONSTANTS
################################################################################

RESPONSE_USERNAME = '\r\nUsername: '
RESPONSE_PASSWORD = '\r\nPassword: '
RESPONSE_TERMINATOR = '\r\n=> '
COMMAND_TERMINATOR = '\r'
STATUS_COMMAND = 1
STATUS_QUERY = 2

################################################################################
## DICTIONARIES/ENUMS
################################################################################

SignalTypes = {
    'Digital': 0,
    'Analog': 1,
    'Serial': 2,
    'Null': 3
}

RampTypes = {
    'Stop': 0,
    'Raise': 1,
    'Lower': 2
}

GainTypes = {
    'GAIN': 0,
    'GAIN_FINE': 1,
    'Null': 2
}

ResponseMessageType = {
    'STACK': 0,
    'BOX': 1,
    'ROOM': 2,
    'EP': 3,
    'MT': 4,
    'MCCF': 5,
    'PROMPT_ONLY': 6,
    'UNKNOWN': 7
}

EndpointType = {
    'MIC': 0,
    'TELCO_RX': 3,
    'TELCO_TX': 4,
    'VOIP_RX': 5,
    'VOIP_TX': 6,
    'OUTPUT': 7,
    'SPEAKER': 8,
    'PROC': 9,
    'FADER': 10,
    'BFM': 13,
    'USB_RX': 14,
    'USB_TX': 15,
    'UA': 20,
    'GPIO': 21,
    'SGEN': 22,
    'SRMIC': 23,
    'DANTE_RX': 24,
    'DANTE_TX': 25
}

CrosspointTypes = {
    'None': 0,
    'Crosspoint': 1,
    'Gated': 3,
    'NonGated': 4,
    'PreAec': 5,
    'Unknown': 255
}

RoomOptions = {
    'Mode': 1,
    'WallState': 2,
    'Select': 7,
    'PresetState': 9,
    'IndividualDividerState': 11,
    'DividerPolarity': 12,
    'DividerState': 13,
    'Unknown': 255
}

ModeTypes = {
    'Normal': 0,
    'GPIO': 1
}

StateChangeTypes = {
    'UNKNOWN': 0,
    'IDLE': 1,
    'DIAL_TONE': 2,
    'DIALING': 3,
    'INPROCESS': 4,
    'RINGING': 5,
    'BUSY': 6,
    'ACTIVE': 7,
    'HOLD': 8,
    'INCOMING': 9,
    'CFW': 10,
    'CONFERENCE_ACTIVE': 11,
    'CONFERENCE_HOLD': 12,
    'TRANSFER_ACTIVE': 13,
    'TRANSFER_HOLD': 14,
    'TRANSFERRING_DIAL_TONE': 15,
    'TRANSFERRING_DIALING': 16,
    'TRANSFERRING_INPROCESS': 17,
    'TRANSFERRING_RINGING': 18,
    'TRANSFERRING_BUSY': 19,
    'TRANSFERRING_ACTIVE': 20,
    'TRANSFERRING_HOLD': 21,
    'BLIND_TRANSFER_HOLD': 22,
    'BLIND_TRANSFERRING_DIAL_TONE': 23,
    'BLIND_TRANSFERRING_DIALING': 24,
    'BLIND_TRANSFERRING_INPROCESS': 25,
    'BLIND_TRANSFERRING_RINGING': 26
}

NotificationIndicationTypes = {
    'HOLD': 0,
    'MUTE': 1,
    'PARTY_LINE': 2,
    'RINGBACK': 3,
    'RINGING': 4,
    'CALL_WAITING_CALLEE': 5,
    'BUSY': 6,
    'WARNING_ERR': 7
}

NotificationIndicationStatusTypes = {
    'OFF': 0,
    'ON': 1,
    'BLINK': 2
}

CallForwardWaitingStates = {
    'INACTIVE': 0,
    'INPROCESS': 1,
    'ACTIVE': 2
}

CallForwardWaitingStatus = {
    'DISABLED': 0,
    'UNCONDITIONAL': 1,
    'BUSY': 2,
    'NO_REPLY': 3
}

DNDStatus = {
    'DND_NOT_SET': 0,
    'DND_CALL_MUTE': 1,
    'DND_CALL_REJECT': 2
}

S4BListType = {
    'UNKNOWN': 0,
    'GROUPS': 1,
    'CONTACTS': 2,
    'INFO_ONLY': 3
}

S4BListFunction = {
    'UNKNOWN': 0,
    'SELECT_GROUP': 1,
    'SELECT_CONTACT': 2,
    'ADD_TO_GROUP': 3,
    'SEARCHING': 4
}

S4BRegState = {
    'UNKNOWN': 0,
    'REGISTERED': 1,
    'NOT_REGISTERED': 2,
    'NO_PROXY_DEFINED': 3
}

S4BSessionStatus = {
    'UNKNOWN': 0,
    'IDLE': 1,
    'CONNECTING': 2,
    'RINGING': 3,
    'BUSY': 4,
    'ACTIVE': 5,
    'HOLD': 6,
    'INCOMING': 7,
    'CONFERENCE_JOIN': 8,
    'INVITE_JOIN_AUDIO': 9
}

################################################################################
## CLASS: DeviceClass
################################################################################

class DeviceClass():

# --------------------------------------------------------------------------
# CONSTRUCTOR
# --------------------------------------------------------------------------

    def __init__(self, *args, **kwargs):
        self.Username = ''
        self.Password = ''
        self.ReceiveData = self.__ReceiveData
        self.Buffer = b''
        self.MyQueue = Queue(self)
        self.EnableDebug = False
        self.EnablePoll = False
        self.IsCommunicating = False
        self.IsInitialized = False
        self.IsInitializing = False
        self.HeartbeatTimer = None
        self.HeartbeatTime = 30
        self.ConnectTimer = None
        self.ConnectTime = 10
        self.PollTimer = None
        self.PollTime = 15
        self.ResponseTimer = None
        self.ResponseTime = 5
        self.StrikeCount = 0
        self.IsParsing = False
        self.LastOut = None
        self.LastIn = None
        self.ComponentID = 0
        self.Components = []
        self.ComponentLock = False
        
        self.__ModuleInfo = self.__unassigned    
        
    """Default callback for non-defined events """
    def __unassigned(self, *args):
        pass
    
    @property
    def ModuleInfo(self):
        return self.__ModuleInfo
    
    @ModuleInfo.setter
    def ModuleInfo(self, handler):
        if callable(handler):
            self.__ModuleInfo = handler
        else:
            raise ValueError('handler must be a function')
            
# --------------------------------------------------------------------------
# CREDENTIAL HELPERS
# --------------------------------------------------------------------------

    def SetCredentials(self, *args, **kwargs):
        try:
            self.Username = kwargs.get('Username', '')
            self.Password = kwargs.get('Password', '')
            
        except Exception as e:
            print('Exception.SetCredentials: ' + str(e))

# --------------------------------------------------------------------------
# COMMAND HELPERS
# --------------------------------------------------------------------------

    def SendNextItem(self):
        try:
            temp = self.MyQueue.Dequeue()
            if isinstance(temp, MessageBundle):
                if(temp.message.find(COMMAND_TERMINATOR) < 0):
                    temp.message += COMMAND_TERMINATOR

                self.LastOut = temp
                self.DebugPrint('SendNextItem: ' + temp.message.replace(COMMAND_TERMINATOR, ''))
                self.Send(temp.message)
                self.StartResponseTimer()
        
        except Exception as e:
            print('Exception.SendNextItem: ' + str(e))

# --------------------------------------------------------------------------
# RESPONSE HELPERS
# --------------------------------------------------------------------------

    def FailedResponse(self):
        if self.LastOut is not None:
            self.DebugPrint('FailedResponse')
            self.StrikeCount += 1
        
            if self.StrikeCount ==  3:
                self.Reinit()
            
            else:
                self.KillResponseTimer()
                self.LastOut = None
                self.LastIn = None
                self.MyQueue.IsBusy = False
                self.SendNextItem()
        
            if self.MyQueue.IsEmpty():
                self.HeartbeatTimer = Wait(3, self.SendHeartbeat)

    def GoodResponse(self):
        self.DebugPrint('GoodResponse')
        self.StrikeCount = 0
        self.KillConnectTimer()
        self.KillResponseTimer()
        self.LastOut = None
        self.LastIn = None
        self.MyQueue.IsBusy = False
        self.SendNextItem()

    def StartResponseTimer(self):
        self.ResponseTimer = Wait(self.ResponseTime, self.FailedResponse)

    def KillResponseTimer(self):        
        if self.ResponseTimer is not None:
            self.ResponseTimer.Cancel()
            self.ResponseTimer = None

    def __ReceiveData(self, interface, RcvString):
        try:
            self.DebugPrint('__ReceiveData: ' + str(RcvString))
            if not self.IsCommunicating:
                self.IsCommunicating = True
                self.__ModuleInfo(self, self.IsCommunicating, self.IsInitialized)
            
            self.Buffer += RcvString
            location = -1
            
            if not self.IsParsing:
                self.IsParsing = True
                
                if self.Buffer.decode().find(RESPONSE_USERNAME) >= 0:
                    self.DebugPrint('ParseResponse: ' + RESPONSE_USERNAME.replace('\r\n', ''))
                    self.Buffer = self.Buffer[len(RESPONSE_USERNAME):]
                    self.MyQueue.IsBusy = False
                    self.MyQueue.Enqueue(MessageBundle(self, self.Username), STATUS_COMMAND)
                    self.IsParsing = False
                    self.LastOut = None
                    self.KillResponseTimer()
                
                elif self.Buffer.decode().find(RESPONSE_PASSWORD) >= 0:
                    self.DebugPrint('ParseResponse: ' + RESPONSE_PASSWORD.replace('\r\n', ''))
                    self.Buffer = self.Buffer[len(RESPONSE_PASSWORD):]
                    self.MyQueue.IsBusy = False
                    self.MyQueue.Enqueue(MessageBundle(self, self.Password), STATUS_COMMAND)
                    self.IsParsing = False
                    self.LastOut = None
                    self.KillResponseTimer()
                    
                elif self.Buffer.decode().find(RESPONSE_TERMINATOR) == 0:
                    self.Buffer = b''
                    self.MyQueue.IsBusy = False
                    self.IsParsing = False
                    
                    if self.LastOut is None:
                        self.KillResponseTimer()
                        
                        if not self.IsInitialized:
                            self.GetInitialized()
                            if self.IsInitializing and self.IsInitializedCheck():
                                self.IsInitializing = False
                                self.IsInitialized = True
                                self.__ModuleInfo(self, self.IsCommunicating, self.IsInitialized)
                                self.StartPollTimer()
                                    
                    elif self.LastOut.message == COMMAND_TERMINATOR:
                        self.GoodResponse()
                        
                        if isinstance(self, SerialClass) and not self.IsInitialized:
                            self.GetInitialized()
                
                else:
                    while True:
                        location = self.Buffer.decode().find(RESPONSE_TERMINATOR)
                        
                        if location < 0:
                            break
                            
                        end = location + len(RESPONSE_TERMINATOR)                    
                        response = self.Buffer[:end].decode()
                        self.Buffer = self.Buffer[end:]
                        self.LastIn = response
                        
                        if self.LastOut is not None:
                            self.LastOut.sender.ProcessResponse(response)
                        
                        self.DebugPrint('ParseResponse: ' + response.replace(RESPONSE_TERMINATOR, ''))
                        
                        parsed = ParserUtil.ParseResponse(response.replace(RESPONSE_TERMINATOR, ''))
                        self.DebugPrint('Parsed: ' + str(parsed))
                        
                        if isinstance(parsed, ResponseMsg) and parsed.Type != ResponseMessageType['UNKNOWN']:
                            self.ProcessSubscription(parsed)

                            if self.LastOut is None:
                                self.KillResponseTimer()
                                
                                if not self.IsInitialized:
                                    self.GetInitialized()
    
                                    if self.IsInitializing and self.IsInitializedCheck():
                                        self.IsInitializing = False
                                        self.IsInitialized = True
                                        self.__ModuleInfo(self, self.IsCommunicating, self.IsInitialized)
                                        self.StartPollTimer()
                            
                            else:
                                if isinstance(self.LastOut, MessageBundle) and self.LastOut.sender.IsGoodResponse(parsed):
                                    self.GoodResponse()
                                    
                                    if self.IsInitializing and self.IsInitializedCheck():
                                        self.IsInitializing = False
                                        self.IsInitialized = True
                                        self.__ModuleInfo(self, self.IsCommunicating, self.IsInitialized)
                                        self.StartPollTimer()
                    
                    self.IsParsing = False
        
        except Exception as e:
            print('Exception.ReceiveData: ' + str(e))              

    def ProcessSubscription(self, msg):
        for component in self.Components:
            component.ProcessSubscription(msg)

# --------------------------------------------------------------------------
# DEBUG HELPERS
# --------------------------------------------------------------------------

    def SetDebug(self, state):
        self.EnableDebug = state

    def DebugPrint(self, msg):
        if self.EnableDebug:
            print('ClearOnePro2[' + str(msg)  + ']')

# --------------------------------------------------------------------------
# INITIALIZATION HELPERS
# --------------------------------------------------------------------------

    def Reinit(self):
        self.DebugPrint('Reinit')
        self.IsCommunicating = False
        self.IsInitializing = False
        self.IsInitialized = False
        self.StrikeCount = 0
        self.IsParsing = False
        self.LastIn = None
        self.LastOut = None
        self.Buffer = b''
        self.MyQueue = Queue(self)
        self.KillHeartbeatTimer()
        self.KillResponseTimer()
        self.KillPollTimer()
        
        self.__ModuleInfo(self, self.IsCommunicating, self.IsInitialized)
        
        for component in self.Components:
            component.Reinitialize()
            
        self.ConnectToDevice()
        self.StartHeartbeatTimer()

    def IsInitializedCheck(self):
        if self.IsCommunicating == False:
            return False
        
        for component in self.Components:
            if component.IsInitialized() == False:
                return False
        
        return True

    def GetInitialized(self):
        if not self.IsInitializing:
            self.IsInitializing = True
            
            for component in self.Components:
                component.GetInitialized()

# --------------------------------------------------------------------------
# HEARTBEAT HELPERS
# --------------------------------------------------------------------------

    def SendHeartbeat(self):
        self.DebugPrint('SendHeartbeat')
        if self.MyQueue.IsEmpty():
            self.DebugPrint('queue is empty')
            self.MyQueue.Enqueue(MessageBundle(self, COMMAND_TERMINATOR), STATUS_QUERY)
            self.StartHeartbeatTimer()

    def StartHeartbeatTimer(self):
        self.HeartbeatTimer = Wait(self.HeartbeatTime, self.SendHeartbeat)

    def KillHeartbeatTimer(self):
        if self.HeartbeatTimer is not None:
            self.HeartbeatTimer.Cancel()
            self.HeartbeatTimer = None

# --------------------------------------------------------------------------
# CONNECTION HELPERS
# --------------------------------------------------------------------------

    def ConnectIP(self):
        self.DebugPrint('ConnectIP')
        result = self.Connect()

        if result != 'Connected':
            self.StartConnectTimer()
        
        else:
            self.KillConnectTimer()

    def Connect232(self):
        self.DebugPrint('Connect232')
        self.SendHeartbeat()

    def ConnectToDevice(self):
        self.DebugPrint('ConnectToDevice')
        
        if not self.IsCommunicating:
            if isinstance(self, EthernetClass):
                self.ConnectIP()
            
            else:
                self.Connect232()

    def StartConnectTimer(self):
        self.ConnectTimer = Wait(self.ConnectTime, self.ConnectToDevice)

    def KillConnectTimer(self):
        if self.ConnectTimer is not None:
            self.ConnectTimer.Cancel()
            self.ConnectTimer = None

# --------------------------------------------------------------------------
# POLLING HELPERS
# --------------------------------------------------------------------------

    def SetPoll(self, state):
        self.EnablePoll = state
        
    def SendPoll(self):
        if self.MyQueue.IsEmpty() and self.EnablePoll:
            for component in self.Components:
                component.Poll()
            
        self.StartPollTimer()

    def StartPollTimer(self):
        self.PollTimer = Wait(self.PollTime, self.SendPoll)

    def KillPollTimer(self):
        if self.PollTimer is not None:
            self.PollTimer.Cancel()
            self.PollTimer = None

# --------------------------------------------------------------------------
# COMPONENT HELPERS
# --------------------------------------------------------------------------

    def RegisterComponent(self, component):
        try:
            if not self.ComponentLock:
                self.ComponentLock = True
                
                if not component in self.Components:
                    self.ComponentID += 1
                    component.ID = self.ComponentID
                    self.Components.insert(0, component)
        
        except Exception as e:
            print('Exception.RegisterComponent: ' + str(e))
        
        finally:
            self.ComponentLock = False

    def AddEPGenericComponent(self, SignalType, EndPointName, BlockNumber, ParameterName, IsPollable=True):
        try:
            newComponent = EPGenericComponent(self, SignalType, EndPointName, BlockNumber, ParameterName, IsPollable)
            self.RegisterComponent(newComponent)
            
            return newComponent
        
        except Exception as e:
            print('Exception.AddEPGenericComponent: ' + str(e))
    
    def AddRoomGenericComponent(self, SignalType, RoomNumber, Option, P1=None, P2=None):
        try:
            newComponent = RoomGenericComponent(self, SignalType, RoomNumber, Option, P1, P2)
            self.RegisterComponent(newComponent)
            
            return newComponent
        
        except Exception as e:
            print('Exception.AddRoomGenericComponent: ' + str(e))
    
    def AddGainMuteComponent(self, GainType, EndPointName, StepValue):
        try:
            newComponent = GainMuteComponent(self, GainType, EndPointName, StepValue)
            self.RegisterComponent(newComponent)
            
            return newComponent
        
        except Exception as e:
            print('Exception.AddGainMuteComponent: ' + str(e))

    def AddMacroComponent(self):
        try:
            newComponent = MacroComponent(self)
            self.RegisterComponent(newComponent)
            
            return newComponent
        
        except Exception as e:
            print('Exception.AddMacroComponent: ' + str(e))
     
    def AddPresetComponent(self, RoomNumber, RoomPartitionName):
        try:
            newComponent = PresetComponent(self, RoomNumber, RoomPartitionName)
            self.RegisterComponent(newComponent)
            
            return newComponent
        
        except Exception as e:
            print('Exception.AddPresetComponent: ' + str(e))
    
    def AddMatrixComponent(self, InputEndPointName, OutputEndPointName, CrosspointType, StepValue):
        try:
            newComponent = MatrixComponent(self, InputEndPointName, OutputEndPointName, CrosspointType, StepValue)
            self.RegisterComponent(newComponent)
            
            return newComponent
        
        except Exception as e:
            print('Exception.AddGainMuteComponent: ' + str(e))
    
    def AddAnalogDialerComponent(self, EndPointName):
        try:
            newComponent = AnalogDialerComponent(self, EndPointName)
            self.RegisterComponent(newComponent)
            
            return newComponent
        
        except Exception as e:
            print('Exception.AddAnalogDialerComponent: ' + str(e))

    def AddVoIPDialerComponent(self, EndPointName):
        try:
            newComponent = VoIPDialerComponent(self, EndPointName)
            self.RegisterComponent(newComponent)
            
            return newComponent
        
        except Exception as e:
            print('Exception.AddVoIPDialerComponent: ' + str(e))

    def AddSkypeComponent(self, EndPointName):
        try:
            newComponent = SkypeComponent(self, EndPointName)
            self.RegisterComponent(newComponent)
            
            return newComponent
        
        except Exception as e:
            print('Exception.AddSkypeComponent: ' + str(e))
            
################################################################################
## CLASS: MessageBundle
################################################################################

class MessageBundle():
    def __init__(self, sender, message):
        try:
            self.sender = sender
            self.message = message
        
        except Exception as e:
            print('Exception.MessageBundle.__init__: ' + str(e)) 
        
################################################################################
## CLASS: EPMsg
################################################################################

class EPMsg():
    def __init__(self, EndPointName=None, BlockNumber=None, ParameterName=None, Msg=None):
        try:
            self.Value = None
            self.EndPointName = EndPointName
            self.BlockNumber = BlockNumber
            self.ParameterName = ParameterName
                
            if isinstance(Msg, EPMsg):
                self.EndPointName = Msg.EndPointName
                self.BlockNumber = Msg.BlockNumber
                self.ParameterName = Msg.ParameterName
        
        except Exception as e:
            print('Exception.EPMsg.__init__: ' + str(e)) 
            
    def BuildQueryMsg(self):
        return 'EP {0} {1} {2}'.format(self.EndPointName, self.BlockNumber, self.ParameterName)
    
    def BuildSetMsg(self):
        return 'EP {0} {1} {2} {3}'.format(self.EndPointName, self.BlockNumber, self.ParameterName, self.Value)
    
    def Reinitialize(self):
        self.Value = None
    
    def IsInitialized(self):
        return self.Value is not None

################################################################################
## CLASS: RoomMsg
################################################################################

class RoomMsg():
    def __init__(self, RoomNumber=None, Option=None, P1=None, P2=None, Msg=None):
        try:
            self.Value = None
            self.RoomNumber = RoomNumber
            self.Option = RoomOptions.get(Option, RoomOptions['Unknown'])
            self.P1 = P1
            self.P2 = P2
                
            if isinstance(Msg, RoomMsg):
                self.RoomNumber = Msg.RoomNumber
                self.Option = Msg.Option
                self.P1 = Msg.P1
                self.P2 = Msg.P2     
        
        except Exception as e:
            print('Exception.RoomMsg.__init__: ' + str(e))                

    def BuildQueryMsg(self):
        if self.Option == RoomOptions['IndividualDividerState'] or self.Option == RoomOptions['DividerPolarity']:
            return 'ROOM {0} {1} {2}'.format(self.RoomNumber, self.Option, self.P1)
        
        elif self.Option == RoomOptions['Mode'] or self.Option == RoomOptions['DividerState']:
            return 'ROOM {0} {1}'.format(self.RoomNumber, self.Option)
        
        else:
            return ''
    
    def BuildSetMsg(self):
        if self.Option == RoomOptions['IndividualDividerState'] or self.Option == RoomOptions['DividerPolarity']:
            return 'ROOM {0} {1} {2} {3}'.format(self.RoomNumber, self.Option, self.P1, self.Value)
        
        elif self.Option == RoomOptions['Select']:
            return 'ROOM {0} {1} {2} {3} 1'.format(self.RoomNumber, self.Option, self.P1, self.Value)
        
        elif self.Option == RoomOptions['Mode'] or self.Option == RoomOptions['DividerState']:
            return 'ROOM {0} {1} {2}'.format(self.RoomNumber, self.Option, self.Value)
        
        else:
            return ''
    
    def Reinitialize(self):
        self.Value = None
    
    def IsInitialized(self):
        return self.Value is not None

################################################################################
## CLASS: MTMsg
################################################################################

class MTMsg():
    def __init__(self, InputEndPointName=None, OutputEndPointName=None, Msg=None):
        try:
            self.State = None
            self.Attenuation = None
            self.CrosspointType = CrosspointTypes['Unknown']
            self.InputEndPointName = InputEndPointName
            self.OutputEndPointName = OutputEndPointName
                
            if isinstance(Msg, MTMsg):
                self.InputEndPointName = Msg.InputEndPointName
                self.OutputEndPointName = Msg.OutputEndPointName 
        
        except Exception as e:
            print('Exception.MTMsg.__init__: ' + str(e))                 

    def BuildQueryMsg(self):
        return 'MT {0} {1}'.format(self.InputEndPointName, self.OutputEndPointName)
    
    def BuildSetMsg(self):
        return 'MT {0} {1} {2} {3} {4}'.format(self.InputEndPointName, self.OutputEndPointName, self.State, self.Attenuation, self.CrosspointType)
    
    def Reinitialize(self):
        self.State = None
        self.Attenuation = None
        self.CrosspointType = CrosspointTypes['Unknown']
    
    def IsInitialized(self):
        if self.State is None or self.Attenuation is None or self.CrosspointType == CrosspointTypes['Unknown']:
            return False
        
        return True

################################################################################
## CLASS: MCCFMsg
################################################################################

class MCCFMsg():
    def __init__(self, Msg=None):
        try:
            self.MacroName = None
            
            if isinstance(Msg, MCCFMsg):
                self.MacroName = Msg.MacroName  
        
        except Exception as e:
            print('Exception.MCCFMsg.__init__: ' + str(e))                

    def BuildQueryMsg(self):
        return ''
    
    def BuildSetMsg(self):
        return 'MCCF {0}'.format(self.MacroName)
    
    def Reinitialize(self):
        pass
    
    def IsInitialized(self):
        return True
        
################################################################################
## CLASS: BoxMsg
################################################################################

class BoxMsg():
    def __init__(self, BoxName=None, BlockNumber=None, ParameterName=None, Msg=None):
        try:
            self.Value = None
            self.BoxName = BoxName
            self.BlockNumber = BlockNumber
            self.ParameterName = ParameterName
                
            if isinstance(Msg, BoxMsg):
                self.BoxName = Msg.BoxName
                self.BlockNumber = Msg.BlockNumber
                self.ParameterName = Msg.ParameterName
        
        except Exception as e:
            print('Exception.BoxMsg.__init__: ' + str(e))                 

    def BuildQueryMsg(self):
        return 'BOX {0} {1} {2}'.format(self.EndPointName, self.BlockNumber, self.ParameterName)
    
    def BuildSetMsg(self):
        return 'BOX {0} {1} {2} {3}'.format(self.EndPointName, self.BlockNumber, self.ParameterName, self.Value)
    
    def Reinitialize(self):
        self.Value = None
    
    def IsInitialized(self):
        return self.Value is not None

################################################################################
## CLASS: StackMsg
################################################################################

class StackMsg():
    def __init__(self, BlockNumber=None, ParameterName=None, Msg=None):
        try:
            self.Value = None
            self.BlockNumber = BlockNumber
            self.ParameterName = ParameterName
                
            if isinstance(Msg, StackMsg):
                self.BlockNumber = Msg.BlockNumber
                self.ParameterName = Msg.ParameterName      
        
        except Exception as e:
            print('Exception.StackMsg.__init__: ' + str(e))                

    def BuildQueryMsg(self):
        return 'STACK {0} {1}'.format(self.BlockNumber, self.ParameterName)
    
    def BuildSetMsg(self):
        return 'STACK {0} {1} {2}'.format(self.BlockNumber, self.ParameterName, self.Value)
    
    def Reinitialize(self):
        self.Value = None
    
    def IsInitialized(self):
        return self.Value is not None

################################################################################
## CLASS: ResponseMsg
################################################################################

class ResponseMsg():
    def __init__(self, Msg=None):
        try:
            self.Type = ResponseMessageType['UNKNOWN']
            self.Elements = []
            
            if isinstance(Msg, str):
                if len(Msg) == 0:
                    self.Type = ResponseMessageType['PROMPT_ONLY']
                    return
                
                tempList = Msg.split()
                
                if isinstance(tempList, list) and len(tempList) > 0:
                    if tempList[0] in ResponseMessageType.keys():
                        self.Type = ResponseMessageType.get(tempList[0], ResponseMessageType['UNKNOWN'])
                        self.Elements = tempList[:]
                        self.Elements.pop(0)
                        
                        if self.Type == ResponseMessageType['EP'] and len(self.Elements) > 1:
                            if self.Elements[0] in EndpointType.keys():
                                self.Elements[0] += ' {0}'.format(self.Elements[1])
                                self.Elements.pop(1)
                        
                        if len(self.Elements) > 0:
                            if len(self.Elements[len(self.Elements) - 1]) == 0:
                                self.Elements.pop(len(self.Elements - 1))     
        
        except Exception as e:
            print('Exception.ResponseMsg.__init__: ' + str(e))

    def __str__(self):
        msgList = []
        msgList.append('Type: {0} | '.format(self.Type))
     
        count = 0
     
        for item in self.Elements:
            msgList.append(',')
            msgList.append('Elements[{0}]: {1} | '.format(count, item))
            count += 1
        
        return ''.join(msgList).replace(',', '')

################################################################################
## CLASS: EPResponseMsg
################################################################################

class EPResponseMsg(ResponseMsg):
    def __init__(self, Msg=None):
        try:
            ResponseMsg.__init__(self, Msg)
            self.Message = EPMsg()
            
            if isinstance(Msg, ResponseMsg):
                self.Type = Msg.Type
                self.Elements = Msg.Elements[:]
                
            self.ParseElements()
        
        except Exception as e:
            print('Exception.EPResponseMsg.__init__: ' + str(e))

    def ParseElements(self):
        try:
            if isinstance(self.Elements, list):
                if len(self.Elements) >= 3:
                    tempList = self.Elements[:]
                    
                    self.Message.EndPointName = tempList[0]
                    tempList.pop(0)
                    
                    self.Message.BlockNumber = tempList[0]
                    tempList.pop(0)
                    
                    self.Message.ParameterName = tempList[0]
                    tempList.pop(0)
                    
                    self.Message.Value = ''
                    
                    for i in range(len(tempList)):
                        self.Message.Value += tempList[i]
                        
                        if i < (len(tempList) - 1):
                            self.Message.Value += ' '
        
        except Exception as e:
            print('Exception.EPResponseMsg.ParseElements: ' + str(e))
        
    def __str__(self):
        msgList = []
        msgList.append(ResponseMsg.__str__(self))
        msgList.append('EndPointName: {0} | '.format(self.Message.EndPointName))
        msgList.append('BlockNumber: {0} | '.format(self.Message.BlockNumber))
        msgList.append('ParameterName: {0} | '.format(self.Message.ParameterName))
        msgList.append('Value: {0}'.format(self.Message.Value))
                
        return ''.join(msgList)
    
    def __eq__(self, other): 
        return self.Message.BlockNumber == other.BlockNumber and \
               self.Message.EndPointName == other.EndPointName and \
               self.Message.ParameterName == other.ParameterName

################################################################################
## CLASS: RoomResponseMsg
################################################################################

class RoomResponseMsg(ResponseMsg):
    def __init__(self, Msg=None):
        try:
            ResponseMsg.__init__(self, Msg)
            self.Message = RoomMsg()
            
            if isinstance(Msg, ResponseMsg):
                self.Type = Msg.Type
                self.Elements = Msg.Elements[:]
                
            self.ParseElements()
        
        except Exception as e:
            print('Exception.RoomResponseMsg.__init__: ' + str(e))

    def ParseElements(self):
        try:
            if isinstance(self.Elements, list):
                if len(self.Elements) >= 2:
                    tempList = self.Elements[:]
                    
                    self.Message.RoomNumber = int(tempList[0])
                    tempList.pop(0)
                    
                    self.Message.Option = int(tempList[0])
                    tempList.pop(0)
                    
                    if self.Message.Option == RoomOptions['IndividualDividerState'] or self.Message.Option == RoomOptions['DividerPolarity']:
                        self.Message.P1 = tempList[0]
                        tempList.pop(0)
                        
                        self.Message.P2 = ''
                    
                    elif self.Message.Option == RoomOptions['Mode'] or self.Message.Option == RoomOptions['DividerState'] or self.Message.Option == RoomOptions['WallState']:
                        self.Message.P1 = ''
                        self.Message.P2 = ''
                    
                    elif self.Message.Option == RoomOptions['Select']:
                        self.Message.P1 = tempList[0]
                        tempList.pop(0)
                        
                    elif self.Message.Option == RoomOptions['PresetState']:
                        self.Message.P1 = tempList[0]
                        tempList.pop(0)
                        
                        self.Message.P2 = tempList[0]
                        tempList.pop(0)
                    
                    self.Message.Value = ''
                    
                    for i in range(len(tempList)):
                        self.Message.Value += tempList[i]
                        
                        if i < (len(tempList) - 1):
                            self.Message.Value += ' '
        
        except Exception as e:
            print('Exception.RoomResponseMsg.ParseElements: ' + str(e))
        
    def __str__(self):
        msgList = []
        msgList.append(ResponseMsg.__str__(self))
        msgList.append('RoomNumber: {0} | '.format(self.Message.RoomNumber))
        msgList.append('Option: {0} | '.format(self.Message.Option))
        msgList.append('P1: {0} | '.format(self.Message.P1))
        msgList.append('P2: {0} | '.format(self.Message.P2))
        msgList.append('Value: {0}'.format(self.Message.Value))
                
        return ''.join(msgList)
    
    def __eq__(self, other): 
        response = False
        
        if self.Message.Option == RoomOptions['IndividualDividerState'] or self.Message.Option == RoomOptions['DividerPolarity']:     
            response = self.Message.RoomNumber == other.RoomNumber and \
                       self.Message.Option == other.Option and \
                       str(self.Message.P1) == str(other.P1)
            
        elif self.Message.Option == RoomOptions['Mode'] or self.Message.Option == RoomOptions['DividerState']:
            response = self.Message.RoomNumber == other.RoomNumber and \
                       self.Message.Option == other.Option
        
        elif self.Message.Option == RoomOptions['Select']:
            response = self.Message.RoomNumber == other.RoomNumber and \
                       self.Message.Option == other.Option and \
                       str(self.Message.P1) == str(other.P1)
        
        return response

################################################################################
## CLASS: MTResponseMsg
################################################################################

class MTResponseMsg(ResponseMsg):
    def __init__(self, Msg=None):
        try:
            ResponseMsg.__init__(self, Msg)
            self.Message = MTMsg()
            
            if isinstance(Msg, ResponseMsg):
                self.Type = Msg.Type
                self.Elements = Msg.Elements[:]
                
            self.ParseElements()
        
        except Exception as e:
            print('Exception.MTResponseMsg.__init__: ' + str(e))

    def ParseElements(self):
        try:
            if isinstance(self.Elements, list):
                if len(self.Elements) >= 5:
                    tempList = self.Elements[:]
                    
                    self.Message.InputEndPointName = tempList[0]
                    tempList.pop(0)
                    
                    self.Message.OutputEndPointName = tempList[0]
                    tempList.pop(0)
                    
                    self.Message.State = tempList[0]
                    tempList.pop(0)
                    
                    self.Message.Attenuation = tempList[0]
                    tempList.pop(0)
                    
                    self.Message.CrosspointType = int(tempList[0]) if (int(tempList[0]) in CrosspointTypes.values()) else CrosspointTypes['Unknown']
                    tempList.pop(0)
        
        except Exception as e:
            print('Exception.MTResponseMsg.ParseElements: ' + str(e))
        
    def __str__(self):
        msgList = []
        msgList.append(ResponseMsg.__str__(self))
        msgList.append('InputEndPointName: {0} | '.format(self.Message.InputEndPointName))
        msgList.append('OutputEndPointName: {0} | '.format(self.Message.OutputEndPointName))
        msgList.append('State: {0} | '.format(self.Message.State))
        msgList.append('Attenuation: {0} | '.format(self.Message.Attenuation))
        msgList.append('CrosspointType: {0}'.format(self.Message.CrosspointType))
                
        return ''.join(msgList)
    
    def __eq__(self, other): 
            return self.Message.InputEndPointName == other.InputEndPointName and \
                   self.Message.OutputEndPointName == other.OutputEndPointName
        
################################################################################
## CLASS: MCCFResponseMsg
################################################################################

class MCCFResponseMsg(ResponseMsg):
    def __init__(self, Msg=None):
        try:
            ResponseMsg.__init__(self, Msg)
            self.Message = MCCFMsg()
            
            if isinstance(Msg, ResponseMsg):
                self.Type = Msg.Type
                self.Elements = Msg.Elements[:]
                
            self.ParseElements()
        
        except Exception as e:
            print('Exception.EPResponseMsg.__init__: ' + str(e))

    def ParseElements(self):
        try:
            if isinstance(self.Elements, list):
                if len(self.Elements) >= 1:
                    tempList = self.Elements[:]
                    
                    self.Message.MacroName = tempList[0]
                    tempList.pop(0)
        
        except Exception as e:
            print('Exception.MCCFResponseMsg.ParseElements: ' + str(e))
        
    def __str__(self):
        msgList = []
        msgList.append(ResponseMsg.__str__(self))
        msgList.append('MacroName: {0} | '.format(self.Message.MacroName))
                
        return ''.join(msgList)
    
    def __eq__(self, other): 
        return isinstance(other, MCCFMsg)

################################################################################
## CLASS: BoxResponseMsg
################################################################################

class BoxResponseMsg(ResponseMsg):
    def __init__(self, Msg=None):
        try:
            ResponseMsg.__init__(self, Msg)
            self.Message = BoxMsg()
            
            if isinstance(Msg, ResponseMsg):
                self.Type = Msg.Type
                self.Elements = Msg.Elements[:]
                
            self.ParseElements()
        
        except Exception as e:
            print('Exception.BoxResponseMsg.__init__: ' + str(e))

    def ParseElements(self):
        try:
            if isinstance(self.Elements, list):
                if len(self.Elements) >= 3:
                    tempList = self.Elements[:]
                    
                    self.Message.BoxName = tempList[0]
                    tempList.pop(0)
                    
                    self.Message.BlockNumber = tempList[0]
                    tempList.pop(0)
                    
                    self.Message.ParameterName = tempList[0]
                    tempList.pop(0)
                    
                    self.Message.Value = ''
                    
                    for i in range(len(tempList)):
                        self.Message.Value += tempList[i]
                        
                        if i < (len(tempList) - 1):
                            self.Message.Value += ' '
        
        except Exception as e:
            print('Exception.BoxResponseMsg.ParseElements: ' + str(e))
        
    def __str__(self):
        msgList = []
        msgList.append(ResponseMsg.__str__(self))
        msgList.append('BoxName: {0} | '.format(self.Message.BoxName))
        msgList.append('BlockNumber: {0} | '.format(self.Message.BlockNumber))
        msgList.append('ParameterName: {0} | '.format(self.Message.ParameterName))
        msgList.append('Value: {0}'.format(self.Message.Value))
                
        return ''.join(msgList)
    
    def __eq__(self, other): 
        return self.Message.BlockNumber == other.BlockNumber and \
               self.Message.BoxName == other.BoxName and \
               self.Message.ParameterName == other.ParameterName

################################################################################
## CLASS: StackResponseMsg
################################################################################

class StackResponseMsg(ResponseMsg):
    def __init__(self, Msg=None):
        try:
            ResponseMsg.__init__(self, Msg)
            self.Message = StackMsg()
            
            if isinstance(Msg, ResponseMsg):
                self.Type = Msg.Type
                self.Elements = Msg.Elements[:]
                
            self.ParseElements()
        
        except Exception as e:
            print('Exception.StackResponseMsg.__init__: ' + str(e))

    def ParseElements(self):
        try:
            if isinstance(self.Elements, list):
                if len(self.Elements) >= 2:
                    tempList = self.Elements[:]
                    
                    self.Message.BlockNumber = tempList[0]
                    tempList.pop(0)
                    
                    self.Message.ParameterName = tempList[0]
                    tempList.pop(0)
                    
                    self.Message.Value = ''
                    
                    for i in range(len(tempList)):
                        self.Message.Value += tempList[i]
                        
                        if i < (len(tempList) - 1):
                            self.Message.Value += ' '
        
        except Exception as e:
            print('Exception.StackResponseMsg.ParseElements: ' + str(e))
        
    def __str__(self):
        msgList = []
        msgList.append(ResponseMsg.__str__(self))
        msgList.append('BlockNumber: {0} | '.format(self.Message.BlockNumber))
        msgList.append('ParameterName: {0} | '.format(self.Message.ParameterName))
        msgList.append('Value: {0}'.format(self.Message.Value))
                
        return ''.join(msgList)
    
    def __eq__(self, other): 
        return self.Message.BlockNumber == other.BlockNumber and \
               self.Message.ParameterName == other.ParameterName

################################################################################
## CLASS: Component
################################################################################

class Component():
    def __init__(self, parent):
        try:
            self.ID = None
            self.parent = parent
        
        except Exception as e:
            print('Exception.Component.__init__: ' + str(e))

################################################################################
## CLASS: EPGenericComponent
################################################################################

class EPGenericComponent(Component):
    def __init__(self, parent, SignalType, EndPointName, BlockNumber, ParameterName, IsPollable):
        try:
            super().__init__(parent)
            self.IsPollable = IsPollable
            self.MySignalType = SignalTypes.get(SignalType, SignalTypes['Null'])
            self.MyControlPoint = EPMsg(EndPointName=EndPointName, BlockNumber=BlockNumber, ParameterName=ParameterName)
            
            self.__OnEPDigitalStateChange = self.__unassigned
            self.__OnEPAnalogStateChange = self.__unassigned
            self.__OnEPSerialStateChange = self.__unassigned
        
        except Exception as e:
            print('Exception.EPGenericComponent.__init__: ' + str(e))
    
    def __unassigned(self, *args):
        pass
        
    @property
    def OnEPDigitalStateChange(self):
        return self.__OnEPDigitalStateChange
    
    @OnEPDigitalStateChange.setter
    def OnEPDigitalStateChange(self, handler):
        if callable(handler):
            self.__OnEPDigitalStateChange = handler
        else:
            raise ValueError('handler must be a function')
         
    @property
    def OnEPAnalogStateChange(self):
        return self.__OnEPAnalogStateChange
    
    @OnEPAnalogStateChange.setter
    def OnEPAnalogStateChange(self, handler):
        if callable(handler):
            self.__OnEPAnalogStateChange = handler
        else:
            raise ValueError('handler must be a function')
    
    @property
    def OnEPSerialStateChange(self):
        return self.__OnEPSerialStateChange
    
    @OnEPSerialStateChange.setter
    def OnEPSerialStateChange(self, handler):
        if callable(handler):
            self.__OnEPSerialStateChange = handler
        else:
            raise ValueError('handler must be a function')

    def SetDigital(self, state):
        if self.IsInitialized() and self.MySignalType == SignalTypes['Digital']:
            req = EPMsg(Msg=self.MyControlPoint)
            req.Value = 1 if int(state) == 1 else 0
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def GetDigital(self):
        if self.IsInitialized() and self.MySignalType == SignalTypes['Digital']:
            return int(self.MyControlPoint.Value) if self.MyControlPoint.Value is not None else 0
    
    def ToggleDigital(self):
        if self.IsInitialized() and self.MySignalType == SignalTypes['Digital']:
            req = EPMsg(Msg=self.MyControlPoint)          
            req.Value = 1 if int(self.MyControlPoint.Value) == 0 else 0
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)

    def SetAnalog(self, value):
        if self.IsInitialized() and self.MySignalType == SignalTypes['Analog']:
            req = EPMsg(Msg=self.MyControlPoint)
            req.Value = value
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def GetAnalog(self):
        if self.IsInitialized() and self.MySignalType == SignalTypes['Analog']:
            return int(self.MyControlPoint.Value) if self.MyControlPoint.Value is not None else 0
            
    def SetSerial(self, value):
        if self.IsInitialized() and self.MySignalType == SignalTypes['Serial']:
            if value != '':
                req = EPMsg(Msg=self.MyControlPoint)
                req.Value = value
                
                self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def GetSerial(self):
        if self.IsInitialized() and self.MySignalType == SignalTypes['Serial']:
            return str(self.MyControlPoint.Value) if self.MyControlPoint.Value is not None else ''
            
    def Poll(self):
        if self.IsInitialized() and self.IsPollable:
            self.parent.MyQueue.Enqueue(MessageBundle(self, self.MyControlPoint.BuildQueryMsg()), STATUS_QUERY)
        
    def Reinitialize(self):
        self.MyControlPoint.Reinitialize()  
        self.UpdateStateChange(None, True)
    
    def Refresh(self):
        self.UpdateStateChange(self.MyControlPoint.Value, True)
    
    def GetInitialized(self):
        if self.IsPollable:
            self.parent.MyQueue.Enqueue(MessageBundle(self, self.MyControlPoint.BuildQueryMsg()), STATUS_QUERY)
    
    def IsInitialized(self):
        if self.IsPollable:
            return self.MyControlPoint.IsInitialized()

        return True
    
    def ProcessSubscription(self, msg):
        if isinstance(msg, EPResponseMsg) and msg == self.MyControlPoint:
            self.UpdateStateChange(msg.Message.Value, False)
        
    def ProcessResponse(self, response):
        pass
    
    def IsGoodResponse(self, response):
        if isinstance(response, EPResponseMsg):
            return response == self.MyControlPoint
        
        return False
    
    def UpdateStateChange(self, value, force):
        try:
            if force or value != self.MyControlPoint.Value:
                if self.MySignalType == SignalTypes['Digital']:
                    self.MyControlPoint.Value = int(value) if value is not None else 0
                    self.__OnEPDigitalStateChange(self, self.MyControlPoint.Value)
                
                elif self.MySignalType == SignalTypes['Analog']:
                    self.MyControlPoint.Value = int(value) if value is not None else 0
                    self.__OnEPAnalogStateChange(self, self.MyControlPoint.Value)
                
                elif self.MySignalType == SignalTypes['Serial']:
                    self.MyControlPoint.Value = str(value) if value is not None else ''
                    self.__OnEPSerialStateChange(self, self.MyControlPoint.Value)
        
        except Exception as e:
            print('Exception.EPGenericComponent.UpdateStateChange: ' + str(e))

################################################################################
## CLASS: RoomGenericComponent
################################################################################

class RoomGenericComponent(Component):
    def __init__(self, parent, SignalType, RoomNumber, Option, P1, P2):
        try:
            super().__init__(parent)
            self.MySignalType = SignalTypes.get(SignalType, SignalTypes['Null'])
            self.MyControlPoint = RoomMsg(RoomNumber=RoomNumber, Option=Option, P1=P1, P2=P2)
            
            self.__OnRoomDigitalStateChange = self.__unassigned
            self.__OnRoomAnalogStateChange = self.__unassigned
            self.__OnRoomSerialStateChange = self.__unassigned
        
        except Exception as e:
            print('Exception.RoomGenericComponent.__init__: ' + str(e))
    
    def __unassigned(self, *args):
        pass
        
    @property
    def OnRoomDigitalStateChange(self):
        return self.__OnRoomDigitalStateChange
    
    @OnRoomDigitalStateChange.setter
    def OnRoomDigitalStateChange(self, handler):
        if callable(handler):
            self.__OnRoomDigitalStateChange = handler
        else:
            raise ValueError('handler must be a function')
    
    @property
    def OnRoomAnalogStateChange(self):
        return self.__OnRoomAnalogStateChange
    
    @OnRoomAnalogStateChange.setter
    def OnRoomAnalogStateChange(self, handler):
        if callable(handler):
            self.__OnRoomAnalogStateChange = handler
        else:
            raise ValueError('handler must be a function')
    
    @property
    def OnRoomSerialStateChange(self):
        return self.__OnRoomSerialStateChange
    
    @OnRoomSerialStateChange.setter
    def OnRoomSerialStateChange(self, handler):
        if callable(handler):
            self.__OnRoomSerialStateChange = handler
        else:
            raise ValueError('handler must be a function')
            
    def SetDigital(self, state):
        if self.IsInitialized() and self.MySignalType == SignalTypes['Digital']:
            req = RoomMsg(Msg=self.MyControlPoint)
            req.Value = 1 if int(state) == 1 else 0
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def GetDigital(self):
        if self.IsInitialized() and self.MySignalType == SignalTypes['Digital']:
            return int(self.MyControlPoint.Value) if self.MyControlPoint.Value is not None else 0
    
    def ToggleDigital(self):
        if self.IsInitialized() and self.MySignalType == SignalTypes['Digital']:
            req = RoomMsg(Msg=self.MyControlPoint)          
            req.Value = 1 if int(self.MyControlPoint.Value) == 0 else 0
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)

    def SetAnalog(self, value):
        if self.IsInitialized() and self.MySignalType == SignalTypes['Analog']:
            req = RoomMsg(Msg=self.MyControlPoint)
            req.Value = value
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def GetAnalog(self):
        if self.IsInitialized() and self.MySignalType == SignalTypes['Analog']:
            return int(self.MyControlPoint.Value) if self.MyControlPoint.Value is not None else 0
            
    def SetSerial(self, value):
        if self.IsInitialized() and self.MySignalType == SignalTypes['Serial']:
            if value != '':
                req = RoomMsg(Msg=self.MyControlPoint)
                req.Value = value
                
                self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def GetSerial(self):
        if self.IsInitialized() and self.MySignalType == SignalTypes['Serial']:
            return str(self.MyControlPoint.Value) if self.MyControlPoint.Value is not None else ''
            
    def Poll(self):
        if self.IsInitialized() and self.MyControlPoint.Option != RoomOptions['Select']:
            self.parent.MyQueue.Enqueue(MessageBundle(self, self.MyControlPoint.BuildQueryMsg()), STATUS_QUERY)
        
    def Reinitialize(self):
        self.MyControlPoint.Reinitialize()
        self.UpdateStateChange(None, True)
    
    def Refresh(self):
        self.UpdateStateChange(self.MyControlPoint.Value, True)
    
    def GetInitialized(self):
        if self.MyControlPoint.Option != RoomOptions['Select']:
            self.parent.MyQueue.Enqueue(MessageBundle(self, self.MyControlPoint.BuildQueryMsg()), STATUS_QUERY)
    
    def IsInitialized(self):
        if self.MyControlPoint.Option == RoomOptions['Select']:
            return True

        return self.MyControlPoint.IsInitialized()
    
    def ProcessSubscription(self, msg):
        if isinstance(msg, RoomResponseMsg) and msg == self.MyControlPoint:
            self.UpdateStateChange(msg.Message.Value, False)
        
    def ProcessResponse(self, response):
        pass
    
    def IsGoodResponse(self, response):
        if isinstance(response, RoomResponseMsg):
            return response == self.MyControlPoint
        
        return False
    
    def UpdateStateChange(self, value, force):
        try:
            if force or value != self.MyControlPoint.Value:
                if self.MySignalType == SignalTypes['Digital']:
                    self.MyControlPoint.Value = int(value) if value is not None else 0
                    self.__OnRoomDigitalStateChange(self, self.MyControlPoint.Value)
                
                elif self.MySignalType == SignalTypes['Analog']:
                    self.MyControlPoint.Value = int(value) if value is not None else 0
                    self.__OnRoomAnalogStateChange(self, self.MyControlPoint.Value)
                
                elif self.MySignalType == SignalTypes['Serial']:
                    self.MyControlPoint.Value = str(value) if value is not None else ''
                    self.__OnRoomSerialStateChange(self, self.MyControlPoint.Value)
        
        except Exception as e:
            print('Exception.RoomGenericComponent.UpdateStateChange: ' + str(e))
            
################################################################################
## CLASS: GainMuteComponent
################################################################################

class GainMuteComponent(Component):
    def __init__(self, parent, GainType, EndPointName, StepValue):
        try:
            super().__init__(parent)
            self.RampTimer = None
            self.RampTime = .3
            self.IsPollable = True
            self.MyGainType = GainTypes.get(GainType, GainTypes['Null'])
            self.MyStepValue = StepValue if StepValue > 0 else 1
            self.MyCurrentMute = None
            self.MyCurrentLevel = None
            self.MyMaxLevel = 20
            self.MyMinLevel = -65
            self.MyCurrentRampMode = RampTypes['Stop']
            self.MyMuteControlPoint = EPMsg(EndPointName=EndPointName, BlockNumber='LEVEL', ParameterName='MUTE')
            self.MyLevelControlPoint = EPMsg(EndPointName=EndPointName, BlockNumber='LEVEL', ParameterName=GainType)
            self.MyMaxLevelControlPoint = EPMsg(EndPointName=EndPointName, BlockNumber='LEVEL', ParameterName='MAX_GAIN')
            self.MyMinLevelControlPoint = EPMsg(EndPointName=EndPointName, BlockNumber='LEVEL', ParameterName='MIN_GAIN')
            self.MyControlPoints = [self.MyMuteControlPoint, self.MyMaxLevelControlPoint, self.MyMinLevelControlPoint, self.MyLevelControlPoint]
            
            self.__OnLevelDBChange = self.__unassigned
            self.__OnLevelPercentChange = self.__unassigned
            self.__OnMuteStateChange = self.__unassigned
        
        except Exception as e:
            print('Exception.GainMuteComponent.__init__: ' + str(e))
    
    def __unassigned(self, *args):
        pass
        
    @property
    def OnLevelDBChange(self):
        return self.__OnLevelDBChange
    
    @OnLevelDBChange.setter
    def OnLevelDBChange(self, handler):
        if callable(handler):
            self.__OnLevelDBChange = handler
        else:
            raise ValueError('handler must be a function')
    
    @property
    def OnLevelPercentChange(self):
        return self.__OnLevelPercentChange
    
    @OnLevelPercentChange.setter
    def OnLevelPercentChange(self, handler):
        if callable(handler):
            self.__OnLevelPercentChange = handler
        else:
            raise ValueError('handler must be a function')
    
    @property
    def OnMuteStateChange(self):
        return self.__OnMuteStateChange
    
    @OnMuteStateChange.setter
    def OnMuteStateChange(self, handler):
        if callable(handler):
            self.__OnMuteStateChange = handler
        else:
            raise ValueError('handler must be a function')
            
    def SetLevelPercent(self, value):
        if self.IsInitialized():
            req = EPMsg(Msg=self.MyLevelControlPoint)
            req.Value = self.ConvertFromPercent(value)
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def GetLevelPercent(self):
        if self.IsInitialized():
            return self.ConvertToPercent(self.MyCurrentLevel) if self.MyCurrentLevel is not None else 0
            
    def SetLevelDB(self, value):
        if self.IsInitialized():
            req = EPMsg(Msg=self.MyLevelControlPoint)
            req.Value = value
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def GetLevelDB(self):
        if self.IsInitialized():
            return float(self.MyCurrentLevel) if self.MyCurrentLevel is not None else self.MyMinLevel
            
    def Raise(self):
        if self.IsInitialized():
            self.RampLevel(RampTypes['Raise'])
    
    def Lower(self):
        if self.IsInitialized():
            self.RampLevel(RampTypes['Lower'])
    
    def Stop(self):
        if self.IsInitialized():
            self.RampLevel(RampTypes['Stop'])
    
    def RampLevel(self, rampType):
        self.MyCurrentRampMode = rampType
        
        if self.MyCurrentRampMode == RampTypes['Stop']:
            self.StopRampTimer()
            
        elif self.MyCurrentRampMode == RampTypes['Raise']:
            self.ResetRampTimer()
            self.SetLevelDB(self.GetNextRampLevel())
        
        elif self.MyCurrentRampMode == RampTypes['Lower']:
            self.ResetRampTimer()
            self.SetLevelDB(self.GetPrevRampLevel())

    def GetNextRampLevel(self):
        if self.MyCurrentLevel is not None:
            if (self.MyCurrentLevel + self.MyStepValue) > self.MyMaxLevel:
                return self.MyMaxLevel
                        
            return self.MyCurrentLevel + self.MyStepValue

    def GetPrevRampLevel(self):
        if self.MyCurrentLevel is not None:
            if (self.MyCurrentLevel - self.MyStepValue) < self.MyMinLevel:
                return self.MyMinLevel
                       
            return self.MyCurrentLevel - self.MyStepValue
            
    def StartRampTimer(self):
        if self.RampTimer is None:
            self.RampTimer = Wait(self.RampTime, self.DoRamp)

    def StopRampTimer(self):
        if self.RampTimer is not None:
            self.RampTimer.Cancel()
            self.RampTimer = None        

    def ResetRampTimer(self):
        self.StopRampTimer()
        self.StartRampTimer()

    def DoRamp(self):
        self.RampLevel(self.MyCurrentRampMode)

    def ConvertFromPercent(self, percent):
        a = float(((percent * (self.MyMaxLevel - self.MyMinLevel)) / 100) + self.MyMinLevel)

        return float(math.ceil(a * 2) / 2)

    def ConvertToPercent(self, db):
        a = float(((db - self.MyMinLevel) * 100) / (self.MyMaxLevel - self.MyMinLevel))

        return int(a)
    
    def SetMute(self, state):
        if self.IsInitialized():
            req = EPMsg(Msg=self.MyMuteControlPoint)
            req.Value = 1 if state == True else 0
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def GetMute(self):
        if self.IsInitialized():
            return int(self.MyMuteControlPoint.Value) if self.MyMuteControlPoint.Value is not None else 0
    
    def ToggleMute(self):
        if self.IsInitialized():
            req = EPMsg(Msg=self.MyMuteControlPoint)          
            req.Value = 1 if int(self.MyMuteControlPoint.Value) == 0 else 0

            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
            
    def Poll(self):
        if self.IsInitialized():
            for item in self.MyControlPoints:
                self.parent.MyQueue.Enqueue(MessageBundle(self, item.BuildQueryMsg()), STATUS_QUERY)
        
    def Reinitialize(self):
        if self.IsPollable:
            for item in self.MyControlPoints:
                item.Reinitialize()
        
        self.UpdateMuteChange(None, True)
        self.UpdateLevelChange(None, True)
    
    def Refresh(self):
        self.UpdateMuteChange(self.MyMuteControlPoint.Value, True)
        self.UpdateLevelChange(self.MyLevelControlPoint.Value, True)
    
    def GetInitialized(self):
        for item in self.MyControlPoints:
            self.parent.MyQueue.Enqueue(MessageBundle(self, item.BuildQueryMsg()), STATUS_QUERY)
    
    def IsInitialized(self):
        for item in self.MyControlPoints:
            if not item.IsInitialized():
                return False
        
        return True
    
    def ProcessSubscription(self, msg):
        if isinstance(msg, EPResponseMsg):
            for item in self.MyControlPoints:
                if msg == item:
                    if msg.Message.Value != item.Value:
                        item.Value = msg.Message.Value

                        if item == self.MyMuteControlPoint:
                            self.UpdateMuteChange(msg.Message.Value, False)
                        else:
                            self.UpdateLevelChange(msg.Message.Value, False)
                    break

    def ProcessResponse(self, response):
        pass
    
    def IsGoodResponse(self, response):
        if isinstance(response, EPResponseMsg):
            for item in self.MyControlPoints:
                if response == item:
                    return True

        return False
    
    def UpdateMuteChange(self, value, force):
        try:
            if force or value != self.MyCurrentMute:
                self.MyCurrentMute = int(value) if value is not None else 0
                self.__OnMuteStateChange(self, self.MyCurrentMute)

        except Exception as e:
            print('Exception.GainMuteComponent.UpdateMuteChange: ' + str(e))
    
    def UpdateLevelChange(self, value, force):
        try:
            if self.MyMaxLevelControlPoint.Value is not None:
                self.MyMaxLevel = float(self.MyMaxLevelControlPoint.Value)
            
            if self.MyMinLevelControlPoint.Value is not None:
                self.MyMinLevel = float(self.MyMinLevelControlPoint.Value)
            
            if force or value != self.MyCurrentLevel:
                self.MyCurrentLevel = float(value) if value is not None else self.MyMinLevel
                
                self.__OnLevelDBChange(self, self.MyCurrentLevel)
                self.__OnLevelPercentChange(self, self.ConvertToPercent(self.MyCurrentLevel))
                
                if self.MyCurrentLevel > self.MyMaxLevel:
                    self.SetLevelDB(self.MyMaxLevel)
                
                if self.MyCurrentLevel < self.MyMinLevel:
                    self.SetLevelDB(self.MyMinLevel)
        
        except Exception as e:
            print('Exception.GainMuteComponent.UpdateLevelChange: ' + str(e))

################################################################################
## CLASS: MatrixComponent
################################################################################

class MatrixComponent(Component):
    def __init__(self, parent, InputEndPointName, OutputEndPointName, CrosspointType, StepValue):
        try:
            super().__init__(parent)
            self.RampTimer = None
            self.RampTime = .3
            self.IsPollable = True
            self.MyCrosspointType = CrosspointTypes.get(CrosspointType, CrosspointTypes['Unknown'])
            self.MyStepValue = StepValue if StepValue > 0 else 1
            self.MyCurrentState = None
            self.MyCurrentLevel = None
            self.MyMaxLevel = 12
            self.MyMinLevel = -20
            self.MyCurrentRampMode = RampTypes['Stop']
            self.MyControlPoint = MTMsg(InputEndPointName=InputEndPointName, OutputEndPointName=OutputEndPointName)
            
            self.__OnAttenuationDBChange = self.__unassigned
            self.__OnAttenuationPercentChange = self.__unassigned
            self.__OnCrosspointChange = self.__unassigned
        
        except Exception as e:
            print('Exception.MatrixComponent.__init__: ' + str(e))
    
    def __unassigned(self, *args):
        pass
        
    @property
    def OnAttenuationDBChange(self):
        return self.__OnAttenuationDBChange
    
    @OnAttenuationDBChange.setter
    def OnAttenuationDBChange(self, handler):
        if callable(handler):
            self.__OnAttenuationDBChange = handler
        else:
            raise ValueError('handler must be a function')

    @property
    def OnAttenuationPercentChange(self):
        return self.__OnAttenuationPercentChange
    
    @OnAttenuationPercentChange.setter
    def OnAttenuationPercentChange(self, handler):
        if callable(handler):
            self.__OnAttenuationPercentChange = handler
        else:
            raise ValueError('handler must be a function')

    @property
    def OnCrosspointChange(self):
        return self.__OnCrosspointChange
    
    @OnCrosspointChange.setter
    def OnCrosspointChange(self, handler):
        if callable(handler):
            self.__OnCrosspointChange = handler
        else:
            raise ValueError('handler must be a function')
            
    def SetCrosspoint(self, state):
        if self.IsInitialized():
            req = MTMsg(Msg=self.MyControlPoint)
            req.CrosspointType = self.myCrosspointType
            req.Attenuation = self.myControlPoint.Attenuation
            req.State = 1 if state == True else 0
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def GetCrosspoint(self):
        if self.IsInitialized():
            return int(self.MyControlPoint.State) if self.MyControlPoint.State is not None else 0
    
    def ToggleCrosspoint(self):
        if self.IsInitialized():
            req = MTMsg(Msg=self.MyControlPoint)
            req.CrosspointType = self.MyCrosspointType
            req.Attenuation = self.MyControlPoint.Attenuation
            req.State = 1 if int(self.MyControlPoint.State) == 0 else 0
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)

    def ClearCrosspoint(self):
        if self.IsInitialized():
            req = MTMsg(Msg=self.MyControlPoint)
            req.CrosspointType = CrosspointTypes['None']
            req.Attenuation = 0
            req.State = 0
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def SetAttenuationDB(self, value):
        if self.IsInitialized():
            if int(value) >= self.MyMinLevel and int(value) <= self.MyMaxLevel:
                req = MTMsg(Msg=self.MyControlPoint)
                req.State = self.MyControlPoint.State
                req.CrosspointType = self.MyCrosspointType
                req.Attenuation = value
                
                self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def GetAttenuationDB(self):
        if self.IsInitialized():
            return float(self.MyCurrentLevel) if self.MyCurrentLevel is not None else self.MyMinLevel
            
    def SetAttenuationPercent(self, value):
        if self.IsInitialized():
            req = MTMsg(Msg=self.MyControlPoint)
            req.State = self.MyControlPoint.State
            req.CrosspointType = self.MyCrosspointType
            req.Attenuation = self.ConvertFromPercent(value)
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def GetAttenuationPercent(self):
        if self.IsInitialized():
            return self.ConvertToPercent(self.MyCurrentLevel) if self.MyCurrentLevel is not None else 0
            
    def Raise(self):
        if self.IsInitialized():
            self.RampLevel(RampTypes['Raise'])
    
    def Lower(self):
        if self.IsInitialized():
            self.RampLevel(RampTypes['Lower'])
    
    def Stop(self):
        if self.IsInitialized():
            self.RampLevel(RampTypes['Stop'])
    
    def RampLevel(self, rampType):
        self.MyCurrentRampMode = rampType
        
        if self.MyCurrentRampMode == RampTypes['Stop']:
            self.StopRampTimer()
            
        elif self.MyCurrentRampMode == RampTypes['Raise']:
            self.ResetRampTimer()
            self.SetAttenuationDB(self.GetNextRampLevel())
        
        elif self.MyCurrentRampMode == RampTypes['Lower']:
            self.ResetRampTimer()
            self.SetAttenuationDB(self.GetPrevRampLevel())

    def GetNextRampLevel(self):
        if self.MyCurrentLevel is not None:
            if (self.MyCurrentLevel + self.MyStepValue) > self.MyMaxLevel:
                return self.MyMaxLevel
                        
            return self.MyCurrentLevel + self.MyStepValue

    def GetPrevRampLevel(self):
        if self.MyCurrentLevel is not None:
            if (self.MyCurrentLevel - self.MyStepValue) < self.MyMinLevel:
                return self.MyMinLevel
                       
            return self.MyCurrentLevel - self.MyStepValue
            
    def StartRampTimer(self):
        if self.RampTimer is None:
            self.RampTimer = Wait(self.RampTime, self.DoRamp)

    def StopRampTimer(self):
        if self.RampTimer is not None:
            self.RampTimer.Cancel()
            self.RampTimer = None        

    def ResetRampTimer(self):
        self.StopRampTimer()
        self.StartRampTimer()

    def DoRamp(self):
        self.RampLevel(self.MyCurrentRampMode)

    def ConvertFromPercent(self, percent):
        a = float(((percent * (self.MyMaxLevel - self.MyMinLevel)) / 100) + self.MyMinLevel)

        return float(math.ceil(a * 2) / 2)

    def ConvertToPercent(self, db):
        a = float(((db - self.MyMinLevel) * 100) / (self.MyMaxLevel - self.MyMinLevel))

        return int(a)
     
    def Poll(self):
        if self.IsInitialized():
            self.parent.MyQueue.Enqueue(MessageBundle(self, self.MyControlPoint.BuildQueryMsg()), STATUS_QUERY)
        
    def Reinitialize(self):
        if self.IsPollable:
            self.MyControlPoint.Reinitialize()
        
        self.UpdateCrosspointChange(None, True)
        self.UpdateLevelChange(None, True)
    
    def Refresh(self):
        self.UpdateCrosspointChange(self.MyControlPoint.State, True)
        self.UpdateLevelChange(self.MyControlPoint.Attenuation, True)
    
    def GetInitialized(self):
        self.parent.MyQueue.Enqueue(MessageBundle(self, self.MyControlPoint.BuildQueryMsg()), STATUS_QUERY)
    
    def IsInitialized(self):
        return self.MyControlPoint.IsInitialized()
    
    def ProcessSubscription(self, msg):
        if isinstance(msg, MTResponseMsg):
            if msg == self.MyControlPoint:
                if msg.Message.State != self.MyControlPoint.State:
                    self.MyControlPoint.State = msg.Message.State
                    self.UpdateCrosspointChange(msg.Message.State, False)

                if msg.Message.Attenuation != self.MyControlPoint.Attenuation:
                    self.MyControlPoint.Attenuation = msg.Message.Attenuation
                    self.UpdateLevelChange(msg.Message.Attenuation, False)
                
                if msg.Message.CrosspointType != self.MyControlPoint.CrosspointType:
                    self.MyControlPoint.CrosspointType = msg.Message.CrosspointType
                
    def ProcessResponse(self, response):
        pass
    
    def IsGoodResponse(self, response):
        if isinstance(response, MTResponseMsg):
            return response == self.MyControlPoint
    
    def UpdateCrosspointChange(self, value, force):
        try:
            if force or value != self.MyCurrentState:
                self.MyCurrentState = int(value) if value is not None else 0
                self.__OnCrosspointChange(self, self.MyCurrentState)

        except Exception as e:
            print('Exception.MatrixComponent.UpdateMuteChange: ' + str(e))
    
    def UpdateLevelChange(self, value, force):
        try:
            if force or value != self.MyCurrentLevel:
                self.MyCurrentLevel = float(value) if value is not None else self.MyMinLevel
                
                self.__OnAttenuationDBChange(self, self.MyCurrentLevel)
                self.__OnAttenuationPercentChange(self, self.ConvertToPercent(self.MyCurrentLevel))
                
                if self.MyCurrentLevel > self.MyMaxLevel:
                    self.SetLevelDB(self.MyMaxLevel)
                
                if self.MyCurrentLevel < self.MyMinLevel:
                    self.SetLevelDB(self.MyMinLevel)
        
        except Exception as e:
            print('Exception.MatrixComponent.UpdateLevelChange: ' + str(e))
            
################################################################################
## CLASS: MacroComponent
################################################################################

class MacroComponent(Component):
    def __init__(self, parent):
        try:
            super().__init__(parent)
            self.MyControlPoint = MCCFMsg()
        
        except Exception as e:
            print('Exception.MacroComponent.__init__: ' + str(e))

    def RunMacro(self, name):
        if self.IsInitialized():
            req = MCCFMsg(Msg=self.MyControlPoint)
            req.MacroName = name
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
            
    def Poll(self):
        pass
        
    def Reinitialize(self):
        pass
    
    def Refresh(self):
        pass
    
    def GetInitialized(self):
        pass
    
    def IsInitialized(self):
        return self.MyControlPoint.IsInitialized()

        return True
    
    def ProcessSubscription(self, msg):
        pass
        
    def ProcessResponse(self, response):
        pass
    
    def IsGoodResponse(self, response):
        if isinstance(response, EPResponseMsg):
            return response == self.MyControlPoint
        
        return False

################################################################################
## CLASS: PresetComponent
################################################################################

class PresetComponent(RoomGenericComponent):
    def __init__(self, parent, RoomNumber, RoomPartitionName):
        try:
            super().__init__(parent, SignalType='Serial', RoomNumber=RoomNumber, Option='Select', P1=RoomPartitionName, P2=None)

        except Exception as e:
            print('Exception.PresetComponent.__init__: ' + str(e))
    
    def RecallPreset(self, name):
        self.SetSerial(name)

################################################################################
## CLASS: AnalogDialerComponent
################################################################################

class CallerID():
    def __init__(self, Msg=None):
        try:
            self.Time = ''
            self.Date = ''
            self.PhoneNumber = ''
            self.Name = ''
                   
            if isinstance(Msg, str):
                tempList = Msg.split()
                
                if isinstance(tempList, list) and len(tempList) > 2:
                    self.Date = tempList[0]
                    tempList.pop(0)
                    
                    self.Time = tempList[0]
                    tempList.pop(0)
                    
                    self.PhoneNumber = tempList[0]
                    tempList.pop(0)
                    
                    for item in tempList:
                        self.Name += '{0} '.format(item)
                    
                    self.Name.strip()     
        
        except Exception as e:
            print('Exception.CallerID.__init__: ' + str(e))
            
class AnalogDialerComponent(Component):
    def __init__(self, parent, EndPointName):
        try:
            super().__init__(parent)
            self.MyKeypadText = ''
            self.MyHookStatus = 0
            self.MyIncomingCallNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='INCOMING_CALL')
            self.MyCallerIDNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='CALLER_ID')
            self.MyHookNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='HOOK')
            self.MyRingNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='RING')
            self.MyCallDurationNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='CALL_DURATION')
            self.MyAutoAnswerSettings = EPMsg(EndPointName=EndPointName, BlockNumber='SETTINGS', ParameterName='AUTO_ANSWER_RINGS')
            self.MyLocalNumberSettings = EPMsg(EndPointName=EndPointName, BlockNumber='SETTINGS', ParameterName='LOCAL_NUMBER')
            self.MyLastNumberDialedSettings = EPMsg(EndPointName=EndPointName, BlockNumber='SETTINGS', ParameterName='LAST_DIALED_DIGITS')
            self.MyDTMFDigitPressed = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_DIGIT_PRESSED')
            self.MyDTMFDigitReleased = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_DIGIT_RELEASED')
            self.MyHookFlash = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_HOOK_FLASH')
            self.MyHook = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_HOOK')
            self.MyCallNumber = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_CALL')
            self.MyRedial = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_REDIAL')
            self.MyDigitsDialedInquire = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE', ParameterName='DIGITS_DIALED_SINCE_OFF_HOOK')
            self.MyCallerIDInquire = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE', ParameterName='CALLER_ID')
            self.MyDirectionInquire = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE', ParameterName='DIRECTION')
            self.MyOffHookDurationInquire = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE', ParameterName='OFF_HOOK_DURATION')
            self.MyHookInquire = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE', ParameterName='HOOK')
            
            self.MyControlPoints = [self.MyIncomingCallNotification, self.MyCallerIDNotification, self.MyHookNotification, self.MyRingNotification, 
                                    self.MyCallDurationNotification, self.MyAutoAnswerSettings, self.MyLocalNumberSettings, self.MyLastNumberDialedSettings, 
                                    self.MyDTMFDigitPressed, self.MyDTMFDigitReleased, self.MyHookFlash, self.MyHook, self.MyCallNumber, self.MyRedial,
                                    self.MyDigitsDialedInquire, self.MyCallerIDInquire, self.MyDirectionInquire, self.MyOffHookDurationInquire, self.MyHookInquire]
            
            self.__OnAnalogKeypadTextChange = self.__unassigned
            self.__OnAnalogAutoAnswerChange = self.__unassigned
            self.__OnAnalogCallDurationChange = self.__unassigned
            self.__OnAnalogCallerIDNotification = self.__unassigned
            self.__OnAnalogHookChange = self.__unassigned
            self.__OnAnalogIncomingCallChange = self.__unassigned
            self.__OnAnalogIncomingRingChange = self.__unassigned
            self.__OnAnalogLastNumberDialedChange = self.__unassigned
            self.__OnAnalogLocalNumberChange = self.__unassigned
        
        except Exception as e:
            print('Exception.AnalogDialerComponent.__init__: ' + str(e))
    
    def __unassigned(self, *args):
        pass
        
    @property
    def OnAnalogKeypadTextChange(self):
        return self.__OnAnalogKeypadTextChange
    
    @OnAnalogKeypadTextChange.setter
    def OnAnalogKeypadTextChange(self, handler):
        if callable(handler):
            self.__OnAnalogKeypadTextChange = handler
        else:
            raise ValueError('handler must be a function')
    
    @property
    def OnAnalogAutoAnswerChange(self):
        return self.__OnAnalogAutoAnswerChange
    
    @OnAnalogAutoAnswerChange.setter
    def OnAnalogAutoAnswerChange(self, handler):
        if callable(handler):
            self.__OnAnalogAutoAnswerChange = handler
        else:
            raise ValueError('handler must be a function')
    
    @property
    def OnAnalogCallDurationChange(self):
        return self.__OnAnalogCallDurationChange
    
    @OnAnalogCallDurationChange.setter
    def OnAnalogCallDurationChange(self, handler):
        if callable(handler):
            self.__OnAnalogCallDurationChange = handler
        else:
            raise ValueError('handler must be a function')
    
    @property
    def OnAnalogCallerIDNotification(self):
        return self.__OnAnalogCallerIDNotification
    
    @OnAnalogCallerIDNotification.setter
    def OnAnalogCallerIDNotification(self, handler):
        if callable(handler):
            self.__OnAnalogCallerIDNotification = handler
        else:
            raise ValueError('handler must be a function')
    
    @property
    def OnAnalogHookChange(self):
        return self.__OnAnalogHookChange
    
    @OnAnalogHookChange.setter
    def OnAnalogHookChange(self, handler):
        if callable(handler):
            self.__OnAnalogHookChange = handler
        else:
            raise ValueError('handler must be a function')
    
    @property
    def OnAnalogIncomingCallChange(self):
        return self.__OnAnalogIncomingCallChange
    
    @OnAnalogIncomingCallChange.setter
    def OnAnalogIncomingCallChange(self, handler):
        if callable(handler):
            self.__OnAnalogIncomingCallChange = handler
        else:
            raise ValueError('handler must be a function')
    
    @property
    def OnAnalogIncomingRingChange(self):
        return self.__OnAnalogIncomingRingChange
    
    @OnAnalogIncomingRingChange.setter
    def OnAnalogIncomingRingChange(self, handler):
        if callable(handler):
            self.__OnAnalogIncomingRingChange = handler
        else:
            raise ValueError('handler must be a function')

    @property
    def OnAnalogLastNumberDialedChange(self):
        return self.__OnAnalogLastNumberDialedChange
    
    @OnAnalogLastNumberDialedChange.setter
    def OnAnalogLastNumberDialedChange(self, handler):
        if callable(handler):
            self.__OnAnalogLastNumberDialedChange = handler
        else:
            raise ValueError('handler must be a function')
            
    @property
    def OnAnalogLocalNumberChange(self):
        return self.__OnAnalogLocalNumberChange
    
    @OnAnalogLocalNumberChange.setter
    def OnAnalogLocalNumberChange(self, handler):
        if callable(handler):
            self.__OnAnalogLocalNumberChange = handler
        else:
            raise ValueError('handler must be a function')
            
    def OnHook(self):
        if self.IsInitialized():
            req = EPMsg(Msg=self.MyHook)
            req.Value = 0
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def OffHook(self):
        if self.IsInitialized():
            req = EPMsg(Msg=self.MyHook)
            req.Value = 1
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)        
    
    def HookFlash(self):
        if self.IsInitialized():
            req = EPMsg(Msg=self.MyHookFlash)
            req.Value = 0
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def GetHookStatus(self):
        return int(self.MyHookStatus) if self.MyHookStatus is not None else 0

    def Redial(self):
        if self.IsInitialized():
            req = EPMsg(Msg=self.MyRedial)
            req.Value = 0
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
        
    def DialNumber(self, number):
        if self.IsInitialized():
            if re.search(r'^[\d*#]+$', number) is not None:
                req = EPMsg(Msg=self.MyCallNumber)
                req.Value = number
                
                self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def Dial(self):
        if self.IsInitialized() and self.MyHookStatus == False:
            self.DialNumber(self.MyKeypadText)
            self.MyKeypadText = ''
    
    def ClearKeypadText(self):
        self.MyKeypadText = ''
        self.UpdateKeypadTextChange()
    
    def BackSpaceKeypadText(self):
        if isinstance(self.MyKeypadText, str) and len(self.MyKeypadText) > 0:
            self.MyKeypadText = self.MyKeypadText[:-1]
            self.UpdateKeypadTextChange()
    
    def DigitPressed(self, digit):
        if self.IsInitialized():
            if re.search(r'[\d*#]', digit) is not None and len(digit) == 1:
                self.MyKeypadText += str(digit)
    
                if self.MyHookStatus:
                    req = EPMsg(Msg=self.MyDTMFDigitPressed)
                    req.Value = digit
                
                    self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
                
                self.UpdateKeypadTextChange()
    
    def DigitReleased(self, digit):
        if self.IsInitialized():
            if re.search(r'[\d*#]', digit) is not None and len(digit) == 1:
                if self.MyHookStatus:
                    req = EPMsg(Msg=self.MyDTMFDigitReleased)
                    req.Value = digit
                
                    self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def GetKeypadText(self):
        return str(self.MyKeypadText) if self.MyKeypadText is not None else ''
    
    def SetAutoAnswerRingCount(self, ringCount):
        if self.IsInitialized():
            if ringCount >= 0 and ringCount <= 3:
                req = EPMsg(Msg=self.MyAutoAnswerSettings)
                req.Value = ringCount
            
                self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def GetAutoAnswerRingCount(self):
        return int(self.MyAutoAnswerSettings.Value) if self.MyAutoAnswerSettings.Value is not None else 0
    
    def GetCallDuration(self):
        return int(self.MyOffHookDurationInquire.Value) if self.MyOffHookDurationInquire.Value is not None else 0
    
    def GetCallerIDInformation(self):
        return CallerID(self.MyCallerIDInquire.Value) if self.MyCallerIDInquire.Value is not None else CallerID()
    
    def GetLastNumberDialed(self):
        return str(self.MyLastNumberDialedSettings.Value) if self.MyLastNumberDialedSettings.Value is not None else ''
    
    def GetLocalNumber(self):
        return str(self.MyLocalNumberSettings.Value) if self.MyLocalNumberSettings.Value is not None else ''
    
    def GetLocalNumber(self):
        return str(self.MyLocalNumberSettings.Value) if self.MyLocalNumberSettings.Value is not None else ''
    
    def Poll(self):
        if self.IsInitialized():
            for item in self.MyControlPoints:
                if item.BlockNumber == 'INQUIRE':
                    self.parent.MyQueue.Enqueue(MessageBundle(self, item.BuildQueryMsg()), STATUS_QUERY)
        
    def Reinitialize(self):
        for item in self.MyControlPoints:
            item.Reinitialize()
            
            if item.BlockNumber == 'SETTINGS' or item.BlockNumber == 'INQUIRE':
                self.UpdateStatusFeedback(item)
                
        self.UpdateKeypadTextChange()
    
    def Refresh(self):
        for item in self.MyControlPoints:
            if item.BlockNumber == 'SETTINGS' or item.BlockNumber == 'INQUIRE':
                self.UpdateStatusFeedback(item)
                
        self.UpdateKeypadTextChange()
    
    def GetInitialized(self):
        for item in self.MyControlPoints:
            if item.BlockNumber == 'SETTINGS' or item.BlockNumber == 'INQUIRE':
                self.parent.MyQueue.Enqueue(MessageBundle(self, item.BuildQueryMsg()), STATUS_QUERY)
    
    def IsInitialized(self):
        for item in self.MyControlPoints:
            if item.BlockNumber == 'SETTINGS' or item.BlockNumber == 'INQUIRE':
                if not item.IsInitialized():
                    return False
        
        return True
    
    def ProcessSubscription(self, msg):
        if isinstance(msg, EPResponseMsg):
            for item in self.MyControlPoints:
                if item.BlockNumber == 'SETTINGS' or item.BlockNumber == 'INQUIRE' or item.BlockNumber == 'NOTIFICATION':
                    if msg == item:
                        if msg.Message.Value != item.Value:
                            item.Value = msg.Message.Value
                            self.UpdateStatusFeedback(item)
                        break
                
    def ProcessResponse(self, response):
        pass
    
    def IsGoodResponse(self, response):
        if isinstance(response, EPResponseMsg):
            for item in self.MyControlPoints:
                if response == item:
                    return True

        return False
    
    def UpdateKeypadTextChange(self):
        try:
            self.__OnAnalogKeypadTextChange(self, self.MyKeypadText)

        except Exception as e:
            print('Exception.AnalogDialerComponent.UpdateKeypadTextChange: ' + str(e))
    
    def UpdateStatusFeedback(self, item):
        try:
            if item == self.MyIncomingCallNotification:
                self.__OnAnalogIncomingCallChange(self, item.Value if item.Value is not None else '')
            
            elif item == self.MyRingNotification:
                self.__OnAnalogIncomingRingChange(self, int(item.Value) if item.Value is not None else 0)
            
            elif item == self.MyCallerIDNotification:
                self.__OnAnalogCallerIDNotification(self, CallerID(item.Value))
            
            elif item == self.MyHookNotification:
                self.MyHookStatus = int(item.Value) if item.Value is not None else 0
                self.__OnAnalogHookChange(self, self.MyHookStatus)
                
                self.MyIncomingCallNotification.Value = ''
                self.MyCallerIDInquire.Value = ''
                self.UpdateStatusFeedback(self.MyIncomingCallNotification)
                
                self.ClearKeypadText()
                
                self.Poll()
                
            elif item == self.MyCallDurationNotification:
                self.__OnAnalogCallDurationChange(self, int(item.Value) if item.Value is not None else 0)    
            
            elif item == self.MyAutoAnswerSettings:
                self.__OnAnalogAutoAnswerChange(self, int(item.Value) if item.Value is not None else 0)            
    
            elif item == self.MyLocalNumberSettings:
                self.__OnAnalogLocalNumberChange(self, str(item.Value) if item.Value is not None else '') 
    
            elif item == self.MyLastNumberDialedSettings:
                self.__OnAnalogLastNumberDialedChange(self, str(item.Value) if item.Value is not None else '')
            
            elif item == self.MyCallerIDInquire:
                self.__OnAnalogCallerIDNotification(self, CallerID(item.Value))
            
            elif item == self.MyOffHookDurationInquire:
                self.__OnAnalogCallDurationChange(self, int(item.Value) if item.Value is not None else 0)
            
            elif item == self.MyHookInquire:
                self.MyHookStatus = int(item.Value) if item.Value is not None else 0
                self.__OnAnalogHookChange(self, self.MyHookStatus)
                    
        except Exception as e:
            print('Exception.AnalogDialerComponent.UpdateStatusFeedback: ' + str(e))

################################################################################
## CLASS: VoIPDialerComponent
################################################################################

class VoipLineState():
    def __init__(self, lineID=None):
        try:
            self.LineLEDState = 0
            self.LineStatus = 0
            self.LineStatusText = ''
            self.LineInfo = ''
            self.LineID = lineID if lineID is not None else 0
            
        except Exception as e:
            print('Exception.VoipLineState.__init__: ' + str(e))  
    
    def __str__(self):
        msgList = []
        msgList.append('LineID: {0} | '.format(self.LineID))
        msgList.append('LineLEDState: {0} | '.format(self.LineLEDState))
        msgList.append('LineStatus: {0} | '.format(self.LineStatus))
        msgList.append('LineStatusText: {0} | '.format(self.LineStatusText)) 
        msgList.append('LineInfo: {0}'.format(self.LineInfo)) 

        return ''.join(msgList)

class VoIPStatus():
    def __init__(self):
        try:
            self.HoldStatus = 0
            self.MuteStatus = 0
            self.RingBackStatus = 0
            self.RingingStatus = 0
            self.CallForwardState = 0
            self.CallForwardStatus = 0
            self.CallForwardStatusText = ''
            self.DNDStatus = 0
            self.DNDStatusText = ''
            self.BusyStatus = 0
            self.WarningErrorStatus = 0
            self.CallWaiting = 0
        
            self.LineStatus = []
            
            for i in range(5):
                self.LineStatus.append(VoipLineState(lineID=i))
        
        except Exception as e:
            print('Exception.VoIPStatus.__init__: ' + str(e)) 
    
    def UpdateLineStatus(self, line, state):
        try:
            tempState = ''
            tempData = ''
            tempType = StateChangeTypes['UNKNOWN']
            dataMatch = re.search(r'(?P<state>\w+)(:(?P<data>.+))?', state)
            
            if dataMatch is not None:
                tempState = dataMatch.group('state') if dataMatch.group('state') is not None else ''
                tempData = dataMatch.group('data') if dataMatch.group('data') is not None else ''
                tempType = StateChangeTypes.get(tempState, StateChangeTypes['UNKNOWN'])
            else:
                tempType = StateChangeTypes.get(state, StateChangeTypes['UNKNOWN'])
            
            if line > 0 and line <= 5:
                self.LineStatus[line - 1].LineStatus = int(tempType)
                self.LineStatus[line - 1].LineStatusText = tempState
                self.LineStatus[line - 1].LineInfo = tempData
        
        except Exception as e:
            print('Exception.VoIPStatus.UpdateLineStatus: ' + str(e))
    
    def UpdateStatus(self, msg):
        try:
            if isinstance(msg, EPMsg):
                if msg.BlockNumber == 'INQUIRE_RESULT' and msg.ParameterName == 'ACTIVE_PARTIES':
                    itemMatch = re.search(r'(?P<line1>.+);(?P<line2>.+);(?P<line3>.+);(?P<line4>.+);(?P<line5>.+);', msg.Value)
                    
                    if itemMatch is not None:
                        self.UpdateLineStatus(1, itemMatch.group('line1'))
                        self.UpdateLineStatus(2, itemMatch.group('line2'))
                        self.UpdateLineStatus(3, itemMatch.group('line3'))
                        self.UpdateLineStatus(4, itemMatch.group('line4'))
                        self.UpdateLineStatus(5, itemMatch.group('line5'))
                        
                        return True
                
                elif msg.BlockNumber == 'INQUIRE_RESULT' and msg.ParameterName == 'DND_STATUS':
                    state = DNDStatus.get(msg.Value, DNDStatus['DND_NOT_SET'])
                    self.DNDStatus = int(state)
                    self.DNDStatusText = [key for key, value in DNDStatus.items() if value == state][0]
                    
                    return True
                
                elif msg.BlockNumber == 'INQUIRE_RESULT' and msg.ParameterName == 'CFW_STATUS':
                    itemMatch = re.search(r'(?P<state>(INACTIVE|INPROCESS|ACTIVE))', msg.Value)
                    
                    if itemMatch is not None:
                        state = CallForwardWaitingStates.get(itemMatch.group('state'), CallForwardWaitingStates['INACTIVE'])
                        self.CallForwardState = int(state)
                        
                        if state == CallForwardWaitingStates['INACTIVE']:
                            self.CallForwardStatus = CallForwardWaitingStatus['DISABLED']
                            self.CallForwardStatusText = [key for key, value in CallForwardWaitingStatus.items() if value == CallForwardWaitingStatus['DISABLED']][0]
                            
                        elif state == CallForwardWaitingStates['INPROCESS']:
                            itemMatch = re.search(r'(?P<state>.+);(?P<status>.+);', msg.Value)
                            
                            if itemMatch is not None:
                                status = itemMatch.group('status')
                                status = status.replace('CFW-', '')
                                s = CallForwardWaitingStatus.get(status, CallForwardWaitingStatus['DISABLED'])
                                self.CallForwardStatus = int(s)
                                self.CallForwardStatusText = [key for key, value in CallForwardWaitingStatus.items() if value == s][0]
                                
                        elif state == CallForwardWaitingStates['ACTIVE']:
                            itemMatch = re.search(r'(?P<state>.+);(?P<status>.+);(?P<info>.+)', msg.Value)
                            
                            if itemMatch is not None:
                                status = itemMatch.group('status')
                                status = status.replace('CFW-', '')
                                s = CallForwardWaitingStatus.get(status, CallForwardWaitingStatus['DISABLED'])
                                self.CallForwardStatus = int(s)
                                self.CallForwardStatusText = [key for key, value in CallForwardWaitingStatus.items() if value == s][0]
                                
                        return True
                            
                elif msg.BlockNumber == 'NOTIFICATION' and msg.ParameterName == 'STATE_CHANGE':
                    itemMatch = re.search(r'PL (?P<line>.+);(?P<state>.+)', msg.Value)
                    
                    if itemMatch is not None:
                        l = int(itemMatch.group('line'))
                        s = itemMatch.group('state')
                        
                        self.UpdateLineStatus(l, s)
                        
                        return True
                
                elif msg.BlockNumber == 'NOTIFICATION' and msg.ParameterName == 'INDICATION':
                    itemMatch = re.search(r'PL (?P<party>.+);(?P<type>.+):(?P<state>.+)', msg.Value)
                    
                    if itemMatch is not None:
                        t = NotificationIndicationTypes.get(itemMatch.group('type'), NotificationIndicationTypes['WARNING_ERR'])
                        s = NotificationIndicationStatusTypes.get(itemMatch.group('state'), NotificationIndicationStatusTypes['OFF'])
                        
                        if t == NotificationIndicationTypes['BUSY']:
                            self.BusyStatus = int(s)
                            
                        elif t == NotificationIndicationTypes['CALL_WAITING_CALLEE']:
                            self.CallWaiting = int(s)
                            
                        elif t == NotificationIndicationTypes['HOLD']:
                            self.HoldStatus = int(s)
                            
                        elif t == NotificationIndicationTypes['MUTE']:
                            self.MuteStatus = int(s)
                            
                        elif t == NotificationIndicationTypes['PARTY_LINE']:
                            index = int(itemMatch.group('party'))
                            
                            if index > 0 and index <= 5:
                                self.LineStatus[index - 1].LineLEDState = int(s)
                        
                        elif t == NotificationIndicationTypes['RINGBACK']:
                            self.RingBackStatus = int(s)
                        
                        elif t == NotificationIndicationTypes['RINGING']:
                            self.RingingStatus = int(s)
                        
                        elif t == NotificationIndicationTypes['WARNING_ERR']:
                            self.WarningErrorStatus = int(s)
                        
                        return True
                        
            return False                
            
        except Exception as e:
            print('Exception.VoIPStatus.UpdateStatus: ' + str(e))
        
    def ClearStatus(self):
        self.HoldStatus = 0
        self.MuteStatus = 0
        self.RingBackStatus = 0
        self.RingingStatus = 0
        self.CallForwardState = 0
        self.CallForwardStatus = 0
        self.CallForwardStatusText = ''
        self.DNDStatus = 0
        self.DNDStatusText = ''
        self.BusyStatus = 0
        self.WarningErrorStatus = 0
        self.CallWaiting = 0
        
        self.LineStatus = []
            
        for i in range(5):
            self.LineStatus.append(VoipLineState(lineID=i))

class VoIPError():
    def __init__(self):
        try:
            self.ErrorText1 = ''
            self.ErrorText2 = ''
            self.ErrorText3 = ''
            self.ErrorText4 = ''
        
        except Exception as e:
            print('Exception.VoIPError.__init__: ' + str(e))                 
        
    def UpdateStatus(self, msg):
        try:
            if isinstance(msg, EPMsg):
                if msg.BlockNumber == 'NOTIFICATION' and msg.ParameterName == 'ERROR':
                    itemMatch = re.search(r'(?P<error1>.+)?;(?P<error2>.+)?;(?P<error3>.+)?;(?P<error4>.+)?', msg.Value)
                    
                    if itemMatch is not None:
                        self.ErrorText1 = itemMatch.group('error1') if itemMatch.group('error1') is not None else ''
                        self.ErrorText2 = itemMatch.group('error2') if itemMatch.group('error2') is not None else ''
                        self.ErrorText3 = itemMatch.group('error3') if itemMatch.group('error3') is not None else ''
                        self.ErrorText4 = itemMatch.group('error4') if itemMatch.group('error4') is not None else ''
                        
                        return True
                                        
            return False                
            
        except Exception as e:
            print('Exception.VoIPError.UpdateStatus: ' + str(e))
        
    def ClearStatus(self):
        self.ErrorText1 = ''
        self.ErrorText2 = ''
        self.ErrorText3 = ''
        self.ErrorText4 = ''
            
class VoIPDialerComponent(Component):
    def __init__(self, parent, EndPointName):
        try:
            super().__init__(parent)
            self.MyKeypadText = ''
            self.MyStatus = VoIPStatus()
            self.MyErrors = VoIPError()
            self.MyRegStatusInquire = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE', ParameterName='REG_STATUS')
            self.MyActivePartiesInquire = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE', ParameterName='ACTIVE_PARTIES')
            self.MyCallForwardStatusInquire = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE', ParameterName='CFW_STATUS')
            self.MyDNDStatusInquire = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE', ParameterName='DND_STATUS')
            self.MyRegStatusInquireResult = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE_RESULT', ParameterName='REG_STATUS')
            self.MyActivePartiesInquireResult = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE_RESULT', ParameterName='ACTIVE_PARTIES')
            self.MyCallForwardStatusInquireResult = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE_RESULT', ParameterName='CFW_STATUS')
            self.MyDNDStatusInquireResult = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE_RESULT', ParameterName='DND_STATUS')
            self.MyKeyCallCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_CALL')
            self.MyKeyRejectCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_REJECT')
            self.MyKeyHoldCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_HOLD')
            self.MyKeyTransferCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_TRANSFER')
            self.MyKeyBlindTransferCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_BLIND_TRANSFER')
            self.MyKeyConferenceCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_CONFERENCE')
            self.MyKeyDigitPressedCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_DIGIT_PRESSED')
            self.MyKeyDigitReleasedCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_DIGIT_RELEASED')
            self.MyKeyRedialCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_REDIAL')
            self.MyKeySelectPartyCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_PARTY')
            self.MyKeyHookCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_HOOK')
            self.MyKeyMuteCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_MUTE')
            self.MyKeyForwardCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_FORWARD')
            self.MyKeyDNDCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_DO_NOT_DISTURB')
            self.MyStateChangeNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='STATE_CHANGE')
            self.MyIndicationNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='INDICATION')
            self.MyErrorNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='ERROR')
            self.MyRegFailedNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='REG_FAILED')
            self.MyRegSucceedNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='REG_SUCCEED')
            self.MyUnregSucceedNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='UNREG_SUCCEED')
            self.MyDisplayNameSettings = EPMsg(EndPointName=EndPointName, BlockNumber='SETTINGS', ParameterName='DISPLAY_NAME')
    
            self.MyControlPoints = [self.MyRegStatusInquire, self.MyActivePartiesInquire, self.MyCallForwardStatusInquire, 
                                    self.MyDNDStatusInquire, self.MyRegStatusInquireResult, self.MyActivePartiesInquireResult,
                                    self.MyCallForwardStatusInquireResult, self.MyDNDStatusInquireResult, self.MyKeyCallCommand,
                                    self.MyKeyRejectCommand, self.MyKeyHoldCommand, self.MyKeyTransferCommand, self.MyKeyBlindTransferCommand,
                                    self.MyKeyConferenceCommand, self.MyKeyDigitPressedCommand, self.MyKeyDigitReleasedCommand, 
                                    self.MyKeyRedialCommand, self.MyKeySelectPartyCommand, self.MyKeyHookCommand, self.MyKeyMuteCommand,
                                    self.MyKeyForwardCommand, self.MyKeyDNDCommand, self.MyStateChangeNotification, self.MyIndicationNotification,
                                    self.MyErrorNotification, self.MyRegFailedNotification, self.MyRegSucceedNotification, 
                                    self.MyUnregSucceedNotification, self.MyDisplayNameSettings]
            
            self.__OnVoIPKeypadTextChange = self.__unassigned
            self.__OnVoIPStatusChange = self.__unassigned
            self.__OnVoIPErrorChange = self.__unassigned
        
        except Exception as e:
            print('Exception.VoIPDialerComponent.__init__: ' + str(e))
    
    def __unassigned(self, *args):
        pass
        
    @property
    def OnVoIPKeypadTextChange(self):
        return self.__OnVoIPKeypadTextChange
    
    @OnVoIPKeypadTextChange.setter
    def OnVoIPKeypadTextChange(self, handler):
        if callable(handler):
            self.__OnVoIPKeypadTextChange = handler
        else:
            raise ValueError('handler must be a function')
    
    @property
    def OnVoIPStatusChange(self):
        return self.__OnVoIPStatusChange
    
    @OnVoIPStatusChange.setter
    def OnVoIPStatusChange(self, handler):
        if callable(handler):
            self.__OnVoIPStatusChange = handler
        else:
            raise ValueError('handler must be a function')
    
    @property
    def OnVoIPErrorChange(self):
        return self.__OnVoIPErrorChange
    
    @OnVoIPErrorChange.setter
    def OnVoIPErrorChange(self, handler):
        if callable(handler):
            self.__OnVoIPErrorChange = handler
        else:
            raise ValueError('handler must be a function')
 
    def SelectLine(self, line):
        if self.IsInitialized():
            if line > 0 and line <= 5:
                req = EPMsg(Msg=self.MyKeySelectPartyCommand)
                req.Value = line
                
                self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def GetActiveLine(self):
        for line in self.MyStatus.LineStatus:
            if line.LineLEDState != NotificationIndicationStatusTypes['OFF']:
                return line.LineID + 1
        
        return 0
    
    def DialNumber(self, number):
        if self.IsInitialized():
            req = EPMsg(Msg=self.MyKeyCallCommand)
            req.Value = number
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def Dial(self):
        if self.IsInitialized():
            self.DialNumber(self.MyKeypadText)
            self.MyKeypadText = ''
 
    def ClearKeypadText(self):
        self.MyKeypadText = ''
        self.UpdateKeypadTextChange()
    
    def BackSpaceKeypadText(self):
        if isinstance(self.MyKeypadText, str) and len(self.MyKeypadText) > 0:
            self.MyKeypadText = self.MyKeypadText[:-1]
            self.UpdateKeypadTextChange()
    
    def DigitPressed(self, digit):
        if self.IsInitialized():
            if re.search(r'[\d*#]', digit) is not None and len(digit) == 1:
                self.MyKeypadText += str(digit)
    
                if self.GetActiveLine() != 0:
                    req = EPMsg(Msg=self.MyKeyDigitPressedCommand)
                    req.Value = digit
                
                    self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
                
                self.UpdateKeypadTextChange()
    
    def DigitReleased(self, digit):
        if self.IsInitialized():
            if re.search(r'[\d*#]', digit) is not None and len(digit) == 1:
                if self.GetActiveLine() != 0:
                    req = EPMsg(Msg=self.MyKeyDigitReleasedCommand)
                    req.Value = digit
                
                    self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
 
    def Reject(self, line):
        if self.IsInitialized():
            req = EPMsg(Msg=self.MyKeyRejectCommand)
            req.Value = line
        
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
 
    def Hold(self):
        if self.IsInitialized():
            req = EPMsg(Msg=self.MyKeyHoldCommand)
            req.Value = ''
        
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
 
    def Transfer(self):
        if self.IsInitialized():
            req = EPMsg(Msg=self.MyKeyTransferCommand)
            req.Value = ''
        
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def BlindTransfer(self):
        if self.IsInitialized():
            req = EPMsg(Msg=self.MyKeyBlindTransferCommand)
            req.Value = ''
        
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def Conference(self, line):
        if self.IsInitialized():
            req = EPMsg(Msg=self.MyKeyConferenceCommand)
            req.Value = line
        
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def Redial(self):
        if self.IsInitialized():
            req = EPMsg(Msg=self.MyKeyRedialCommand)
            req.Value = ''
        
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def MuteOn(self):
        if self.IsInitialized():
            req = EPMsg(Msg=self.MyKeyMuteCommand)
            req.Value = 1
        
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
            
    def MuteOff(self):
        if self.IsInitialized():
            req = EPMsg(Msg=self.MyKeyMuteCommand)
            req.Value = 0
        
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)       
            
    def MuteToggle(self):
        if self.IsInitialized():
            req = EPMsg(Msg=self.MyKeyMuteCommand)
            req.Value = 2
        
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND) 
    
    def OnHook(self):
        if self.IsInitialized():
            req = EPMsg(Msg=self.MyKeyHookCommand)
            req.Value = 0
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def OffHook(self):
        if self.IsInitialized():
            req = EPMsg(Msg=self.MyKeyHookCommand)
            req.Value = 1
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)        
    
    def HookToggle(self):
        if self.IsInitialized():
            req = EPMsg(Msg=self.MyKeyHookCommand)
            req.Value = 2
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND) 
  
    def CallForwardDisable(self):
        if self.IsInitialized():
            req = EPMsg(Msg=self.MyKeyForwardCommand)
            req.Value = 0
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND) 

    def CallForwardEnable(self):
        if self.IsInitialized():
            req = EPMsg(Msg=self.MyKeyForwardCommand)
            req.Value = 1
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
            self.parent.MyQueue.Enqueue(MessageBundle(self, self.MyCallForwardStatusInquire.BuildSetMsg()), STATUS_COMMAND)

    def CallForwardBusy(self):
        if self.IsInitialized():
            req = EPMsg(Msg=self.MyKeyForwardCommand)
            req.Value = 2
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
            self.parent.MyQueue.Enqueue(MessageBundle(self, self.MyCallForwardStatusInquire.BuildSetMsg()), STATUS_COMMAND)
            
    def CallForwardNoAnswer(self):
        if self.IsInitialized():
            req = EPMsg(Msg=self.MyKeyForwardCommand)
            req.Value = 3
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
            self.parent.MyQueue.Enqueue(MessageBundle(self, self.MyCallForwardStatusInquire.BuildSetMsg()), STATUS_COMMAND)
            
    def DNDDisable(self):
        if self.IsInitialized():
            req = EPMsg(Msg=self.MyKeyDNDCommand)
            req.Value = 0
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
            self.parent.MyQueue.Enqueue(MessageBundle(self, self.MyDNDStatusInquire.BuildSetMsg()), STATUS_COMMAND)
            
    def DNDCallMute(self):
        if self.IsInitialized():
            req = EPMsg(Msg=self.MyKeyDNDCommand)
            req.Value = 1
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
            self.parent.MyQueue.Enqueue(MessageBundle(self, self.MyDNDStatusInquire.BuildSetMsg()), STATUS_COMMAND)
            
    def DNDCallReject(self):
        if self.IsInitialized():
            req = EPMsg(Msg=self.MyKeyDNDCommand)
            req.Value = 2
            
            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
            self.parent.MyQueue.Enqueue(MessageBundle(self, self.MyDNDStatusInquire.BuildSetMsg()), STATUS_COMMAND)
            
    def ClearErrors(self):
        if self.IsInitialized():
            self.MyErrors.ClearStatus()
            self.OnVoIPErrorChange(self.MyErrors)

    def GetStatus(self):
        return self.MyStatus if self.MyStatus is not None else VoIPStatus
        
    def GetErrors(self):
        return self.MyErrors if self.MyErrors is not None else VoIPError
    
    def Poll(self):
        if self.IsInitialized():
            for item in self.MyControlPoints:
                if item.BlockNumber == 'INQUIRE':
                    self.parent.MyQueue.Enqueue(MessageBundle(self, item.BuildQueryMsg()), STATUS_QUERY)
        
    def Reinitialize(self):
        for item in self.MyControlPoints:
            item.Reinitialize()
            
            if item.BlockNumber == 'SETTINGS' or item.BlockNumber == 'INQUIRE':
                self.UpdateStatusFeedback(item)
        
        self.MyKeypadText = ''
        self.MyErrors = VoIPError()
        self.MyStatus = VoIPStatus()        
        
        self.UpdateKeypadTextChange()
        self.UpdateStatusFeedback()
    
    def Refresh(self):
        for item in self.MyControlPoints:
            if item.BlockNumber == 'SETTINGS' or item.BlockNumber == 'INQUIRE':
                self.UpdateStatusFeedback(item)
                
        self.UpdateKeypadTextChange()
    
    def GetInitialized(self):
        for item in self.MyControlPoints:
            if item.BlockNumber == 'SETTINGS' or item.BlockNumber == 'INQUIRE':
                self.parent.MyQueue.Enqueue(MessageBundle(self, item.BuildQueryMsg()), STATUS_QUERY)
    
    def IsInitialized(self):
        for item in self.MyControlPoints:
            if item.BlockNumber == 'SETTINGS' or item.BlockNumber == 'INQUIRE_RESULT':
                if not item.IsInitialized():
                    return False
        
        return True
    
    def ProcessSubscription(self, msg):
        if isinstance(msg, EPResponseMsg):
            if msg.Message.ParameterName == 'MESSAGE':
                self.parent.MyQueue.Enqueue(MessageBundle(self, self.MyCallForwardStatusInquire.BuildQueryMsg()), STATUS_COMMAND)
                return            
            
            for item in self.MyControlPoints:
                if item.BlockNumber == 'SETTINGS' or item.BlockNumber == 'INQUIRE_RESULT' or item.BlockNumber == 'NOTIFICATION':
                    if msg == item:
                        if msg.Message.Value != item.Value:
                            item.Value = msg.Message.Value
                            self.UpdateStatusFeedback(item)
                        break
                
    def ProcessResponse(self, response):
        pass
    
    def IsGoodResponse(self, response):
        if isinstance(response, EPResponseMsg):
            for item in self.MyControlPoints:
                if response == item:
                    return True

        return False
    
    def UpdateKeypadTextChange(self):
        try:
            self.__OnVoIPKeypadTextChange(self, self.MyKeypadText)

        except Exception as e:
            print('Exception.VoIPDialerComponent.UpdateKeypadTextChange: ' + str(e))
    
    def UpdateStatusFeedback(self, item=None):
        try:
            if isinstance(item, EPMsg):
                success = self.MyStatus.UpdateStatus(item)
    
                if success:
                    self.__OnVoIPStatusChange(self, self.MyStatus)
                
                success = self.MyErrors.UpdateStatus(item)
    
                if success:
                    self.__OnVoIPErrorChange(self, self.MyErrors)
            
            else:
                self.__OnVoIPStatusChange(self, self.MyStatus)
                self.__OnVoIPErrorChange(self, self.MyErrors)
                    
        except Exception as e:
            print('Exception.VoIPDialerComponent.UpdateStatusFeedback: ' + str(e))

################################################################################
## CLASS: SkypeComponent
################################################################################

class S4BActiveSessionInfo():
    def __init__(self):
        try:
            self.MeetingID = ''
            self.Url = ''
            self.Status = S4BSessionStatus['UNKNOWN']
            self.IsConference = False 
            self.StatusText = 'UNKNOWN'

        except Exception as e:
            print('Exception.S4BActiveSessionInfo.__init__: ' + str(e))
    
    def SetStatus(self, key):
        try:
            self.Status = S4BSessionStatus.get(key, S4BSessionStatus['UNKNOWN'])
            self.StatusText = next((key for key, value in S4BSessionStatus.items() if value == self.Status), 'UNKNOWN')

        except Exception as e:
            print('Exception.S4BActiveSessionInfo.SetStatus: ' + str(e))

class S4BContactListResponse():
    def __init__(self):
        try:
            self.List = []
            self.Type = S4BListType['UNKNOWN']
            self.Function = S4BListFunction['UNKNOWN']
            self.HelpText = ''
        
        except Exception as e:
            print('Exception.S4BContactListResponse.__init__: ' + str(e))

    def GetItem(self, index):
        try:
            if isinstance(self.List, list):
                if index < len(self.List):
                    return self.List[index]
                return ''
            return ''
        
        except Exception as e:
            print('Exception.S4BContactListResponse.GetItem: ' + str(e))
        
    def Count(self):
        try:
            if isinstance(self.List, list):
                return len(self.List)
            return 0
        
        except Exception as e:
            print('Exception.S4BContactListResponse.Count: ' + str(e))

class S4BGroupMember():
    def __init__(self):
        try:
            self.Name = ''
            self.DialNumber = ''
            self.OtherNumbers = []
            self.Tag = None 
        
        except Exception as e:
            print('Exception.S4BGroupMember.__init__: ' + str(e))

class S4BGroup():
    def __init__(self):
        try:
            self.ID = ''
            self.Name = ''
            self.Members = [] 
        
        except Exception as e:
            print('Exception.S4BGroup.__init__: ' + str(e))

class S4BRegistrationStatus():
    def __init__(self):
        try:
            self.State = S4BRegState['UNKNOWN']
            self.Status = S4BRegState['UNKNOWN']
            self.StatusText = 'UNKNOWN' 
        
        except Exception as e:
            print('Exception.S4BRegistrationStatus.__init__: ' + str(e))

    def SetState(self, key):
        try:
            self.State = S4BRegState.get(key, S4BRegState['UNKNOWN'])
            self.Status = self.State
            self.StatusText = next((key for key, value in S4BRegState.items() if value == self.State), 'UNKNOWN')

        except Exception as e:
            print('Exception.S4BRegistrationStatus.SetState: ' + str(e))

class S4BSelectedItem():
    def __init__(self):
        try:
            self.Type = S4BListType['UNKNOWN']
            self.DisplayName = ''
            self.Group = ''
            self.Tag = None 
        
        except Exception as e:
            print('Exception.S4BSelectedItem.__init__: ' + str(e))

class S4BSession():
    def __init__(self):
        try:
            self.Conference = False
            self.ID = ''
            self.Status = S4BSessionStatus['UNKNOWN']
            self.ContactUrl = ''
            self.DisplayName = ''
            self.StatusValue = S4BSessionStatus['UNKNOWN']
            self.StatusText = 'UNKNOWN' 
        
        except Exception as e:
            print('Exception.S4BSession.__init__: ' + str(e))

    def SetStatus(self, key):
        try:
            self.Status = S4BSessionStatus.get(key, S4BSessionStatus['UNKNOWN'])
            self.StatusValue = self.Status
            self.StatusText = next((key for key, value in S4BSessionStatus.items() if value == self.Status), 'UNKNOWN')

        except Exception as e:
            print('Exception.S4BSession.SetStatus: ' + str(e))

class S4BSessionListResponse():
    def __init__(self):
        try:
            self.List = []
            self.ActiveItem = 0
        
        except Exception as e:
            print('Exception.S4BSessionListResponse.__init__: ' + str(e))

    def GetItem(self, index, session):
        try:
            if isinstance(self.List, list):
                if index < len(self.List):
                    session = self.List[index]
                    return
                
                session = S4BSession()
        
        except Exception as e:
            print('Exception.S4BSessionListResponse.GetItem: ' + str(e))
        
    def Count(self):
        try:
            if isinstance(self.List, list):
                return len(self.List)
            return 0
        
        except Exception as e:
            print('Exception.S4BSessionListResponse.Count: ' + str(e))

class SkypeComponent(Component):
    def __init__(self, parent, EndPointName):
        try:
            super().__init__(parent)

            self.myCurrentList = S4BContactListResponse()
            self.myCurrentListSelectedItem = S4BSelectedItem()
            self.myRegistrationStatus = None
            self.myKeypadText = ''
            self.myFilterSearchText = ''
            self.myCurrentActiveSessionID = ''
            self.mySessions = None
            self.myGroups = []
            self.myRecentSearches = None
            self.myContactListRefreshTimer = None
            self.mySessionListRefreshTimer = None

            self.myKeyCallCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_CALL')
            self.myKeyDigitPressedCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_DIGIT_PRESSED')
            self.myKeyDigitReleasedCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_DIGIT_RELEASED')
            self.myKeyHookCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_HOOK')
            self.myKeyHoldCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_HOLD')
            self.myKeyResumeCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_RESUME')
            self.myKeyAddParticipantCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_ADD_PARTICIPANT_TO_SESSION')
            self.myKeyRejectCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_REJECT')
            self.myKeyCreateGroupCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_CREATE_GROUP')
            self.myKeyDeleteGroupCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_DELETE_GROUP')
            self.myKeyAddToGroupCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_ADD_TO_GROUP')
            self.myKeyRemoveFromGroupCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_REMOVE_FROM_GROUP')
            self.myKeyRemoveFromAllGroupsCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_REMOVE_FROM_ALL_GROUPS')
            self.myKeySetRingToneCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_SET_RINGTONE')
            self.myKeyPlayRingToneCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_PLAY_RINGTONE')
            self.myKeyAutoAnswerCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_AUTO_ANSWER')
            self.myKeySetActiveSessionCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_SET_ACTIVE_SESSION')
            self.myKeyCreateSessionCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_CREATE_SESSION')
            self.myKeyAcceptSessionCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_ACCEPT')
            self.myKeyLeaveSessionCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_LEAVE_SESSION')
            self.myKeyAddAudioToSessionCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_ADD_AUDIO_TO_SESSION')
            self.myKeyJoinMeetingCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_JOIN_MEETING')
            self.myKeyJoinConferenceCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_JOIN_CONFERENCE')
            self.myKeyMeetNowCommand = EPMsg(EndPointName=EndPointName, BlockNumber='KEY', ParameterName='KEY_MEET_NOW')

            self.myContactSearchInquire = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE', ParameterName='CONTACT_SEARCH')
            self.myGroupsInquire = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE', ParameterName='GROUPS')
            self.myGroupsMembersInquire = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE', ParameterName='GROUP_MEMBERS')
            self.myRegStatusInquire = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE', ParameterName='REG_STATUS')
            self.myActiveRingToneInquire = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE', ParameterName='ACTIVE_RINGTONE')
            self.myAutoAnswerModeInquire = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE', ParameterName='AUTO_ANSWER_MODE')
            self.myCurrentSessionsInquire = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE', ParameterName='CURRENT_SESSIONS')
            self.mySessionCallStateInquire = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE', ParameterName='SESSION_CALL_STATE')
            self.myContactDisplayNameInquire = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE', ParameterName='CONTACT_DISPLAY_NAME')
            self.myContactPhoneNumberListInquire = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE', ParameterName='CONTACT_PHONE_NUMBER_LIST')
            self.myActiveSessionIDInquire = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE', ParameterName='ACTIVE_SESSION_ID')
            self.myMeetUrlInquire = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE', ParameterName='MEET_URL')
            self.mySessionParticipantsInquire = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE', ParameterName='SESSION_PARTICIPANTS')
            self.mySessionInfoInquire = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE', ParameterName='SESSION_INFO')
            self.mySessionRegStatusInquire = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE', ParameterName='REG_STATUS')

            self.myContactSearchInquireResult = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE_RESULT', ParameterName='CONTACT_SEARCH')
            self.myGroupsInquireResult = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE_RESULT', ParameterName='GROUPS')
            self.myGroupsMembersInquireResult = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE_RESULT', ParameterName='GROUP_MEMBERS')
            self.myRegStatusInquireResult = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE_RESULT', ParameterName='REG_STATUS')
            self.myActiveRingToneInquireResult = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE_RESULT', ParameterName='ACTIVE_RINGTONE')
            self.myAutoAnswerModeInquireResult = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE_RESULT', ParameterName='AUTO_ANSWER_MODE')
            self.myCurrentSessionsInquireResult = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE_RESULT', ParameterName='CURRENT_SESSIONS')
            self.mySessionCallStateInquireResult = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE_RESULT', ParameterName='SESSION_CALL_STATE')
            self.myContactDisplayNameInquireResult = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE_RESULT', ParameterName='CONTACT_DISPLAY_NAME')
            self.myContactPhoneNumberListInquireResult = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE_RESULT', ParameterName='CONTACT_PHONE_NUMBER_LIST')
            self.myActiveSessionIDInquireResult = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE_RESULT', ParameterName='ACTIVE_SESSION_ID')
            self.myMeetUrlInquireResult = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE_RESULT', ParameterName='MEET_URL')
            self.mySessionParticipantsInquireResult = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE_RESULT', ParameterName='SESSION_PARTICIPANTS')
            self.mySessionInfoInquireResult = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE_RESULT', ParameterName='SESSION_INFO')
            self.mySessionRegStatusInquireResult = EPMsg(EndPointName=EndPointName, BlockNumber='INQUIRE_RESULT', ParameterName='REG_STATUS')

            self.myRegSucceedNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='REG_SUCCEED')
            self.myRegFailedNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='REG_FAILED')
            self.myGroupCreatedNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='GROUP_CREATED')
            self.myContactAddedToGroupNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='CONTACT_ADDED_TO_GROUP')
            self.myContactRemovedFromGroupNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='CONTACT_REMOVED_FROM_GROUP')
            self.myGroupRemovedNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='GROUP_REMOVED')
            self.mySessionCreatedNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='SESSION_CREATED')
            self.myParticipantAddedToSessionNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='PARTICIPANT_ADDED_TO_SESSION')
            self.myParticipantRemovedFromSessionNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='PARTICIPANT_REMOVED_FROM_SESSION')
            self.mySessionCallStateChangeNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='SESSION_CALL_STATE_CHANGE')
            self.myActiveSessionNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='ACTIVE_SESSION')
            self.mySessionEndedNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='SESSION_ENDED')
            self.myErrorNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='ERROR')
            self.myContactRemovedFromAllGroupsNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='CONTACT_REMOVED_FROM_ALL_GROUPS')
            self.mySessionUpdatedNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='SESSION_UPDATED')
            self.myIWasRemovedFromSessionNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='I_WAS_REMOVED_FROM_SESSION')
            self.mySFBReadyNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='SFB_READY')
            self.myInviteJoinAudioNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='INVITE_JOIN_AUDIO')
            self.mySFBInitializingNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='SFB_INITIALIZING')
            self.mySFBResettingNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='SFB_RESETTING')
            self.mySFBShuttingDownNotification = EPMsg(EndPointName=EndPointName, BlockNumber='NOTIFICATION', ParameterName='SFB_SHUTTING_DOWN')

            self.MyControlPoints = [self.myKeyCallCommand, self.myKeyDigitPressedCommand, self.myKeyDigitReleasedCommand,
                                    self.myKeyHookCommand, self.myKeyHoldCommand, self.myKeyResumeCommand,
                                    self.myKeyAddParticipantCommand, self.myKeyRejectCommand, self.myKeyCreateGroupCommand,
                                    self.myKeyDeleteGroupCommand, self.myKeyAddToGroupCommand, self.myKeyRemoveFromGroupCommand,
                                    self.myKeyRemoveFromAllGroupsCommand, self.myKeySetRingToneCommand, self.myKeyPlayRingToneCommand,
                                    self.myKeyAutoAnswerCommand, self.myKeySetActiveSessionCommand, self.myKeyCreateSessionCommand,
                                    self.myKeyAcceptSessionCommand, self.myKeyLeaveSessionCommand, self.myKeyAddAudioToSessionCommand,
                                    self.myKeyJoinMeetingCommand, self.myKeyJoinConferenceCommand, self.myKeyMeetNowCommand,
                                    self.myContactSearchInquire, self.myGroupsInquire, self.myGroupsMembersInquire,
                                    self.myRegStatusInquire, self.myActiveRingToneInquire, self.myAutoAnswerModeInquire,
                                    self.myCurrentSessionsInquire, self.mySessionCallStateInquire, self.myContactDisplayNameInquire,
                                    self.myContactPhoneNumberListInquire, self.myActiveSessionIDInquire, self.myMeetUrlInquire,
                                    self.mySessionParticipantsInquire, self.mySessionInfoInquire, self.myRegStatusInquire,
                                    self.myContactSearchInquireResult, self.myGroupsInquireResult, self.myGroupsMembersInquireResult,
                                    self.myRegStatusInquireResult, self.myActiveRingToneInquireResult, self.myAutoAnswerModeInquireResult,
                                    self.myCurrentSessionsInquireResult, self.mySessionCallStateInquireResult, self.myContactDisplayNameInquireResult,
                                    self.myContactPhoneNumberListInquireResult, self.myActiveSessionIDInquireResult, self.myMeetUrlInquireResult,
                                    self.mySessionParticipantsInquireResult, self.mySessionInfoInquireResult, self.myRegStatusInquireResult,
                                    self.myRegSucceedNotification, self.myRegFailedNotification, self.myGroupCreatedNotification,
                                    self.myContactAddedToGroupNotification, self.myContactRemovedFromGroupNotification, self.myGroupRemovedNotification,
                                    self.mySessionCreatedNotification, self.myParticipantAddedToSessionNotification, self.myParticipantRemovedFromSessionNotification,
                                    self.mySessionCallStateChangeNotification, self.myActiveSessionNotification, self.mySessionEndedNotification,
                                    self.myErrorNotification, self.myContactRemovedFromAllGroupsNotification, self.mySessionUpdatedNotification,
                                    self.myIWasRemovedFromSessionNotification, self.mySFBReadyNotification, self.myInviteJoinAudioNotification,
                                    self.mySFBInitializingNotification, self.mySFBResettingNotification, self.mySFBShuttingDownNotification]

            self.__OnKeypadTextChange = self.__unassigned
            self.__OnFilterSearchTextChange = self.__unassigned
            self.__OnContactListUpdate = self.__unassigned
            self.__OnSelectedItemChange = self.__unassigned
            self.__OnSessionListUpdate = self.__unassigned
            self.__OnActiveSessionInfoUpdate = self.__unassigned
            self.__OnErrorNotification = self.__unassigned
            self.__OnRegistrationStatusChange = self.__unassigned

        except Exception as e:
            print('Exception.SkypeComponent.__init__: ' + str(e))
    
    def __unassigned(self, *args):
        pass

    @property
    def OnKeypadTextChange(self):
        return self.__OnKeypadTextChange
    
    @OnKeypadTextChange.setter
    def OnKeypadTextChange(self, handler):
        if callable(handler):
            self.__OnKeypadTextChange = handler
        else:
            raise ValueError('handler must be a function')

    @property
    def OnFilterSearchTextChange(self):
        return self.__OnFilterSearchTextChange
    
    @OnFilterSearchTextChange.setter
    def OnFilterSearchTextChange(self, handler):
        if callable(handler):
            self.__OnFilterSearchTextChange = handler
        else:
            raise ValueError('handler must be a function')

    @property
    def OnContactListUpdate(self):
        return self.__OnContactListUpdate
    
    @OnContactListUpdate.setter
    def OnContactListUpdate(self, handler):
        if callable(handler):
            self.__OnContactListUpdate = handler
        else:
            raise ValueError('handler must be a function')
    
    @property
    def OnSelectedItemChange(self):
        return self.__OnSelectedItemChange
    
    @OnSelectedItemChange.setter
    def OnSelectedItemChange(self, handler):
        if callable(handler):
            self.__OnSelectedItemChange = handler
        else:
            raise ValueError('handler must be a function')
 
    @property
    def OnSessionListUpdate(self):
        return self.__OnSessionListUpdate
    
    @OnSessionListUpdate.setter
    def OnSessionListUpdate(self, handler):
        if callable(handler):
            self.__OnSessionListUpdate = handler
        else:
            raise ValueError('handler must be a function')

    @property
    def OnActiveSessionInfoUpdate(self):
        return self.__OnActiveSessionInfoUpdate
    
    @OnActiveSessionInfoUpdate.setter
    def OnActiveSessionInfoUpdate(self, handler):
        if callable(handler):
            self.__OnActiveSessionInfoUpdate = handler
        else:
            raise ValueError('handler must be a function')

    @property
    def OnErrorNotification(self):
        return self.__OnErrorNotification
    
    @OnErrorNotification.setter
    def OnErrorNotification(self, handler):
        if callable(handler):
            self.__OnErrorNotification = handler
        else:
            raise ValueError('handler must be a function')

    @property
    def OnRegistrationStatusChange(self):
        return self.__OnRegistrationStatusChange
    
    @OnRegistrationStatusChange.setter
    def OnRegistrationStatusChange(self, handler):
        if callable(handler):
            self.__OnRegistrationStatusChange = handler
        else:
            raise ValueError('handler must be a function')

    def HasActiveCall(self):
        return self.IsRegistered() and isinstance(self.myCurrentActiveSessionID, str) and len(self.myCurrentActiveSessionID) > 0 and self.myCurrentActiveSessionID != '-1'
    
    def IsRegistered(self):
        return isinstance(self.myRegistrationStatus, S4BRegistrationStatus) and self.myRegistrationStatus.State == S4BRegState['REGISTERED']

    def OnHook(self):
        if self.IsRegistered():
            req = EPMsg(Msg=self.myKeyHookCommand)
            req.Value = 0

            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def OffHook(self):
        if self.IsRegistered():
            req = EPMsg(Msg=self.myKeyHookCommand)
            req.Value = 1

            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)        
    
    def HookToggle(self):
        if self.IsRegistered():
            req = EPMsg(Msg=self.myKeyHookCommand)
            req.Value = 2

            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND) 

    def Hold(self):
        if self.IsRegistered():
            req = EPMsg(Msg=self.myKeyHoldCommand)
            req.Value = ''

            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)

    def Resume(self):
        if self.IsRegistered():
            req = EPMsg(Msg=self.myKeyResumeCommand)
            req.Value = ''

            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)

    def Reject(self):
        if self.IsRegistered():
            req = EPMsg(Msg=self.myKeyRejectCommand)
            req.Value = ''

            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)

    def Accept(self):
        if self.IsRegistered():
            req = EPMsg(Msg=self.myKeyAcceptSessionCommand)
            req.Value = ''

            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)

    def Leave(self):
        if self.IsRegistered():
            req = EPMsg(Msg=self.myKeyLeaveSessionCommand)
            req.Value = ''

            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)

    def KeypadPress(self, digit):
        if re.search(r'[\d*#+]', digit) is not None and len(digit) == 1:
            self.myKeypadText += str(digit)
            self.UpdateKeypadTextChange()

            if self.HasActiveCall() and digit != '+':
                req = EPMsg(Msg=self.myKeyDigitPressedCommand)
                req.Value = digit
            
                self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
            
    def KeypadRelease(self, digit):
        if re.search(r'[\d*#+]', digit) is not None and len(digit) == 1:
            if self.HasActiveCall() and digit != '+':
                req = EPMsg(Msg=self.myKeyDigitReleasedCommand)
                req.Value = digit

                self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)

    def KeypadClear(self):
        self.myKeypadText = ''
        self.UpdateKeypadTextChange()
    
    def KeypadBackspace(self):
        if isinstance(self.myKeypadText, str) and len(self.myKeypadText) > 0:
            self.myKeypadText = self.myKeypadText[:-1]
            self.UpdateKeypadTextChange()

    def KeypadDial(self):
        if self.IsRegistered() and isinstance(self.myKeypadText, str) and len(self.myKeypadText) > 0:
            req = EPMsg(Msg=self.myKeyCallCommand)
            req.Value = self.myKeypadText

            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)

            self.myKeypadText = ''

    def DialString(self, dial):
        self.myKeypadText = dial
        self.UpdateKeypadTextChange()

    def MeetNow(self):
        if self.IsRegistered():
            req = EPMsg(Msg=self.myKeyMeetNowCommand)
            req.Value = ''

            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)

    def JoinMeeting(self, url):
        if self.IsRegistered() and isinstance(url, str) and len(url) > 0:
            req = EPMsg(Msg=self.myKeyJoinMeetingCommand)
            req.Value = url

            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def JoinConference(self, url):
        if self.IsRegistered() and isinstance(url, str) and len(url) > 0:
            req = EPMsg(Msg=self.myKeyJoinConferenceCommand)
            req.Value = url

            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def SelectSession(self, index):
        if self.IsRegistered() and isinstance(self.mySessions, list) and len(self.mySessions) >= index:
            req = EPMsg(Msg=self.myKeySetActiveSessionCommand)
            req.Value = self.mySessions[index - 1].ID

            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
    
    def ResetList(self):
        self.myCurrentList = S4BContactListResponse()
        self.myCurrentListSelectedItem = S4BSelectedItem()
        self.UpdateContactListChange(True, page=1)
        self.UpdateContactListSelectedItemChange()
        self.ResetFilterSearch()

    def SelectListItem(self, index):       
        if isinstance(self.myCurrentList, S4BContactListResponse):
            if self.myCurrentList.Type == S4BListType['GROUPS'] and isinstance(self.myGroups, list) and len(self.myGroups) >= index and self.myCurrentList.Function == S4BListFunction['SELECT_GROUP']:
                grp = self.myGroups[index - 1]
   
                if isinstance(grp.Members, list):
                    self.myCurrentList = S4BContactListResponse()
                    self.myCurrentList.Type = S4BListType['CONTACTS']
                    self.myCurrentList.Function = S4BListFunction['SELECT_CONTACT']
                    self.myCurrentList.HelpText = 'Select a Contact'

                    self.myCurrentListSelectedItem = S4BSelectedItem()
                    self.myCurrentListSelectedItem.Type = S4BListType['GROUPS']
                    self.myCurrentListSelectedItem.DisplayName = grp.Name
                    self.myCurrentListSelectedItem.Group = ''
                    self.myCurrentListSelectedItem.Tag = grp
 
                    for item in grp.Members:
                        self.myCurrentList.List.append(item.Name)
                    
                    self.UpdateContactListChange(False, page=1)
                    self.UpdateContactListSelectedItemChange()
            
            elif self.myCurrentList.Type == S4BListType['CONTACTS'] and isinstance(self.myCurrentListSelectedItem.Tag, S4BGroup) and self.myCurrentList.Function == S4BListFunction['SELECT_CONTACT']:
                grp = self.myCurrentListSelectedItem.Tag

                if isinstance(grp.Members, list) and len(grp.Members) >= index:
                    cnt = grp.Members[index - 1]

                    self.myCurrentListSelectedItem = S4BSelectedItem()
                    self.myCurrentListSelectedItem.DisplayName = cnt.Name
                    self.myCurrentListSelectedItem.Type = S4BListType['CONTACTS']
                    self.myCurrentListSelectedItem.Group = grp.Name
                    self.myCurrentListSelectedItem.Tag = cnt

                    self.UpdateContactListSelectedItemChange()
            
            elif self.myCurrentList.Type == S4BListType['CONTACTS'] and self.myCurrentList.Function == S4BListFunction['SELECT_CONTACT']:
                found = False

                for grp in self.myGroups:
                    if found:
                        break

                    if isinstance(grp, S4BGroup):
                        for member in grp.Members:
                            if isinstance(member, S4BGroupMember):
                                if member.Name == self.myCurrentList.List[index - 1]:
                                    found = True

                                    self.myCurrentListSelectedItem = S4BSelectedItem()
                                    self.myCurrentListSelectedItem.DisplayName = member.Name
                                    self.myCurrentListSelectedItem.Type = S4BListType['CONTACTS']
                                    self.myCurrentListSelectedItem.Group = grp.Name
                                    self.myCurrentListSelectedItem.Tag = member

                                    self.UpdateContactListSelectedItemChange()
                                    break

            elif self.myCurrentList.Type == S4BListType['GROUPS'] and isinstance(self.myCurrentListSelectedItem.Tag, S4BGroupMember) and self.myCurrentList.Function == S4BListFunction['ADD_TO_GROUP']:
                if self.myCurrentList.Count() >= index:
                    found = False

                    for grp in self.myGroups:
                        if found:
                            break

                        if grp.Name == self.myCurrentList.List[index - 1]:
                            found = True

                            if isinstance(grp, S4BGroup) and grp.ID != '-1':
                                req = EPMsg(Msg=self.myKeyAddToGroupCommand)
                                req.Value = '{0}|{1}'.format(grp.ID, self.myCurrentListSelectedItem.Tag.DialNumber)

                                self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)

    def DialContact(self):
        if self.IsRegistered() and isinstance(self.myCurrentListSelectedItem, S4BSelectedItem) and self.myCurrentListSelectedItem.Type == S4BListType['CONTACTS']:
            req = EPMsg(Msg=self.myKeyCallCommand)
            req.Value = self.myCurrentListSelectedItem.Tag.DialNumber

            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)

    def AddContactActiveSession(self):
        if self.IsRegistered() and isinstance(self.myCurrentListSelectedItem, S4BSelectedItem) and self.myCurrentListSelectedItem.Type == S4BListType['CONTACTS']:
            req = EPMsg(Msg=self.myKeyAddParticipantCommand)
            req.Value = self.myCurrentListSelectedItem.Tag.DialNumber

            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
            
    def AddContactToGroup(self):
        if self.IsRegistered() and isinstance(self.myCurrentListSelectedItem, S4BSelectedItem) and self.myCurrentListSelectedItem.Type == S4BListType['CONTACTS']:
            self.myCurrentList = S4BContactListResponse()
            self.myCurrentList.Type = S4BListType['GROUPS']
            self.myCurrentList.Function = S4BListFunction['ADD_TO_GROUP']
            self.myCurrentList.HelpText = 'Select a Group'

            for grp in self.myGroups:
                if grp.ID != '-1':
                    self.myCurrentList.List.append(grp.Name)

            self.UpdateContactListChange(False, page=1)

    def RemoveContactFromGroup(self):
        if self.IsRegistered() and isinstance(self.myCurrentListSelectedItem, S4BSelectedItem) and self.myCurrentListSelectedItem.Type == S4BListType['CONTACTS']:
            if isinstance(self.myCurrentListSelectedItem.Group, str) and len(self.myCurrentListSelectedItem.Group) > 0:
                found = False
                
                for grp in self.myGroups:
                    if found:
                        break
                    
                    if grp.Name == self.myCurrentListSelectedItem.Group:
                        if isinstance(grp, S4BGroup) and grp.ID != '-1':
                            req = EPMsg(Msg=self.myKeyRemoveFromGroupCommand)
                            req.Value = '{0}|{1}'.format(grp.ID, self.myCurrentListSelectedItem.Tag.DialNumber)
  
                            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
            
    def RemoveContactFromAllGroups(self):
        if self.IsRegistered() and isinstance(self.myCurrentListSelectedItem, S4BSelectedItem) and self.myCurrentListSelectedItem.Type == S4BListType['CONTACTS']:
            req = EPMsg(Msg=self.myKeyRemoveFromAllGroupsCommand)
            req.Value = self.myCurrentListSelectedItem.Tag.DialNumber

            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)

    def CreateSessionUsingSelection(self):
        if self.IsRegistered():
            if isinstance(self.myCurrentListSelectedItem, S4BSelectedItem) and self.myCurrentListSelectedItem.Type == S4BListType['GROUPS']:
                if isinstance(self.myCurrentListSelectedItem.Tag, S4BGroup):
                    grp = self.myCurrentListSelectedItem.Tag

                    if isinstance(grp.Members, list) and len(grp.Members) > 0:
                        req_1 = EPMsg(Msg=self.myKeyCreateSessionCommand)
                        req_2 = EPMsg(Msg=self.myKeyAddAudioToSessionCommand)
                        
                        req_1.Value = ''
                        req_2.Value = ''
                        
                        for mem in grp.Members:
                            req_1.Value += '{0}|'.format(mem.DialNumber)
                            req_2.Value += '{0}|'.format(mem.DialNumber)
                            
                        req_1.Value.rstrip('|')
                        req_2.Value.rstrip('|')

                        self.parent.MyQueue.Enqueue(MessageBundle(self, req_1.BuildSetMsg()), STATUS_COMMAND)
                        self.parent.MyQueue.Enqueue(MessageBundle(self, req_2.BuildSetMsg()), STATUS_COMMAND)
            
            elif isinstance(self.myCurrentListSelectedItem, S4BSelectedItem) and self.myCurrentListSelectedItem.Type == S4BListType['CONTACTS']:
                req_1 = EPMsg(Msg=self.myKeyCreateSessionCommand)
                req_2 = EPMsg(Msg=self.myKeyAddAudioToSessionCommand)
                
                req_1.Value = self.myCurrentListSelectedItem.Tag.DialNumber
                req_2.Value = self.myCurrentListSelectedItem.Tag.DialNumber
                
                self.parent.MyQueue.Enqueue(MessageBundle(self, req_1.BuildSetMsg()), STATUS_COMMAND)
                self.parent.MyQueue.Enqueue(MessageBundle(self, req_2.BuildSetMsg()), STATUS_COMMAND)

    def DeleteGroup(self):
        if self.IsRegistered():
            if isinstance(self.myCurrentListSelectedItem, S4BSelectedItem) and self.myCurrentListSelectedItem.Type == S4BListType['GROUPS']:
                if isinstance(self.myCurrentListSelectedItem.Tag, S4BGroup):
                    grp = self.myCurrentListSelectedItem.Tag

                    if grp.ID != '-1':
                        req = EPMsg(Msg=self.myKeyDeleteGroupCommand)
                        req.Value = grp.ID

                        self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)

    def SetFilterSearchText(self, _filter):
        if isinstance(self.myFilterSearchText, str) == False or self.myFilterSearchText != _filter:
            self.myFilterSearchText = _filter
            self.UpdateFilterSearchTextChange()

            if len(_filter) > 0:
                filteredList = []

                for grp in self.myGroups:
                    for item in grp.Members:
                        if _filter.lower() in item.Name.lower():
                            filteredList.append(item.Name)

                self.myCurrentList = S4BContactListResponse()
                self.myCurrentList.Type = S4BListType['CONTACTS']
                self.myCurrentList.Function = S4BListFunction['SELECT_CONTACT']
                self.myCurrentList.HelpText = 'Select a Contact'

                self.myCurrentListSelectedItem = S4BSelectedItem()

                uniqueList = sorted(set(filteredList))
                
                for i in uniqueList:
                    self.myCurrentList.List.append(i)
                
                self.UpdateContactListChange(False, page=1)
                self.UpdateContactListSelectedItemChange()
            
            else:
                self.ResetList()

    def ResetFilterSearch(self):
        self.SetFilterSearchText('')
    
    def PerformSearch(self):
        if self.IsRegistered() and isinstance(self.myFilterSearchText, str) and len(self.myFilterSearchText) > 0:
            req = EPMsg(Msg=self.myContactSearchInquire)
            req.Value = '"{0}"'.format(self.myFilterSearchText)

            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)

            self.myCurrentList = S4BContactListResponse()
            self.myCurrentList.Type = S4BListType['INFO_ONLY']
            self.myCurrentList.Function = S4BListFunction['SEARCHING']
            self.myCurrentList.HelpText = 'Searching Contacts'
            self.myCurrentList.List.append('Searching...')

            self.myCurrentListSelectedItem = S4BSelectedItem()

            self.UpdateContactListChange(False, page=1)
            self.UpdateContactListSelectedItemChange()

    def CreateGroup(self, group):
        if self.IsRegistered() and isinstance(group, str) and len(group) > 0:
            req = EPMsg(Msg=self.myKeyCreateGroupCommand)
            req.Value = '"{0}"'.format(group)

            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)

    def FindMember(self, dialNumber):
        for grp in self.myGroups:
            for item in grp.Members:
                if item.DialNumber == dialNumber:
                    return item

    def UpdateMember(dialNumber, name):
        for grp in self.myGroups:
            for item in grp.Members:
                if item.DialNumber == dialNumber:
                    item.Name = name

    def StartContactListRefreshTimer(self, time):
        self.myContactListRefreshTimer = Wait(time, self.OnElaspedContactListRefreshTimer)

    def StopContactListRefreshTimer(self):        
        if self.myContactListRefreshTimer is not None:
            self.myContactListRefreshTimer.Cancel()
            self.myContactListRefreshTimer = None

    def ResetContactListRefreshTimer(self, time=5):        
        self.StopContactListRefreshTimer()
        self.StartContactListRefreshTimer(time)

    def OnElaspedContactListRefreshTimer(self):
        self.ResetList()
    
    def StartSessionListRefreshTimer(self, time):
        self.mySessionListRefreshTimer = Wait(time, self.OnElaspedSessionRefreshTimer)

    def StopSessionListRefreshTimer(self):        
        if self.mySessionListRefreshTimer is not None:
            self.mySessionListRefreshTimer.Cancel()
            self.mySessionListRefreshTimer = None

    def ResetSessionListRefreshTimer(self, time=.5):        
        self.StopSessionListRefreshTimer()
        self.StartSessionListRefreshTimer(time)

    def OnElaspedSessionRefreshTimer(self):
        sessions = S4BSessionListResponse()
        info = S4BActiveSessionInfo()
        
        if isinstance(self.myCurrentActiveSessionID, str) and len (self.myCurrentActiveSessionID) > 0 and self.myCurrentActiveSessionID != '-1':
            location = -1
            
            for i in range(len(self.mySessions)):
                if self.mySessions[i].ID == self.myCurrentActiveSessionID:
                    location = i
                    break            
            
            sessions.ActiveItem = location + 1

            if location >= 0:
                info.MeetingID = self.mySessions[location].DisplayName
                info.Url = self.mySessions[location].ContactUrl
                info.Status = self.mySessions[location].Status
                info.StatusText = self.mySessions[location].StatusText
                info.IsConference = self.mySessions[location].Conference   

        for session in self.mySessions:
            sessions.List.append(session)
        
        self.UpdateSessionListChange(sessions)
        self.UpdateActiveSessionInfoChange(info)

    def Poll(self):
        if self.IsInitialized() and self.IsRegistered():
            self.parent.MyQueue.Enqueue(MessageBundle(self, self.myRegStatusInquire.BuildQueryMsg()), STATUS_QUERY)
            self.parent.MyQueue.Enqueue(MessageBundle(self, self.myGroupsInquire.BuildQueryMsg()), STATUS_QUERY)
            self.parent.MyQueue.Enqueue(MessageBundle(self, self.myCurrentSessionsInquire.BuildQueryMsg()), STATUS_QUERY)
            self.parent.MyQueue.Enqueue(MessageBundle(self, self.myActiveSessionIDInquire.BuildQueryMsg()), STATUS_QUERY)

    def Reinitialize(self):
        self.myKeypadText = ''
        self.mySessions = None
        self.myGroups = []        
        self.myCurrentActiveSessionID = ''
        
        self.UpdateKeypadTextChange()
        self.UpdateRegistrationStatus()
        
        for item in self.MyControlPoints:
            item.Reinitialize()

    def Refresh(self):
        pass

    def GetInitialized(self):
        self.parent.MyQueue.Enqueue(MessageBundle(self, self.myRegStatusInquire.BuildQueryMsg()), STATUS_QUERY)
        self.parent.MyQueue.Enqueue(MessageBundle(self, self.myGroupsInquire.BuildQueryMsg()), STATUS_QUERY)
        self.parent.MyQueue.Enqueue(MessageBundle(self, self.myCurrentSessionsInquire.BuildQueryMsg()), STATUS_QUERY)
        self.parent.MyQueue.Enqueue(MessageBundle(self, self.myActiveSessionIDInquire.BuildQueryMsg()), STATUS_QUERY)

    def IsInitialized(self):
        if isinstance(self.myRegistrationStatus, S4BRegistrationStatus) == False:
            return False
        
        if isinstance(self.mySessions, list) == False:
            return False

        if isinstance(self.myGroups, list) == False:
            return False
        
        if isinstance(self.myCurrentActiveSessionID, str) == False or len(self.myCurrentActiveSessionID) == 0:
            return False
        
        return True

    def ProcessSubscription(self, msg):
        if isinstance(msg, EPResponseMsg):
            if msg == self.myRegStatusInquireResult:
                self.myRegistrationStatus = S4BRegistrationStatus()

                try:
                    self.myRegistrationStatus.SetState(msg.Message.Value)
                    self.UpdateRegistrationStatus()
                    
                except Exception as e:
                    print('Exception.SkypeComponent.ProcessSubscription: ' + str(e))
            
            elif msg == self.myRegSucceedNotification or msg == self.myRegFailedNotification:
                self.parent.MyQueue.Enqueue(MessageBundle(self, self.myRegStatusInquire.BuildQueryMsg()), STATUS_QUERY)

            elif msg == self.myCurrentSessionsInquireResult:
                items = msg.Message.Value.split('|')
                
                if isinstance(items, list) and len(items) > 0:
                    self.mySessions = []

                    if items[0] != '0':
                        sessionids = items.copy()
                        sessionids.pop(0)

                        for id in sessionids:
                            req = EPMsg(Msg=self.mySessionInfoInquire)
                            req.Value = id

                            self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)
            
            elif msg == self.mySessionInfoInquireResult:
                rm = re.search(r'(?P<id>\d+) (?P<active>1|0)\|(?P<url>.+)\|(?P<displayname>.+)', msg.Message.Value)

                if rm is not None:
                    if isinstance(self.mySessions, list):
                        session = None
                        
                        for item in self.mySessions:
                            if item.ID == rm.group('id'):
                                session = item
                                break

                        if session is None:
                            session = S4BSession()
                            session.ID = rm.group('id')
                            session.Status = S4BSessionStatus['UNKNOWN']

                            self.mySessions.append(session)

                        session.Conference = rm.group('active') == '1'
                        session.ContactUrl = rm.group('url')
                        session.DisplayName = rm.group('displayname')
                        
                        self.ResetSessionListRefreshTimer()

            elif msg == self.mySessionCreatedNotification or msg == self.mySessionUpdatedNotification:
                rm = re.search(r'(?P<id>\d+)\|(?P<active>1|0)\|(?P<url>.+)\|(?P<displayname>.+)', msg.Message.Value)

                if rm is not None:
                    if isinstance(self.mySessions, list):
                        session = None
                        
                        for item in self.mySessions:
                            if item.ID == rm.group('id'):
                                session = item
                                break

                        if session is None:
                            session = S4BSession()
                            session.ID = rm.group('id')
                            session.Status = S4BSessionStatus['UNKNOWN']
                            
                            self.mySessions.append(session)

                        session.Conference = rm.group('active') == '1'
                        session.ContactUrl = rm.group('url')
                        session.DisplayName = rm.group('displayname')

                        self.ResetSessionListRefreshTimer()
            
            elif msg == self.mySessionCallStateInquireResult:
                rm = re.search(r'(?P<id>\d+) (?P<url>.+)\|(?P<status>\w+)(\^(?P<alt>.+))?', msg.Message.Value)

                if rm is not None:
                    if isinstance(self.mySessions, list):
                        session = None
                        
                        for item in self.mySessions:
                            if item.ID == rm.group('id'):
                                session = item
                                break

                        if session is None:
                            session = S4BSession()
                            session.ID = rm.group('id')
                            session.ContactUrl = rm.group('url')
                            session.DisplayName = ''

                            self.mySessions.append(session)
                        
                        session.SetStatus(rm.group('status'))
                        
                        self.ResetSessionListRefreshTimer()

            elif msg == self.mySessionCallStateChangeNotification:
                rm = re.search(r'(?P<id>\d+)\|(?P<url>.+)\|(?P<status>\w+)(\^(?P<duration>.+))?', msg.Message.Value)
                
                if rm is not None:
                    if isinstance(self.mySessions, list):
                        session = None
                        
                        for item in self.mySessions:
                            if item.ID == rm.group('id'):
                                session = item
                                break

                        if isinstance(session, S4BSession):
                            session.SetStatus(rm.group('status'))

                            if session.Status == S4BSessionStatus['INCOMING'] or session.Status == S4BSessionStatus['CONFERENCE_JOIN'] or session.Status == S4BSessionStatus['INVITE_JOIN_AUDIO']:
                                req = EPMsg(Msg=self.myKeySetActiveSessionCommand)
                                req.Value = session.ID

                                self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND)

                            self.ResetSessionListRefreshTimer()

            elif msg == self.mySessionEndedNotification:
                rm = re.search(r'(?P<id>\d+)', msg.Message.Value)

                if rm is not None:
                    if isinstance(self.mySessions, list):
                        session = None
                        
                        for item in self.mySessions:
                            if item.ID == rm.group('id'):
                                session = item
                                break

                        if isinstance(session, S4BSession):
                            self.mySessions.remove(session)
                            self.ResetSessionListRefreshTimer()
            
            elif msg == self.myActiveSessionIDInquireResult or msg == self.myActiveSessionNotification:
                self.myCurrentActiveSessionID = msg.Message.Value
                self.ResetSessionListRefreshTimer()

            elif msg == self.myErrorNotification:
                self.UpdateErrorNotificationChange(msg.Message.Value)
                self.ResetContactListRefreshTimer()

            elif msg == self.myGroupCreatedNotification:
                rm = re.search(r'(?P<id>\d+)\|(?P<groupname>.+)', msg.Message.Value)

                if rm is not None:
                    grp = S4BGroup()
                    grp.ID = rm.group('id')
                    grp.Name = rm.group('groupname')
                    self.myGroups.append(grp)

                    self.ResetContactListRefreshTimer()

            elif msg == self.myGroupRemovedNotification:
                rm = re.search(r'(?P<id>\d+)', msg.Message.Value)

                if rm is not None:
                    if isinstance(self.myGroups, list):
                        grp = None
                        
                        for item in self.myGroups:
                            if item.ID == rm.group('id'):
                                grp = item
                                break

                        if isinstance(grp, S4BGroup):
                            self.myGroups.remove(grp)

                            self.ResetContactListRefreshTimer()

            elif msg == self.myContactSearchInquireResult:
                rm = re.search(r'"(?P<search>.+)" (?P<count>\d+)\|(?P<item>\d+)(\|(?P<id>.+)\|(?P<name>.+))?', msg.Message.Value)

                if rm is not None:
                    cnt = rm.group('count')
                    item = rm.group('item')
                    name = rm.group('name')
                    ID = rm.group('id')

                    if (isinstance(cnt, str) and len(cnt) > 0 and isinstance(item, str) and len(item) > 0 and isinstance(name, str) and len(name) > 0 and isinstance(ID, str) and len(ID) > 0) == False:
                        self.myCurrentList = S4BContactListResponse()
                        self.myCurrentList.Type = S4BListType['INFO_ONLY']
                        self.myCurrentList.Function = S4BListFunction['SEARCHING']
                        self.myCurrentList.HelpText = 'Search returned Zero(0) results.'
                        self.myCurrentList.List.append('Search returned Zero(0) results.')

                        self.myCurrentListSelectedItem = S4BSelectedItem()

                        self.UpdateContactListChange(False, page=1)
                        self.UpdateContactListSelectedItemChange()

                        return
                    
                    if item == '1' or item == '0':
                        self.myRecentSearches = S4BGroup()
                        self.myRecentSearches.Name = 'Recent Search Results'
                        self.myRecentSearches.ID = '-1'
                        self.myRecentSearches.Members = []

                    mem = S4BGroupMember()
                    mem.Name = name
                    mem.DialNumber = ID

                    self.myRecentSearches.Members.append(mem)

                    if cnt == item:
                        if isinstance(self.myGroups, list):
                            grp = None
                            
                            for i in self.myGroups:
                                if i.ID == '-1':
                                    grp = i
                                    break

                            if isinstance(grp, S4BGroup):
                                self.myGroups.remove(grp)
                            
                            self.myGroups.insert(0, self.myRecentSearches)

                            self.ResetList()
                            self.SelectListItem(1)

            elif msg == self.myContactRemovedFromAllGroupsNotification:
                rm = re.search(r'(?P<who>.+)', msg.Message.Value)

                if rm is not None:
                    if isinstance(self.myGroups, list):
                        for grp in self.myGroups:
                            if grp.ID != '-1':
                                for item in grp.Members:
                                    if item.DialNumber == rm.group('who'):
                                        grp.Members.remove(item)

                        self.ResetContactListRefreshTimer()

            elif msg == self.myContactRemovedFromGroupNotification:
                rm = re.search(r'(?P<id>\d+)\|(?P<who>.+)', msg.Message.Value)

                if rm is not None:
                    if isinstance(self.myGroups, list):
                        grp = None
                        
                        for item in self.myGroups:
                            if item.ID == rm.group('id'):
                                grp = item
                                break

                        if isinstance(grp, S4BGroup):
                            if isinstance(grp.Members, list):
                                for i in grp.Members:
                                    if i.DialNumber == rm.group('who'):
                                        grp.Members.remove(i)

                            self.ResetContactListRefreshTimer()
            
            elif msg == self.myContactAddedToGroupNotification:
                rm = re.search(r'(?P<id>\d+)\|(?P<who>.+)', msg.Message.Value)
                
                if rm is not None:
                    if isinstance(self.myGroups, list):
                        grp = None
                        
                        for item in self.myGroups:
                            if item.ID == rm.group('id'):
                                grp = item
                                break

                        if isinstance(grp, S4BGroup):
                            dialnumber = str(rm.group('who'))
                            mem = None

                            if isinstance(grp.Members, list):
                                for i in grp.Members:
                                    if i.DialNumber == rm.group('who'):
                                        mem = S4BGroupMember()
                                        break

                                if isinstance(mem, S4BGroupMember) == False:
                                    mem = self.FindMember(dialnumber)

                                    if isinstance(mem, S4BGroupMember) == False:
                                        mem = S4BGroupMember()
                                        mem.DialNumber = dialnumber
                                        mem.Name = dialnumber

                                        req = EPMsg(Msg=self.myContactDisplayNameInquire)
                                        req.Value = dialnumber

                                        self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND) 
                                    
                                    grp.Members.append(mem)

                                    self.ResetContactListRefreshTimer()
            
            elif msg == self.myGroupsInquireResult:
                rm = re.search(r'(?P<count>\d+)\|(?P<item>\d+)(\|(?P<id>.+)\|(?P<name>.+))?', msg.Message.Value)

                if rm is not None:
                    cnt = rm.group('count')
                    item = rm.group('item')
                    name = rm.group('name')
                    ID = rm.group('id')
                                        
                    if (isinstance(cnt, str) and len(cnt) > 0 and isinstance(item, str) and len(item) > 0 and isinstance(name, str) and len(name) > 0 and isinstance(ID, str) and len(ID) > 0) == False:
                        return
                    
                    if item == '1' or item == '0':
                        self.myGroups = []

                    grp = S4BGroup()
                    grp.ID = ID
                    grp.Name = name
                    
                    self.myGroups.append(grp)

                    if cnt == item:
                        if isinstance(self.myGroups, list):
                            for i in self.myGroups:
                                req = EPMsg(Msg=self.myGroupsMembersInquire)
                                req.Value = i.ID

                                self.parent.MyQueue.Enqueue(MessageBundle(self, req.BuildSetMsg()), STATUS_COMMAND) 
            
            elif msg == self.myGroupsMembersInquireResult:
                rm = re.search(r'(?P<id>\d+) (?P<count>\d+)\|(?P<item>\d+)(\|(?P<num>.+)\|(?P<name>.+))?', msg.Message.Value)
                
                if rm is not None:
                    ID = rm.group('id')
                    cnt = rm.group('count')
                    item = rm.group('item')
                    name = rm.group('name')
                    num = rm.group('num')

                    if (isinstance(cnt, str) and len(cnt) > 0 and isinstance(item, str) and len(item) > 0 and isinstance(ID, str) and len(ID) > 0) == False:
                        return
                    
                    for grp in self.myGroups:
                        if grp.ID == ID:
                            if item == '1' or item == '0':
                                grp.Members = []
                            
                            if isinstance(name, str) and len(name) > 0 and isinstance(num, str) and len(num) > 0:
                                mem = S4BGroupMember()
                                mem.DialNumber = num
                                mem.Name = name

                                grp.Members.append(mem)

                            if cnt == item:
                                self.ResetContactListRefreshTimer(time=2)

            elif msg == self.myContactDisplayNameInquireResult:
                rm = re.search(r'(?P<id>\S+) (?P<name>.+)', msg.Message.Value)

                if rm is not None:
                    if isinstance(self.myGroups, list):
                        ID = rm.group('id')
                        name = rm.group('name')
                        
                        if (isinstance(ID, str) and len(ID) > 0 and isinstance(name, str) and len(name) > 0) == False:
                            return

                        self.UpdateMember(ID, name)
                
    def ProcessResponse(self, response):
        pass
    
    def IsGoodResponse(self, response):
        if isinstance(response, EPResponseMsg):
            for item in self.MyControlPoints:
                if response == item:
                    return True

        return False
    
    def UpdateKeypadTextChange(self):
        try:
            self.__OnKeypadTextChange(self, self.myKeypadText)

        except Exception as e:
            print('Exception.SkypeComponent.UpdateKeypadTextChange: ' + str(e))
    
    def UpdateFilterSearchTextChange(self):
        try:
            self.__OnFilterSearchTextChange(self, self.myFilterSearchText)

        except Exception as e:
            print('Exception.SkypeComponent.UpdateFilterSearchTextChange: ' + str(e))

    def UpdateContactListChange(self, forceReset=False, page=1):
        try:
            if isinstance(self.myCurrentList, S4BContactListResponse):
                if forceReset or isinstance(self.myCurrentList, S4BContactListResponse) == False or self.myCurrentList.Type == S4BListType['UNKNOWN']:
                    self.myCurrentList = S4BContactListResponse()
                    self.myCurrentList.Type = S4BListType['GROUPS']
                    self.myCurrentList.Function = S4BListFunction['SELECT_GROUP']
                    self.myCurrentList.HelpText = "Select Group"
                    
                    for grp in self.myGroups:
                        txt = '{0} ({1}) >'.format(grp.Name, len(grp.Members))
                        self.myCurrentList.List.append(txt)

                self.__OnContactListUpdate(self, self.myCurrentList)
                    
        except Exception as e:
            print('Exception.SkypeComponent.UpdateContactListChange: ' + str(e))
    
    def UpdateContactListSelectedItemChange(self):
        try:
            if isinstance(self.myCurrentListSelectedItem, S4BSelectedItem):
                self.__OnSelectedItemChange(self, self.myCurrentListSelectedItem)

        except Exception as e:
            print('Exception.SkypeComponent.UpdateContactListSelectedItemChange: ' + str(e))

    def UpdateSessionListChange(self, resp):
        try:
            if isinstance(resp, S4BSessionListResponse):
                self.__OnSessionListUpdate(self, resp)

        except Exception as e:
            print('Exception.SkypeComponent.UpdateSessionListChange: ' + str(e))

    def UpdateActiveSessionInfoChange(self, info):
        try:
            if isinstance(info, S4BActiveSessionInfo):
                self.__OnActiveSessionInfoUpdate(self, info)

        except Exception as e:
            print('Exception.SkypeComponent.UpdateActiveSessionInfoChange: ' + str(e))

    def UpdateErrorNotificationChange(self, error):
        try:
            if isinstance(error, str):
                self.__OnErrorNotification(self, error)

        except Exception as e:
            print('Exception.SkypeComponent.UpdateErrorNotificationChange: ' + str(e))

    def UpdateRegistrationStatus(self):
        try:
            self.__OnRegistrationStatusChange(self, self.myRegistrationStatus)

        except Exception as e:
            print('Exception.SkypeComponent.UpdateRegistrationStatus: ' + str(e))

################################################################################
## CLASS: ParserUtil
################################################################################

class ParserUtil():
    def ParseResponse(msg):
        try:
            response = ResponseMsg(msg)
            
            if isinstance(response, ResponseMsg):
                if response.Type == ResponseMessageType['BOX']:
                    return BoxResponseMsg(response)
                
                elif response.Type == ResponseMessageType['EP']:
                    return EPResponseMsg(response)
                
                elif response.Type == ResponseMessageType['ROOM']:
                    return RoomResponseMsg(response)
                
                elif response.Type == ResponseMessageType['STACK']:
                    return StackResponseMsg(response)
                
                elif response.Type == ResponseMessageType['MT']:
                    return MTResponseMsg(response)
                
                elif response.Type == ResponseMessageType['MCCF']:
                    return MCCFResponseMsg(response)
        
        except Exception as e:
            print('Exception.ParserUtil: ' + str(e))

################################################################################
## CLASS: Queue
################################################################################

class Queue():
    def __init__(self, listener):
        self.listener = listener
        self.Command = []
        self.Query = []
        self.IsBusy = False

    def Enqueue(self, MessageBundle, priority):
        wasEmpty = self.IsEmpty()

        if priority == STATUS_COMMAND:
            self.Command.insert(0, MessageBundle)
        
        else:
            self.Query.insert(0, MessageBundle)
    
        if wasEmpty:
            self.listener.SendNextItem()

    def Dequeue(self):
        out = None
    
        if self.HasItems() and not self.IsBusy:
            self.IsBusy = True
            
            if len(self.Command) > 0:
                out = self.Command.pop()
            
            elif len(self.Query) > 0:
                out = self.Query.pop()
        
        return out

    def IsEmpty(self):
        return self.Size() == 0

    def HasItems(self):
        return self.Size() > 0

    def Size(self):
        return len(self.Command) + len(self.Query)

################################################################################
## CLASS: SerialClass/EthernetClass
################################################################################

class SerialClass(SerialInterface, DeviceClass):
    def __init__(self, *args, **kwargs):
        SerialInterface.__init__(self, *args, **kwargs)
        DeviceClass.__init__(self, *args, **kwargs)

class EthernetClass(EthernetClientInterface, DeviceClass):
    def __init__(self, *args, **kwargs):
        EthernetClientInterface.__init__(self, *args, **kwargs)
        DeviceClass.__init__(self, *args, **kwargs)