################################################################################
## ControlScript Imports
################################################################################

from extronlib import event, Version
from extronlib.device import eBUSDevice, ProcessorDevice, UIDevice
from extronlib.interface import (ContactInterface, DigitalIOInterface,
    EthernetClientInterface, EthernetServerInterfaceEx, FlexIOInterface,
    IRInterface, RelayInterface, SerialInterface, SWPowerInterface,
    VolumeInterface)
from extronlib.ui import Button, Knob, Label, Level
from extronlib.system import Clock, MESet, Wait

print(Version())

################################################################################
## User Imports
################################################################################

import Clearone_Converge_Pro_2_v1_0_0 as MyClearOne

################################################################################
## Device/Processor Definition
################################################################################

IPCPPro350  = ProcessorDevice('IPCPPro350')

################################################################################
## Device/User Interface Definition
################################################################################

TLPPro1020T = UIDevice('TLPPro1020T')

################################################################################
## Module Setup
################################################################################

CommTypes = {
    'Serial': 0,
    'Ethernet': 1
}

MyCommType = CommTypes['Serial']

#-------------------------------------------------------------------------------
# Primary Comm Module
#-------------------------------------------------------------------------------

if MyCommType == CommTypes['Ethernet']:
    MyModule = MyClearOne.EthernetClass(Hostname='192.168.187.54', IPPort=23)
    MyModule.SetCredentials(Username='clearone', Password='converge')

elif MyCommType == CommTypes['Serial']:
    MyModule = MyClearOne.SerialClass(Host=IPCPPro350, Port='COM1', Baud=57600, Data=8, Parity='None', Stop=1, FlowControl='Off', CharDelay=0, Mode='RS232')
    
#-------------------------------------------------------------------------------
# Components
#-------------------------------------------------------------------------------

# Presets/Macros
MyPresets = MyModule.AddPresetComponent(1, 'Part_A')
MyMacros = MyModule.AddMacroComponent()

# Gain/Mute
MyVolume1 = MyModule.AddGainMuteComponent('GAIN_FINE', 'Mic-01', 5)
MyVolume2 = MyModule.AddGainMuteComponent('GAIN_FINE', 'Mic-02', 5)
MyVolume3 = MyModule.AddGainMuteComponent('GAIN_FINE', 'Mic-03', 5)
MyVolume4 = MyModule.AddGainMuteComponent('GAIN', 'Fader_1', 5)
MyVolume5 = MyModule.AddGainMuteComponent('GAIN', 'Fader_2', 5)
MyVolume6 = MyModule.AddGainMuteComponent('GAIN', 'Speakers', 5)

# Matrix Mixer

MyIn1Out1 = MyModule.AddMatrixComponent('Mic-01', 'Record', 'Gated', 2)
MyIn1Out2 = MyModule.AddMatrixComponent('Mic-01', 'Spare_1', 'Gated', 2)
MyIn1Out3 = MyModule.AddMatrixComponent('Mic-01', 'Spare_2', 'Gated', 2)
MyIn1Out4 = MyModule.AddMatrixComponent('Mic-01', 'Speakers', 'Gated', 2)

MyIn2Out1 = MyModule.AddMatrixComponent('Mic-02', 'Record', 'Gated', 2)
MyIn2Out2 = MyModule.AddMatrixComponent('Mic-02', 'Spare_1', 'Gated', 2)
MyIn2Out3 = MyModule.AddMatrixComponent('Mic-02', 'Spare_2', 'Gated', 2)
MyIn2Out4 = MyModule.AddMatrixComponent('Mic-02', 'Speakers', 'Gated', 2)

MyIn3Out1 = MyModule.AddMatrixComponent('Fader_1', 'Record', 'Crosspoint', 2)
MyIn3Out2 = MyModule.AddMatrixComponent('Fader_1', 'Spare_1', 'Crosspoint', 2)
MyIn3Out3 = MyModule.AddMatrixComponent('Fader_1', 'Spare_2', 'Crosspoint', 2)
MyIn3Out4 = MyModule.AddMatrixComponent('Fader_1', 'Speakers', 'Crosspoint', 2)

MyIn4Out1 = MyModule.AddMatrixComponent('Fader_2', 'Record', 'Crosspoint', 2)
MyIn4Out2 = MyModule.AddMatrixComponent('Fader_2', 'Spare_1', 'Crosspoint', 2)
MyIn4Out3 = MyModule.AddMatrixComponent('Fader_2', 'Spare_2', 'Crosspoint', 2)
MyIn4Out4 = MyModule.AddMatrixComponent('Fader_2', 'Speakers', 'Crosspoint', 2)

# Analog Dialer
MyADialer = MyModule.AddAnalogDialerComponent('TELCO_RX 201')
MyADialerRx = MyModule.AddGainMuteComponent('GAIN', 'Dialer_1_Rx', 5)
MyADialerPrivacy = MyModule.AddEPGenericComponent('Digital', 'Dialer_1_Tx', 'LEVEL', 'MUTE')

# VoIP Dialer
MyVDialer = MyModule.AddVoIPDialerComponent('Voip_1')
MyVDialerRx = MyModule.AddGainMuteComponent('GAIN', 'Voip_1_Rx', 5)
MyVDialerPrivacy = MyModule.AddEPGenericComponent('Digital', 'Voip_1_Tx', 'LEVEL', 'MUTE')

# Room Combiner
MyWall1 = MyModule.AddRoomGenericComponent('Digital', 1, 'IndividualDividerState', P1=1)
MyWall2 = MyModule.AddRoomGenericComponent('Digital', 1, 'IndividualDividerState', P1=2)
MyWall3 = MyModule.AddRoomGenericComponent('Digital', 1, 'IndividualDividerState', P1=3)

#-------------------------------------------------------------------------------
# Component Grouping
#-------------------------------------------------------------------------------

# Gain/Mute
myVolumes = [
    MyVolume1,
    MyVolume2,
    MyVolume3,
    MyVolume4,
    MyVolume5,
    MyVolume6
]

# Matrix Mixer
myCrosspoints = [
    MyIn1Out1,
    MyIn1Out2,
    MyIn1Out3,
    MyIn1Out4,
    MyIn2Out1,
    MyIn2Out2,
    MyIn2Out3,
    MyIn2Out4,
    MyIn3Out1,
    MyIn3Out2,
    MyIn3Out3,
    MyIn3Out4,
    MyIn4Out1,
    MyIn4Out2,
    MyIn4Out3,
    MyIn4Out4
]

# Room Combiner
myWalls = [
    MyWall1,
    MyWall2,
    MyWall3
]

################################################################################
## Constants
################################################################################

#-------------------------------------------------------------------------------
# General
#-------------------------------------------------------------------------------

BUTTON_STATES = ['Pressed', 'Released', 'Held', 'Repeated']

################################################################################
## Buttons/Labels
################################################################################

#-------------------------------------------------------------------------------
# Communicating/Initialized
#-------------------------------------------------------------------------------

btnCommunicating = Button(TLPPro1020T, 2)
btnInitialized = Button(TLPPro1020T, 3)

#-------------------------------------------------------------------------------
# Menu
#-------------------------------------------------------------------------------

btnMenuList = [
    Button(TLPPro1020T, 11),    # Presets/Macros
    Button(TLPPro1020T, 12),    # Gain/Mute
    Button(TLPPro1020T, 13),    # Matrix Mixer
    Button(TLPPro1020T, 14),    # Analog Dialer
    Button(TLPPro1020T, 15),    # VoIP Dialer
    Button(TLPPro1020T, 16)     # Room Combiner
]

btnMenuListME = MESet(btnMenuList)

puMenuNames = [
    'Presets/Macros',
    'Gain/Mute',
    'Matrix Mixer',
    'Analog Dialer', 
    'VoIP Dialer',
    'Room Combiner' 
]

#-------------------------------------------------------------------------------
# Presets/Macros
#-------------------------------------------------------------------------------

btnPresetList = [
    Button(TLPPro1020T, 21),    # Preset 1
    Button(TLPPro1020T, 22),    # Preset 2
    Button(TLPPro1020T, 23)     # Preset 3
]

txtPresetNames = [
    'Part_A_Preset_1',
    'Part_A_Preset_2',
    'Part_A_Preset_3'
]

btnMacroList = [
    Button(TLPPro1020T, 31),    # Macro 1
    Button(TLPPro1020T, 32),    # Macro 2
    Button(TLPPro1020T, 33)     # Macro 3
]

txtMacroNames = [
    'Macro_1',
    'Macro_2',
    'Macro_3'
]

#-------------------------------------------------------------------------------
# Gain/Mute
#-------------------------------------------------------------------------------

btnVolUpList = [
    Button(TLPPro1020T, 41),    # Mic-01
    Button(TLPPro1020T, 42),    # Mic-02
    Button(TLPPro1020T, 43),    # Mic-03
    Button(TLPPro1020T, 44),    # Fader_1
    Button(TLPPro1020T, 45),    # Fader_2
    Button(TLPPro1020T, 46)     # Speakers
]

btnVolDnList = [
    Button(TLPPro1020T, 51),    # Mic-01
    Button(TLPPro1020T, 52),    # Mic-02
    Button(TLPPro1020T, 53),    # Mic-03
    Button(TLPPro1020T, 54),    # Fader_1
    Button(TLPPro1020T, 55),    # Fader_2
    Button(TLPPro1020T, 56)     # Speakers
]

btnVolMuteList = [
    Button(TLPPro1020T, 61),    # Mic-01
    Button(TLPPro1020T, 62),    # Mic-02
    Button(TLPPro1020T, 63),    # Mic-03
    Button(TLPPro1020T, 64),    # Fader_1
    Button(TLPPro1020T, 65),    # Fader_2
    Button(TLPPro1020T, 66)     # Speakers
]

lvlVolList = [
    Level(TLPPro1020T, 71),     # Mic-01
    Level(TLPPro1020T, 72),     # Mic-02
    Level(TLPPro1020T, 73),     # Mic-03
    Level(TLPPro1020T, 74),     # Fader_1
    Level(TLPPro1020T, 75),     # Fader_2
    Level(TLPPro1020T, 76)      # Speakers
]

lvlVolList[0].SetRange(0, 100)
lvlVolList[1].SetRange(0, 100)
lvlVolList[2].SetRange(0, 100)
lvlVolList[3].SetRange(0, 100)
lvlVolList[4].SetRange(0, 100)
lvlVolList[5].SetRange(0, 100)

#-------------------------------------------------------------------------------
# Matrix Mixer
#-------------------------------------------------------------------------------

btnMixerSelectList = [
    Button(TLPPro1020T, 81),    # XPoint 1.1
    Button(TLPPro1020T, 82),    # XPoint 1.2
    Button(TLPPro1020T, 83),    # XPoint 1.3
    Button(TLPPro1020T, 84),    # XPoint 1.4
    Button(TLPPro1020T, 85),    # XPoint 2.1
    Button(TLPPro1020T, 86),    # XPoint 2.2
    Button(TLPPro1020T, 87),    # XPoint 2.3
    Button(TLPPro1020T, 88),    # XPoint 2.4
    Button(TLPPro1020T, 89),    # XPoint 3.1
    Button(TLPPro1020T, 90),    # XPoint 3.2
    Button(TLPPro1020T, 91),    # XPoint 3.3
    Button(TLPPro1020T, 92),    # XPoint 3.4
    Button(TLPPro1020T, 93),    # XPoint 4.1
    Button(TLPPro1020T, 94),    # XPoint 4.2
    Button(TLPPro1020T, 95),    # XPoint 4.3
    Button(TLPPro1020T, 96)     # XPoint 4.4
]

btnMixerStateList = [
    Button(TLPPro1020T, 101),   # XPoint 1.1
    Button(TLPPro1020T, 102),   # XPoint 1.2
    Button(TLPPro1020T, 103),   # XPoint 1.3
    Button(TLPPro1020T, 104),   # XPoint 1.4
    Button(TLPPro1020T, 105),   # XPoint 2.1
    Button(TLPPro1020T, 106),   # XPoint 2.2
    Button(TLPPro1020T, 107),   # XPoint 2.3
    Button(TLPPro1020T, 108),   # XPoint 2.4
    Button(TLPPro1020T, 109),   # XPoint 3.1
    Button(TLPPro1020T, 110),   # XPoint 3.2
    Button(TLPPro1020T, 111),   # XPoint 3.3
    Button(TLPPro1020T, 112),   # XPoint 3.4
    Button(TLPPro1020T, 113),   # XPoint 4.1
    Button(TLPPro1020T, 114),   # XPoint 4.2
    Button(TLPPro1020T, 115),   # XPoint 4.3
    Button(TLPPro1020T, 116)    # XPoint 4.4
]

btnMixerStateListME = MESet(btnMixerStateList)
btnMixerStateListME.SetCurrent(None)

btnMixerXPointVolUp = Button(TLPPro1020T, 121)
btnMixerXPointVolDn = Button(TLPPro1020T, 122)
btnMixerXPointRoute = Button(TLPPro1020T, 123)
lvlMixerXPoint      = Level(TLPPro1020T, 124)

lvlMixerXPoint.SetRange(0, 100)
lvlMixerXPoint.SetLevel(lvlMixerXPoint.Min)

#-------------------------------------------------------------------------------
# Analog Dialer
#-------------------------------------------------------------------------------

btnADialerDialpadList = [
    Button(TLPPro1020T, 140),   # 0
    Button(TLPPro1020T, 131),   # 1
    Button(TLPPro1020T, 132),   # 2
    Button(TLPPro1020T, 133),   # 3 
    Button(TLPPro1020T, 134),   # 4 
    Button(TLPPro1020T, 135),   # 5
    Button(TLPPro1020T, 136),   # 6
    Button(TLPPro1020T, 137),   # 7
    Button(TLPPro1020T, 138),   # 8
    Button(TLPPro1020T, 139),   # 9
    Button(TLPPro1020T, 141),   # Star
    Button(TLPPro1020T, 142)    # Pound
]

btnADialerDialpadDigitList = [
    '0',
    '1',
    '2',
    '3',
    '4',
    '5',
    '6',
    '7',
    '8',
    '9',
    '*',
    '#' 
]

txtADialerDialpad = Label(TLPPro1020T, 152)

btnADialerClear = Button(TLPPro1020T, 143)
btnADialerBack =  Button(TLPPro1020T, 144)
btnADialerAutoAnswer = Button(TLPPro1020T, 145)
btnADialerDial = Button(TLPPro1020T, 146)
btnADialerPickup = Button(TLPPro1020T, 147)
btnADialerHangup = Button(TLPPro1020T, 148)
btnADialerFlash = Button(TLPPro1020T, 149)
btnADialerRedial = Button(TLPPro1020T, 150)

btnADialerPrivacy = Button(TLPPro1020T, 151)
btnADialerRxVolUp   = Button(TLPPro1020T, 161)
btnADialerRxVolDn   = Button(TLPPro1020T, 162)
btnADialerRxVolMute = Button(TLPPro1020T, 163)
lvlADialerRx        = Level(TLPPro1020T, 164)

lvlADialerRx.SetRange(0, 100)

txtADialerIncomingName = Label(TLPPro1020T, 165)
txtADialerIncomingNumber = Label(TLPPro1020T, 166)
btnADialerIncomingAnswer = Button(TLPPro1020T, 167)
btnADialerIncomingIgnore = Button(TLPPro1020T, 168)

#-------------------------------------------------------------------------------
# VoIP Dialer
#-------------------------------------------------------------------------------

btnVDialerDialpadList = [
    Button(TLPPro1020T, 180),   # 0
    Button(TLPPro1020T, 171),   # 1
    Button(TLPPro1020T, 172),   # 2
    Button(TLPPro1020T, 173),   # 3 
    Button(TLPPro1020T, 174),   # 4 
    Button(TLPPro1020T, 175),   # 5
    Button(TLPPro1020T, 176),   # 6
    Button(TLPPro1020T, 177),   # 7
    Button(TLPPro1020T, 178),   # 8
    Button(TLPPro1020T, 179),   # 9
    Button(TLPPro1020T, 181),   # Star
    Button(TLPPro1020T, 182)    # Pound
]

btnVDialerDialpadDigitList = [
    '0', 
    '1', 
    '2', 
    '3', 
    '4', 
    '5', 
    '6', 
    '7', 
    '8', 
    '9', 
    '*', 
    '#'  
]

txtVDialerDialpad = Label(TLPPro1020T, 187)

btnVDialerClear = Button(TLPPro1020T, 183)
btnVDialerBack =  Button(TLPPro1020T, 184)
btnVDialerDial = Button(TLPPro1020T, 185)
btnVDialerOnHook = Button(TLPPro1020T, 231)
btnVDialerOffHook = Button(TLPPro1020T, 232)
btnVDialerTransfer = Button(TLPPro1020T, 233)
btnVDialerBlindTransfer = Button(TLPPro1020T, 234)
btnVDialerRedial = Button(TLPPro1020T, 235)
btnVDialerHold = Button(TLPPro1020T, 236)
btnVDialerFwdOn = Button(TLPPro1020T, 237)
btnVDialerFwdOff = Button(TLPPro1020T, 238)
btnVDialerDndOn = Button(TLPPro1020T, 239)
btnVDialerDndOff = Button(TLPPro1020T, 240)

btnVDialerCallSelectSelectList = [
    Button(TLPPro1020T, 191),   # Call 1
    Button(TLPPro1020T, 192),   # Call 2
    Button(TLPPro1020T, 193),   # Call 3 
    Button(TLPPro1020T, 194),   # Call 4 
    Button(TLPPro1020T, 195)    # Call 5
]

btnVDialerCallSelectStateList = [
    Button(TLPPro1020T, 211),   # Call 1
    Button(TLPPro1020T, 212),   # Call 2
    Button(TLPPro1020T, 213),   # Call 3 
    Button(TLPPro1020T, 214),   # Call 4 
    Button(TLPPro1020T, 215)    # Call 5
]

txtVDialerCallSelectStateList = [
    Label(TLPPro1020T, 201),    # Call 1
    Label(TLPPro1020T, 202),    # Call 2
    Label(TLPPro1020T, 203),    # Call 3 
    Label(TLPPro1020T, 204),    # Call 4 
    Label(TLPPro1020T, 205)     # Call 5
]

btnVDialerCallSelectConfList = [
    Button(TLPPro1020T, 221),   # Call 1
    Button(TLPPro1020T, 222),   # Call 2
    Button(TLPPro1020T, 223),   # Call 3 
    Button(TLPPro1020T, 224),   # Call 4 
    Button(TLPPro1020T, 225)    # Call 5
]

btnVDialerPrivacy = Button(TLPPro1020T, 186)
btnVDialerRxVolUp   = Button(TLPPro1020T, 244)
btnVDialerRxVolDn   = Button(TLPPro1020T, 245)
btnVDialerRxVolMute = Button(TLPPro1020T, 246)
lvlVDialerRx        = Level(TLPPro1020T, 247)

lvlVDialerRx.SetRange(0, 100)

txtVDialerIncomingName = Label(TLPPro1020T, 226)
btnVDialerIncomingAnswer = Button(TLPPro1020T, 228)
btnVDialerIncomingIgnore = Button(TLPPro1020T, 229)

#-------------------------------------------------------------------------------
# Room Combiner
#-------------------------------------------------------------------------------

btnCombinerCombineList = [
    Button(TLPPro1020T, 251),   # Room 1/2 (Wall 1)
    Button(TLPPro1020T, 252),   # Room 2/3 (Wall 2)
    Button(TLPPro1020T, 253)    # Room 3/4 (Wall 3)
]

lvlCombinerStateList = [
    Level(TLPPro1020T, 261),    # Room 1
    Level(TLPPro1020T, 262),    # Room 2
    Level(TLPPro1020T, 263),    # Room 3
    Level(TLPPro1020T, 264)     # Room 4
]

txtCombinerStateList = [
    Label(TLPPro1020T, 271),    # Room 1
    Label(TLPPro1020T, 272),    # Room 2
    Label(TLPPro1020T, 273),    # Room 3
    Label(TLPPro1020T, 274)     # Room 4
]

lvlCombinerStateList[0].SetRange(0, 2)
lvlCombinerStateList[1].SetRange(0, 2)
lvlCombinerStateList[2].SetRange(0, 2)
lvlCombinerStateList[3].SetRange(0, 2)

################################################################################
## Event Definitions
################################################################################

#-------------------------------------------------------------------------------
# Device Connection
#-------------------------------------------------------------------------------

if MyCommType == CommTypes['Ethernet']:
    @event(MyModule, 'Connected')
    @event(MyModule, 'Disconnected')
    def OnConnectionChange(interface, state):
        print(str(state))
        if state == 'Disconnected':
            MyModule.Reinit()

#-------------------------------------------------------------------------------
# Communicating/Initialization
#-------------------------------------------------------------------------------

@event(MyModule, 'ModuleInfo')
def ModuleInfoEvent(interface, communicating, initialized):
    btnCommunicating.SetState(communicating)
    btnInitialized.SetState(initialized)    
        
#-------------------------------------------------------------------------------
# Menu
#-------------------------------------------------------------------------------

@event(btnMenuList, BUTTON_STATES)
def MenuSelect(button, state):
    if state == 'Pressed':
        index = btnMenuList.index(button)
        btnMenuListME.SetCurrent(index)
        TLPPro1020T.ShowPopup(puMenuNames[index])

#-------------------------------------------------------------------------------
# Presets/Macros
#-------------------------------------------------------------------------------

@event(btnPresetList, BUTTON_STATES)
def PresetSelect(button, state):
    index = btnPresetList.index(button)
      
    if index < len(txtPresetNames):
        if state == 'Pressed':
            MyPresets.RecallPreset(txtPresetNames[index])

@event(btnMacroList, BUTTON_STATES)
def MacroSelect(button, state):
    index = btnMacroList.index(button)
      
    if index < len(txtMacroNames):
        if state == 'Pressed':
            MyMacros.RunMacro(txtMacroNames[index])
      
#-------------------------------------------------------------------------------
# Gain/Mute
#-------------------------------------------------------------------------------

@event(btnVolUpList, BUTTON_STATES)
def VolUpSelect(button, state):
    index = btnVolUpList.index(button)
       
    if index < len(myVolumes):
        if state == 'Pressed':
            myVolumes[index].Raise()
        
        elif state == 'Released':
            myVolumes[index].Stop()
            
@event(btnVolDnList, BUTTON_STATES)
def VolDnSelect(button, state):
    index = btnVolDnList.index(button)
    
    if index < len(myVolumes):
        if state == 'Pressed':
            myVolumes[index].Lower()
        
        elif state == 'Released':
            myVolumes[index].Stop()              

@event(myVolumes, 'OnLevelPercentChange')
def LevelPercentChangeEvent(interface, value):
    index = myVolumes.index(interface)
    
    if index < len(myVolumes):   
        lvlVolList[index].SetLevel(value)

@event(btnVolMuteList, BUTTON_STATES)
def VolMuteSelect(button, state):
    index = btnVolMuteList.index(button)
    
    if index < len(myVolumes):
        if state == 'Pressed':
            myVolumes[index].ToggleMute()

@event(myVolumes, 'OnMuteStateChange')
def MuteChangeEvent(interface, value):
    index = myVolumes.index(interface)
    
    if index < len(myVolumes):   
        btnVolMuteList[index].SetState(value) 
    
#-------------------------------------------------------------------------------
# Matrix Mixer
#-------------------------------------------------------------------------------

def GetCurrentSelection():
    btn = btnMixerStateListME.GetCurrent()
    
    return -1 if btn is None else btnMixerStateList.index(btn)

@event(btnMixerSelectList, BUTTON_STATES)
def MixerBtnSelect(button, state):
    index = btnMixerSelectList.index(button)
       
    if index < len(myCrosspoints):
        if state == 'Pressed':
            current = GetCurrentSelection()
            
            if current == index:
                btnMixerStateListME.SetCurrent(None)
                lvlMixerXPoint.SetLevel(lvlMixerXPoint.Min)
            else:
                btnMixerStateListME.SetCurrent(index)
                lvlMixerXPoint.SetLevel(myCrosspoints[index].GetAttenuationPercent())                    

@event(btnMixerXPointVolUp, BUTTON_STATES)
def MixerXPointVolUpSelect(button, state):
    index = GetCurrentSelection()
    
    if index >= 0:
        if state == 'Pressed':
            myCrosspoints[index].Raise()
        
        elif state == 'Released':
            myCrosspoints[index].Stop()
            
@event(btnMixerXPointVolDn, BUTTON_STATES)
def MixerXPointVolDnSelect(button, state):
    index = GetCurrentSelection()
    
    if index >= 0:
        if state == 'Pressed':
            myCrosspoints[index].Lower()
        
        elif state == 'Released':
            myCrosspoints[index].Stop()

@event(btnMixerXPointRoute, BUTTON_STATES)
def MixerXPointRouteSelect(button, state):
    index = GetCurrentSelection()
    
    if index >= 0:
        if state == 'Pressed':
            myCrosspoints[index].ToggleCrosspoint()

@event(myCrosspoints, 'OnAttenuationDBChange')
def AttenuationDBChangeEvent(interface, value):
    index = myCrosspoints.index(interface)
    
    if index < len(myCrosspoints):
        btnMixerSelectList[index].SetText('{0} db'.format(int(value)))

@event(myCrosspoints, 'OnAttenuationPercentChange')
def AttenuationPercentChangeEvent(interface, value):
    index = myCrosspoints.index(interface)
    current = GetCurrentSelection()
    
    if index == current:   
        lvlMixerXPoint.SetLevel(value)

@event(myCrosspoints, 'OnCrosspointChange')
def CrosspointChangeEvent(interface, value):
    index = myCrosspoints.index(interface)
    
    if index < len(myCrosspoints):
        btnMixerSelectList[index].SetState(value)

#-------------------------------------------------------------------------------
# Analog Dialer
#-------------------------------------------------------------------------------

@event(btnADialerDialpadList, BUTTON_STATES)
def ADialerDialpadSelect(button, state):
    index = btnADialerDialpadList.index(button)
    
    if index <= len(btnADialerDialpadDigitList):
        if state == 'Pressed':      
            MyADialer.DigitPressed(btnADialerDialpadDigitList[index])
            
        elif state == 'Released':
            MyADialer.DigitReleased(btnADialerDialpadDigitList[index])

@event(btnADialerClear, BUTTON_STATES)
def ADialerClearSelect(button, state):
    if state == 'Pressed':
        MyADialer.ClearKeypadText()

@event(btnADialerBack, BUTTON_STATES)
def ADialerBackSelect(button, state):
    if state == 'Pressed':
        MyADialer.BackSpaceKeypadText()

@event(btnADialerAutoAnswer, BUTTON_STATES)
def ADialerAutoAnswerSelect(button, state):
    current = MyADialer.GetAutoAnswerRingCount()
    
    if state == 'Pressed':
        if current == 0:
            MyADialer.SetAutoAnswerRingCount(1)
        
        else:
            MyADialer.SetAutoAnswerRingCount(0)

@event(btnADialerDial, BUTTON_STATES)
def ADialerDialSelect(button, state):
    if state == 'Pressed':
        MyADialer.Dial()

@event(btnADialerPickup, BUTTON_STATES)
def ADialerPickupSelect(button, state):
    if state == 'Pressed':
        MyADialer.OffHook()

@event(btnADialerHangup, BUTTON_STATES)
def ADialerhangupSelect(button, state):
    if state == 'Pressed':
        MyADialer.OnHook()

@event(btnADialerFlash, BUTTON_STATES)
def ADialerFlashSelect(button, state):
    if state == 'Pressed':
        MyADialer.HookFlash()

@event(btnADialerRedial, BUTTON_STATES)
def ADialerRedialSelect(button, state):
    if state == 'Pressed':
        MyADialer.Redial()

@event(btnADialerIncomingAnswer, BUTTON_STATES)
def ADialerIncomingAnswerSelect(button, state):
    if state == 'Pressed':
        TLPPro1020T.HidePopup('Incoming Call')
        MyADialer.OffHook()

@event(btnADialerIncomingIgnore, BUTTON_STATES)
def ADialerIncomingIgnoreSelect(button, state):
    if state == 'Pressed':
        TLPPro1020T.HidePopup('Incoming Call')
        MyADialer.OffHook()
        
        @Wait(2)
        def OnHook():
            MyADialer.OnHook()
        
@event(btnADialerRxVolUp, BUTTON_STATES)
def ADialerRxVolUpSelect(button, state):
    if state == 'Pressed':
        MyADialerRx.Raise()
    
    elif state == 'Released':
        MyADialerRx.Stop()

@event(btnADialerRxVolDn, BUTTON_STATES)
def ADialerRxVolDnSelect(button, state):
    if state == 'Pressed':
        MyADialerRx.Lower()

    elif state == 'Released':
        MyADialerRx.Stop()

@event(MyADialerRx, 'OnLevelPercentChange')
def AnalogLevelPercentChangeEvent(interface, value):
    lvlADialerRx.SetLevel(value)
    
@event(btnADialerRxVolMute, BUTTON_STATES)
def ADialerRxVolMuteSelect(button, state):
    if state == 'Pressed':
        MyADialerRx.ToggleMute()

@event(MyADialerRx, 'OnMuteStateChange')
def AnalogMuteChangeEvent(interface, value):
    btnADialerRxVolMute.SetState(value) 

@event(btnADialerPrivacy, BUTTON_STATES)
def ADialerPrivacySelect(button, state):
    if state == 'Pressed':
        MyADialerPrivacy.ToggleDigital()

@event(MyADialerPrivacy, 'OnEPDigitalStateChange')
def AnalogPrivacyChangeEvent(interface, value):
    btnADialerPrivacy.SetState(value)

@event(MyADialer, 'OnAnalogKeypadTextChange')
def AnalogKeypadTextChangeEvent(interface, value):
    txtADialerDialpad.SetText(value)

@event(MyADialer, 'OnAnalogAutoAnswerChange')
def AnalogAutoAnswerChangeEvent(interface, value):
    if value > 0:
        btnADialerAutoAnswer.SetState(1)
    
    else:
        btnADialerAutoAnswer.SetState(0)

@event(MyADialer, 'OnAnalogCallerIDNotification')
def AnalogCallerIDNotificationEvent(interface, value):
    txtADialerIncomingName.SetText(str(value.Name))
    txtADialerIncomingNumber.SetText(str(value.PhoneNumber))

@event(MyADialer, 'OnAnalogHookChange')
def AnalogHookChangeEvent(interface, value):
        btnADialerHangup.SetState(not value)
        btnADialerDial.SetState(value)
    
@event(MyADialer, 'OnAnalogIncomingRingChange')
def AnalogIncomingRingChangeEvent(interface, value):
    if MyADialer.GetAutoAnswerRingCount() == 0:
        if value == 1:
            TLPPro1020T.ShowPopup('Incoming Call')
 
#-------------------------------------------------------------------------------
# VoIP Dialer
#-------------------------------------------------------------------------------

@event(btnVDialerDialpadList, BUTTON_STATES)
def VDialerDialpadSelect(button, state):
    index = btnVDialerDialpadList.index(button)
    
    if index <= len(btnVDialerDialpadDigitList):
        if state == 'Pressed':      
            MyVDialer.DigitPressed(btnVDialerDialpadDigitList[index])
            
        elif state == 'Released':
            MyVDialer.DigitReleased(btnVDialerDialpadDigitList[index])

@event(btnVDialerClear, BUTTON_STATES)
def VDialerClearSelect(button, state):
    if state == 'Pressed':
        MyVDialer.ClearKeypadText()

@event(btnVDialerBack, BUTTON_STATES)
def VDialerBackSelect(button, state):
    if state == 'Pressed':
        MyVDialer.BackSpaceKeypadText()

@event(btnVDialerDial, BUTTON_STATES)
def VDialerDialSelect(button, state):
    if state == 'Pressed':
        MyVDialer.Dial()

@event(btnVDialerCallSelectSelectList, BUTTON_STATES)
def VDialerCallBtnSelect(button, state):
    if state == 'Pressed':
        index = btnVDialerCallSelectSelectList.index(button) + 1
        
        MyVDialer.SelectLine(index)

@event(btnVDialerCallSelectConfList, BUTTON_STATES)
def VDialerCallConfSelect(button, state):
    if state == 'Pressed':
        index = btnVDialerCallSelectConfList.index(button) + 1
        
        MyVDialer.Conference(index)

@event(btnVDialerOnHook, BUTTON_STATES)
def VDialerOnHookSelect(button, state):
    if state == 'Pressed':
        MyVDialer.OnHook()

@event(btnVDialerOffHook, BUTTON_STATES)
def VDialerOffHookSelect(button, state):
    if state == 'Pressed':
        MyVDialer.OffHook()

@event(btnVDialerTransfer, BUTTON_STATES)
def VDialerTransferSelect(button, state):
    if state == 'Pressed':
        MyVDialer.Transfer()

@event(btnVDialerBlindTransfer, BUTTON_STATES)
def VDialerBlindTransferSelect(button, state):
    if state == 'Pressed':
        MyVDialer.BlindTransfer()

@event(btnVDialerRedial, BUTTON_STATES)
def VDialerRedialSelect(button, state):
    if state == 'Pressed':
        MyVDialer.Redial()

@event(btnVDialerHold, BUTTON_STATES)
def VDialerHoldSelect(button, state):
    if state == 'Pressed':
        MyVDialer.Hold()

@event(btnVDialerFwdOn, BUTTON_STATES)
def VDialerFwdOnSelect(button, state):
    if state == 'Pressed':
        MyVDialer.CallForwardEnable()

@event(btnVDialerFwdOff, BUTTON_STATES)
def VDialerFwdOffSelect(button, state):
    if state == 'Pressed':
        MyVDialer.CallForwardDisable()

@event(btnVDialerDndOn, BUTTON_STATES)
def VDialerDndOnSelect(button, state):
    if state == 'Pressed':
        MyVDialer.DNDCallReject()

@event(btnVDialerDndOff, BUTTON_STATES)
def VDialerDndOffSelect(button, state):
    if state == 'Pressed':
        MyVDialer.DNDDisable()
        
@event(btnVDialerIncomingAnswer, BUTTON_STATES)
def VDialerIncomingAnswerSelect(button, state):
    if state == 'Pressed':
        MyVDialer.OffHook()

@event(btnVDialerIncomingIgnore, BUTTON_STATES)
def VDialerIncomingIgnoreSelect(button, state):
    if state == 'Pressed':
        current = MyVDialer.GetActiveLine()
        
        if current > 0 and current <= 5:
            MyVDialer.Reject(current)
        
@event(btnVDialerRxVolUp, BUTTON_STATES)
def VDialerRxVolUpSelect(button, state):
    if state == 'Pressed':
        MyVDialerRx.Raise()
    
    elif state == 'Released':
        MyVDialerRx.Stop()

@event(btnVDialerRxVolDn, BUTTON_STATES)
def VDialerRxVolDnSelect(button, state):
    if state == 'Pressed':
        MyVDialerRx.Lower()

    elif state == 'Released':
        MyVDialerRx.Stop()

@event(MyVDialerRx, 'OnLevelPercentChange')
def VoIPLevelPercentChangeEvent(interface, value):
    lvlVDialerRx.SetLevel(value)
    
@event(btnVDialerRxVolMute, BUTTON_STATES)
def VDialerRxVolMuteSelect(button, state):
    if state == 'Pressed':
        MyVDialerRx.ToggleMute()

@event(MyVDialerRx, 'OnMuteStateChange')
def VoIPMuteChangeEvent(interface, value):
    btnVDialerRxVolMute.SetState(value) 

@event(btnVDialerPrivacy, BUTTON_STATES)
def VDialerPrivacySelect(button, state):
    if state == 'Pressed':
        MyVDialerPrivacy.ToggleDigital()

@event(MyVDialerPrivacy, 'OnEPDigitalStateChange')
def VoIPPrivacyChangeEvent(interface, value):
    btnVDialerPrivacy.SetState(value)

@event(MyVDialer, 'OnVoIPKeypadTextChange')
def VoIPKeypadTextChangeEvent(interface, value):
    txtVDialerDialpad.SetText(value)

def ProcessCallForwardState(state):
    if state == MyClearOne.CallForwardWaitingStates['INACTIVE']:
        btnVDialerFwdOn.SetState(0)
        btnVDialerFwdOff.SetState(1)
    
    elif state == MyClearOne.CallForwardWaitingStates['INPROCESS']:
        btnVDialerFwdOn.SetBlinking('Medium', [0,1])
        btnVDialerFwdOff.SetState(0)
    
    elif state == MyClearOne.CallForwardWaitingStates['ACTIVE']:
        btnVDialerFwdOn.SetState(1)
        btnVDialerFwdOff.SetState(0)    

def ProcessDNDState(state):
    if state == MyClearOne.DNDStatus['DND_NOT_SET']:
        btnVDialerDndOn.SetState(0)
        btnVDialerDndOff.SetState(1)
    
    else:
        btnVDialerDndOn.SetState(1)
        btnVDialerDndOff.SetState(0)

def ProcessLineState(line):
    index = line.LineID
    
    if index < 5:
        btnVDialerCallSelectStateList[index].SetState(1 if line.LineLEDState > 0 else 0)
        txtVDialerCallSelectStateList[index].SetText(line.LineStatusText)
        
        if line.LineStatus == MyClearOne.StateChangeTypes['INCOMING']:
            txtVDialerIncomingName.SetText(line.LineInfo)
            
@event(MyVDialer, 'OnVoIPStatusChange')
def VoIPStatusChangeEvent(interface, status):
    ProcessCallForwardState(status.CallForwardState)
    ProcessDNDState(status.DNDStatus)
     
    for i in status.LineStatus:
        ProcessLineState(i)
    
    if status.RingingStatus == 1:
        TLPPro1020T.ShowPopup('Incoming Call VoIP')
    
    else:
        TLPPro1020T.HidePopup('Incoming Call VoIP')

#-------------------------------------------------------------------------------
# Room Combiner
#-------------------------------------------------------------------------------

@event(btnCombinerCombineList, BUTTON_STATES)
def RoomCombineSelect1(button, state):
    index = btnCombinerCombineList.index(button)
      
    if index < len(myWalls):
        if state == 'Pressed':
            myWalls[index].ToggleDigital()
            
@event(myWalls, 'OnRoomDigitalStateChange')
def RoomDigitalChangeEvent(interface, value):
    index = myWalls.index(interface)
    
    if index < len(myWalls):   
        btnCombinerCombineList[index].SetState(not value)
        SetRoomCombineFB()

def SetRoomCombineFB():
    wall1 = myWalls[0].GetDigital()
    wall2 = myWalls[1].GetDigital()
    wall3 = myWalls[2].GetDigital()
    
    if wall1 and wall2 and wall3:
        lvlCombinerStateList[0].SetLevel(0)
        lvlCombinerStateList[1].SetLevel(0)
        lvlCombinerStateList[2].SetLevel(0)
        lvlCombinerStateList[3].SetLevel(0)
        txtCombinerStateList[0].SetText('Room 1')
        txtCombinerStateList[1].SetText('Room 2')
        txtCombinerStateList[2].SetText('Room 3')
        txtCombinerStateList[3].SetText('Room 4')
        
    elif not wall1 and wall2 and wall3:
        lvlCombinerStateList[0].SetLevel(1)
        lvlCombinerStateList[1].SetLevel(1)
        lvlCombinerStateList[2].SetLevel(0)
        lvlCombinerStateList[3].SetLevel(0)
        txtCombinerStateList[0].SetText('Room 1')
        txtCombinerStateList[1].SetText('Room 1')
        txtCombinerStateList[2].SetText('Room 2')
        txtCombinerStateList[3].SetText('Room 3')
        
    elif not wall1 and not wall2 and wall3:
        lvlCombinerStateList[0].SetLevel(1)
        lvlCombinerStateList[1].SetLevel(1)
        lvlCombinerStateList[2].SetLevel(1)
        lvlCombinerStateList[3].SetLevel(0)
        txtCombinerStateList[0].SetText('Room 1')
        txtCombinerStateList[1].SetText('Room 1')
        txtCombinerStateList[2].SetText('Room 1')
        txtCombinerStateList[3].SetText('Room 2')
        
    elif wall1 and wall2 and not wall3:
        lvlCombinerStateList[0].SetLevel(0)
        lvlCombinerStateList[1].SetLevel(0)
        lvlCombinerStateList[2].SetLevel(1)
        lvlCombinerStateList[3].SetLevel(1)
        txtCombinerStateList[0].SetText('Room 1')
        txtCombinerStateList[1].SetText('Room 2')
        txtCombinerStateList[2].SetText('Room 3')
        txtCombinerStateList[3].SetText('Room 3')
        
    elif wall1 and not wall2 and not wall3:
        lvlCombinerStateList[0].SetLevel(0)
        lvlCombinerStateList[1].SetLevel(1)
        lvlCombinerStateList[2].SetLevel(1)
        lvlCombinerStateList[3].SetLevel(1)
        txtCombinerStateList[0].SetText('Room 1')
        txtCombinerStateList[1].SetText('Room 2')
        txtCombinerStateList[2].SetText('Room 2')
        txtCombinerStateList[3].SetText('Room 2')
        
    elif not wall1 and wall2 and not wall3:
        lvlCombinerStateList[0].SetLevel(1)
        lvlCombinerStateList[1].SetLevel(1)
        lvlCombinerStateList[2].SetLevel(2)
        lvlCombinerStateList[3].SetLevel(2)
        txtCombinerStateList[0].SetText('Room 1')
        txtCombinerStateList[1].SetText('Room 1')
        txtCombinerStateList[2].SetText('Room 2')
        txtCombinerStateList[3].SetText('Room 2')
        
    elif wall1 and not wall2 and wall3:
        lvlCombinerStateList[0].SetLevel(0)
        lvlCombinerStateList[1].SetLevel(1)
        lvlCombinerStateList[2].SetLevel(1)
        lvlCombinerStateList[3].SetLevel(0)
        txtCombinerStateList[0].SetText('Room 1')
        txtCombinerStateList[1].SetText('Room 2')
        txtCombinerStateList[2].SetText('Room 2')
        txtCombinerStateList[3].SetText('Room 3')
        
    elif not wall1 and not wall2 and not wall3:
        lvlCombinerStateList[0].SetLevel(1)
        lvlCombinerStateList[1].SetLevel(1)
        lvlCombinerStateList[2].SetLevel(1)
        lvlCombinerStateList[3].SetLevel(1)
        txtCombinerStateList[0].SetText('Room 1')
        txtCombinerStateList[1].SetText('Room 1')
        txtCombinerStateList[2].SetText('Room 1')
        txtCombinerStateList[3].SetText('Room 1')
 
#-------------------------------------------------------------------------------
# Program Initialize
#-------------------------------------------------------------------------------

def Initialize():
    TLPPro1020T.HideAllPopups()
    btnMenuListME.SetCurrent(0)
    TLPPro1020T.ShowPopup(puMenuNames[0])
    
    MyModule.SetDebug(False)
    MyModule.SetPoll(False)
    MyModule.Reinit()

################################################################################
## Main
################################################################################

Initialize()