################################################################################
## ControlScript Imports
################################################################################

from extronlib import event, Version
from extronlib.device import eBUSDevice, ProcessorDevice, UIDevice
from extronlib.interface import (ContactInterface, DigitalIOInterface,
    EthernetClientInterface, EthernetServerInterfaceEx, FlexIOInterface,
    IRInterface, RelayInterface, SerialInterface, SWPowerInterface,
    VolumeInterface)
from extronlib.ui import Button, Knob, Label, Level
from extronlib.system import Clock, MESet, Wait

import re

print(Version())

################################################################################
## User Imports
################################################################################

import Clearone_Converge_Pro_2_v1_1_0 as MyClearOne

################################################################################
## Device/Processor Definition
################################################################################

IPCPPro350  = ProcessorDevice('IPCPPro350')

################################################################################
## Device/User Interface Definition
################################################################################

TLPPro1020T = UIDevice('TLPPro1020T')

################################################################################
## Module Setup
################################################################################

CommTypes = {
    'Serial': 0,
    'Ethernet': 1
}

MyCommType = CommTypes['Ethernet']

#-------------------------------------------------------------------------------
# Primary Comm Module
#-------------------------------------------------------------------------------

if MyCommType == CommTypes['Ethernet']:
    MyModule = MyClearOne.EthernetClass(Hostname='192.168.187.53', IPPort=23)
    MyModule.SetCredentials(Username='clearone', Password='converge')

elif MyCommType == CommTypes['Serial']:
    MyModule = MyClearOne.SerialClass(Host=IPCPPro350, Port='COM1', Baud=57600, Data=8, Parity='None', Stop=1, FlowControl='Off', CharDelay=0, Mode='RS232')
   
#-------------------------------------------------------------------------------
# Components
#-------------------------------------------------------------------------------

MySkype = MyModule.AddSkypeComponent('Skype_Name_2_01')

################################################################################
## Constants / Global
################################################################################

#-------------------------------------------------------------------------------
# General
#-------------------------------------------------------------------------------

BUTTON_STATES = ['Pressed', 'Released', 'Held', 'Repeated']

#-------------------------------------------------------------------------------
# Contacts
#-------------------------------------------------------------------------------

CONTACTS_LIST_SIZE  = 5
ContactsFull        = []
ContactsCurrent     = []
ContactsCurrentPage = 1
ContactsTotalPages  = 0
ContactsSearchText  = ''

#-------------------------------------------------------------------------------
# Meetings
#-------------------------------------------------------------------------------

MEETINGS_LIST_SIZE  = 7
MeetingsFull        = []
MeetingsCurrent     = []
MeetingsCurrentPage = 1
MeetingsTotalPages  = 0
MeetingsJoinText    = ''

#-------------------------------------------------------------------------------
# Keyboard
#-------------------------------------------------------------------------------

KeyboardType = {
    'NONE': 0,
    'CONTACTS': 1,
    'MEETINGS': 2
}

CurrentKeyboard = KeyboardType['NONE']

################################################################################
## Buttons/Labels
################################################################################

#-------------------------------------------------------------------------------
# Communication/Initialization/Registration
#-------------------------------------------------------------------------------

btnCommunicating = Button(TLPPro1020T, 1)
btnInitialized   = Button(TLPPro1020T, 2)
btnRegistered    = Button(TLPPro1020T, 3)
btnReinit        = Button(TLPPro1020T, 4)

#-------------------------------------------------------------------------------
# Menu
#-------------------------------------------------------------------------------

btnMenuList = [
    Button(TLPPro1020T, 11),    # Dialer
    Button(TLPPro1020T, 12),    # Contacts
    Button(TLPPro1020T, 13)     # Meeting Control
]

btnMenuListME = MESet(btnMenuList)

puMenuNames = [
    'Dialer',
    'Contacts',
    'Meeting Control' 
]

#-------------------------------------------------------------------------------
# Incoming Call
#-------------------------------------------------------------------------------

puIncomingName = 'Incoming Call'

btnIncomingAnswer    = Button(TLPPro1020T, 21)
btnIncomingIgnore    = Button(TLPPro1020T, 22)
txtIncomingMeetingID = Label(TLPPro1020T, 23)
txtIncomingURL       = Label(TLPPro1020T, 24)

#-------------------------------------------------------------------------------
# Dialer
#-------------------------------------------------------------------------------

btnDialerDialpad = [
    Button(TLPPro1020T, 31),   # 1
    Button(TLPPro1020T, 32),   # 2
    Button(TLPPro1020T, 33),   # 3
    Button(TLPPro1020T, 34),   # 4
    Button(TLPPro1020T, 35),   # 5
    Button(TLPPro1020T, 36),   # 6
    Button(TLPPro1020T, 37),   # 7
    Button(TLPPro1020T, 38),   # 8
    Button(TLPPro1020T, 39),   # 9
    Button(TLPPro1020T, 40),   # 0
    Button(TLPPro1020T, 41),   # Star
    Button(TLPPro1020T, 42),   # Pound
    Button(TLPPro1020T, 43)    # Plus
]

btnDialerDialpadMap = [
    '1',
    '2',
    '3',
    '4',
    '5',
    '6',
    '7',
    '8',
    '9',
    '0',
    '*',
    '#',
    '+'    
]

btnDialerClear = Button(TLPPro1020T, 44)
btnDialerBack  = Button(TLPPro1020T, 45)
btnDialerDial  = Button(TLPPro1020T, 46)
txtDialString  = Label(TLPPro1020T, 47)

#-------------------------------------------------------------------------------
# Contacts
#-------------------------------------------------------------------------------

btnContactsList = [
    Button(TLPPro1020T, 61),   # Contact 1
    Button(TLPPro1020T, 62),   # Contact 2
    Button(TLPPro1020T, 63),   # Contact 3
    Button(TLPPro1020T, 64),   # Contact 4
    Button(TLPPro1020T, 65)    # Contact 5
]

btnContactsListTop    = Button(TLPPro1020T, 66)
btnContactsListUp     = Button(TLPPro1020T, 67)
btnContactsListDown   = Button(TLPPro1020T, 68)
btnContactsListBottom = Button(TLPPro1020T, 69)
btnContactsListHome   = Button(TLPPro1020T, 70)
btnContactsSearch     = Button(TLPPro1020T, 71)
btnContactsString     = Button(TLPPro1020T, 72)

btnContactsGroupCreateSession = Button(TLPPro1020T, 81)
btnContactsGroupDeleteGroup   = Button(TLPPro1020T, 82)
btnContactsGroupName          = Button(TLPPro1020T, 83)
btnContactsGroupOther         = Button(TLPPro1020T, 84)

btnContactsGroupVisibility = [
    btnContactsGroupCreateSession,
    btnContactsGroupDeleteGroup,
    btnContactsGroupName,
    btnContactsGroupOther
]

btnContactsContactDialContact         = Button(TLPPro1020T, 91)
btnContactsContactAddToGroup          = Button(TLPPro1020T, 92)
btnContactsContactRemoveFromGroup     = Button(TLPPro1020T, 93)
btnContactsContactRemoveFromAllGroups = Button(TLPPro1020T, 94)
btnContactsContactAddToActiveSession  = Button(TLPPro1020T, 95)
btnContactsContactCreateSession       = Button(TLPPro1020T, 96)
btnContactsContactName                = Button(TLPPro1020T, 97)
btnContactsContactGroup               = Button(TLPPro1020T, 98)
btnContactsContactOther               = Button(TLPPro1020T, 99)

btnContactsContactVisibility = [
    btnContactsContactDialContact,
    btnContactsContactAddToGroup,
    btnContactsContactRemoveFromGroup,
    btnContactsContactRemoveFromAllGroups,
    btnContactsContactAddToActiveSession,
    btnContactsContactCreateSession,
    btnContactsContactName,
    btnContactsContactGroup,
    btnContactsContactOther
]

#-------------------------------------------------------------------------------
# Meetings
#-------------------------------------------------------------------------------

btnMeetingList = [
    Button(TLPPro1020T, 111),   # Meeting 1
    Button(TLPPro1020T, 112),   # Meeting 2
    Button(TLPPro1020T, 113),   # Meeting 3
    Button(TLPPro1020T, 114),   # Meeting 4
    Button(TLPPro1020T, 115),   # Meeting 5
    Button(TLPPro1020T, 116),   # Meeting 6
    Button(TLPPro1020T, 117)    # Meeting 7
]

btnMeetingListTop    = Button(TLPPro1020T, 118)
btnMeetingListUp     = Button(TLPPro1020T, 119)
btnMeetingListDown   = Button(TLPPro1020T, 120)
btnMeetingListBottom = Button(TLPPro1020T, 121)

btnMeetingActiveAccept = Button(TLPPro1020T, 131)
btnMeetingActiveLeave  = Button(TLPPro1020T, 132)
btnMeetingActiveResume = Button(TLPPro1020T, 133)
btnMeetingActiveHold   = Button(TLPPro1020T, 134)
btnMeetingActiveName   = Button(TLPPro1020T, 135)
btnMeetingActiveID     = Button(TLPPro1020T, 136)
btnMeetingActiveState  = Button(TLPPro1020T, 137)

btnMeetingCreateJoinMeeting    = Button(TLPPro1020T, 141)
btnMeetingCreateJoinConference = Button(TLPPro1020T, 142)
btnMeetingCreateJoinNow        = Button(TLPPro1020T, 143)
btnMeetingCreateJoinID         = Button(TLPPro1020T, 144)

#-------------------------------------------------------------------------------
# Keyboard
#-------------------------------------------------------------------------------

puKeyboard = 'Keyboard'

btnKeyboard = [
    Button(TLPPro1020T, 201),   # 1
    Button(TLPPro1020T, 202),   # 2
    Button(TLPPro1020T, 203),   # 3
    Button(TLPPro1020T, 204),   # 4
    Button(TLPPro1020T, 205),   # 5
    Button(TLPPro1020T, 206),   # 6
    Button(TLPPro1020T, 207),   # 7
    Button(TLPPro1020T, 208),   # 8
    Button(TLPPro1020T, 209),   # 9
    Button(TLPPro1020T, 210),   # 0
    Button(TLPPro1020T, 211),   # Q
    Button(TLPPro1020T, 212),   # W
    Button(TLPPro1020T, 213),   # E
    Button(TLPPro1020T, 214),   # R
    Button(TLPPro1020T, 215),   # T
    Button(TLPPro1020T, 216),   # Y
    Button(TLPPro1020T, 217),   # U
    Button(TLPPro1020T, 218),   # I
    Button(TLPPro1020T, 219),   # O
    Button(TLPPro1020T, 220),   # P
    Button(TLPPro1020T, 221),   # A
    Button(TLPPro1020T, 222),   # S
    Button(TLPPro1020T, 223),   # D
    Button(TLPPro1020T, 224),   # F
    Button(TLPPro1020T, 225),   # G
    Button(TLPPro1020T, 226),   # H
    Button(TLPPro1020T, 227),   # J
    Button(TLPPro1020T, 228),   # K
    Button(TLPPro1020T, 229),   # L
    Button(TLPPro1020T, 230),   # Z
    Button(TLPPro1020T, 231),   # X
    Button(TLPPro1020T, 232),   # C
    Button(TLPPro1020T, 233),   # V
    Button(TLPPro1020T, 234),   # B
    Button(TLPPro1020T, 235),   # N
    Button(TLPPro1020T, 236),   # M
    Button(TLPPro1020T, 237),   # Space
    Button(TLPPro1020T, 241),   # :
    Button(TLPPro1020T, 242),   # /
    Button(TLPPro1020T, 243),   # @
    Button(TLPPro1020T, 244),   # .
    Button(TLPPro1020T, 245),   # #
    Button(TLPPro1020T, 246),   # %
    Button(TLPPro1020T, 247),   # &
    Button(TLPPro1020T, 248),   # ?
    Button(TLPPro1020T, 249),   # ;
    Button(TLPPro1020T, 250)    # =
]

btnKeyboardMap = [
    '1',
    '2',
    '3',
    '4',
    '5',
    '6',
    '7',
    '8',
    '9',
    '0',
    'Q',
    'W',
    'E',
    'R',
    'T',
    'Y',
    'U',
    'I',
    'O',
    'P',
    'A',
    'S',
    'D',
    'F',
    'G',
    'H',
    'J',
    'K',
    'L',
    'Z',
    'X',
    'C',
    'V',
    'B',
    'N',
    'M',
    ' ',
    ':',
    '/',
    '@',
    '.',
    '#',
    '%',
    '&',
    '?',
    ';',
    '='
]

btnKeyboardBack  = Button(TLPPro1020T, 238)
btnKeyboardClear = Button(TLPPro1020T, 239)
btnKeyboardClose = Button(TLPPro1020T, 240)

################################################################################
## Helper Functions
################################################################################

#-------------------------------------------------------------------------------
# Incoming Call
#-------------------------------------------------------------------------------

def ProcessCallActiveSessionChange(info):
    url = info.Url.replace('sip:+', '').replace('sip:', '').replace(';user=phone', '')
    
    if info.Status == MyClearOne.S4BSessionStatus['INCOMING']:
        TLPPro1020T.ShowPopup(puIncomingName)
        txtIncomingMeetingID.SetText(info.MeetingID.title())
        txtIncomingURL.SetText(url)
    else:
        TLPPro1020T.HidePopup(puIncomingName)
        txtIncomingMeetingID.SetText('')
        txtIncomingURL.SetText('')

#-------------------------------------------------------------------------------
# Contacts
#-------------------------------------------------------------------------------

def SetGroupItemVisibility(state):
    for btn in btnContactsGroupVisibility:
        btn.SetVisible(state)


def SetContactItemVisibility(state):
    for btn in btnContactsContactVisibility:
        btn.SetVisible(state)


def SetContactList(page):
    global ContactsFull, ContactsCurrent, ContactsCurrentPage, ContactsTotalPages
    
    if ContactsTotalPages > 0 and page > 0 and page <= ContactsTotalPages:
        ContactsCurrentPage = page

        min_index = (CONTACTS_LIST_SIZE * page) - CONTACTS_LIST_SIZE
        max_index = (CONTACTS_LIST_SIZE * page) - 1
        
        ContactsCurrent = ContactsFull[min_index:max_index + 1]
        
        for i in range(len(btnContactsList)):
            if i >= len(ContactsCurrent):
                btnContactsList[i].SetText('')
            else:    
                btnContactsList[i].SetText(ContactsCurrent[i])

#-------------------------------------------------------------------------------
# Meetings
#-------------------------------------------------------------------------------

def ProcessMeetingActiveSessionChange(info):
    url = info.Url.replace('sip:+', '').replace('sip:', '').replace(';user=phone', '')
        
    btnMeetingActiveName.SetText(info.MeetingID.title())
    btnMeetingActiveID.SetText(url)
    btnMeetingActiveState.SetText(info.StatusText.title())


def SetMeetingsList(page):
    global MeetingsFull, MeetingsCurrent, MeetingsCurrentPage, MeetingsTotalPages
    
    if MeetingsTotalPages == 0:
        for i in range(len(btnMeetingList)):
            btnMeetingList[i].SetText('')
    
    elif page > 0 and page <= MeetingsTotalPages:
        MeetingsCurrentPage = page

        min_index = (MEETINGS_LIST_SIZE * page) - MEETINGS_LIST_SIZE
        max_index = (MEETINGS_LIST_SIZE * page) - 1
        
        MeetingsCurrent = MeetingsFull[min_index:max_index + 1]
        
        for i in range(len(btnMeetingList)):
            if i >= len(MeetingsCurrent):
                btnMeetingList[i].SetText('')
            else:    
                btnMeetingList[i].SetText(MeetingsCurrent[i].DisplayName.title())

################################################################################
## Event Definitions
################################################################################

#-------------------------------------------------------------------------------
# Device Connection
#-------------------------------------------------------------------------------

if MyCommType == CommTypes['Ethernet']:
    @event(MyModule, 'Connected')
    @event(MyModule, 'Disconnected')
    def OnConnectionChange(interface, state):
        print(str(state))
        if state == 'Disconnected':
            MyModule.Reinit()

#-------------------------------------------------------------------------------
# Communication/Initialization/Registration
#-------------------------------------------------------------------------------

@event(btnReinit, BUTTON_STATES)
def ReinitSelect(button, state):
    if state == 'Pressed':
        global MeetingsJoinText
        
        MeetingsJoinText = ''
        btnMeetingCreateJoinID.SetText(MeetingsJoinText)
        
        MyModule.Reinit()


@event(MyModule, 'ModuleInfo')
def ModuleInfoEvent(interface, communicating, initialized):
    btnCommunicating.SetState(communicating)
    btnInitialized.SetState(initialized)   


@event(MySkype, 'OnRegistrationStatusChange')
def RegistrationEvent(interface, status):
    if isinstance(status, MyClearOne.S4BRegistrationStatus):
        IsRegistered = status.State == MyClearOne.S4BRegState['REGISTERED']
        btnRegistered.SetState(IsRegistered) 

#-------------------------------------------------------------------------------
# Menu
#-------------------------------------------------------------------------------

@event(btnMenuList, BUTTON_STATES)
def MenuSelect(button, state):
    if state == 'Pressed':
        index = btnMenuList.index(button)
        btnMenuListME.SetCurrent(index)
        TLPPro1020T.ShowPopup(puMenuNames[index])

#-------------------------------------------------------------------------------
# Dialer
#-------------------------------------------------------------------------------   
        
@event(btnDialerDialpad, BUTTON_STATES)
def DialerDialpadSelect(button, state):
    index = btnDialerDialpad.index(button)
    
    if index <= len(btnDialerDialpadMap):
        if state == 'Pressed':      
            MySkype.KeypadPress(btnDialerDialpadMap[index])
            
        elif state == 'Released':
            MySkype.KeypadRelease(btnDialerDialpadMap[index])


@event(btnDialerClear, BUTTON_STATES)
def DialerClearSelect(button, state):
    if state == 'Pressed':
        MySkype.KeypadClear()


@event(btnDialerBack, BUTTON_STATES)
def DialerBackSelect(button, state):
    if state == 'Pressed':
        MySkype.KeypadBackspace()


@event(btnDialerDial, BUTTON_STATES)
def DialerDialSelect(button, state):
    if state == 'Pressed':
        MySkype.KeypadDial()


@event(btnIncomingAnswer, BUTTON_STATES)
def DialerIncomingAnswerSelect(button, state):
    if state == 'Pressed':
        MySkype.Accept()


@event(btnIncomingIgnore, BUTTON_STATES)
def DialerIncomingIgnoreSelect(button, state):
    if state == 'Pressed':
        MySkype.Reject()


@event(MySkype, 'OnKeypadTextChange')
def KeypadTextChangeEvent(interface, value):
    txtDialString.SetText(value)


@event(MySkype, 'OnActiveSessionInfoUpdate')
def ActiveSessionCallInfoChangeEvent(interface, info):
    if isinstance(info, MyClearOne.S4BActiveSessionInfo):
        ProcessCallActiveSessionChange(info)
        ProcessMeetingActiveSessionChange(info)

#-------------------------------------------------------------------------------
# Keyboard
#-------------------------------------------------------------------------------

@event(btnKeyboardClose, BUTTON_STATES)
def ContactsKeyboardCloseSelect(button, state):
    if state == 'Pressed':
        global CurrentKeyboard
        
        CurrentKeyboard = KeyboardType['NONE']
        TLPPro1020T.HidePopup(puKeyboard)


@event(btnKeyboardBack, BUTTON_STATES)
def ContactsKeyboardBackSelect(button, state):
    if state == 'Pressed':
        global ContactsSearchText, MeetingsJoinText, CurrentKeyboard
        
        if CurrentKeyboard == KeyboardType['CONTACTS']:
            if isinstance(ContactsSearchText, str) and len(ContactsSearchText) > 0:
                ContactsSearchText = ContactsSearchText[:-1]
                MySkype.SetFilterSearchText(ContactsSearchText)
        
        elif CurrentKeyboard == KeyboardType['MEETINGS']:
            if isinstance(MeetingsJoinText, str) and len(MeetingsJoinText) > 0:
                MeetingsJoinText = MeetingsJoinText[:-1]
                btnMeetingCreateJoinID.SetText(MeetingsJoinText)
                

@event(btnKeyboardClear, BUTTON_STATES)
def ContactsKeyboardClearSelect(button, state):
    if state == 'Pressed':
        global ContactsSearchText, MeetingsJoinText, CurrentKeyboard
        
        if CurrentKeyboard == KeyboardType['CONTACTS']:
            ContactsSearchText = ''
            MySkype.SetFilterSearchText(ContactsSearchText)
        
        elif CurrentKeyboard == KeyboardType['MEETINGS']:
            MeetingsJoinText = ''
            btnMeetingCreateJoinID.SetText(MeetingsJoinText)


@event(btnKeyboard, BUTTON_STATES)
def ContactsKeyboardSelect(button, state):
    index = btnKeyboard.index(button)
    
    if index <= len(btnKeyboardMap):
        if state == 'Pressed':      
            global ContactsSearchText, MeetingsJoinText, CurrentKeyboard
        
            if CurrentKeyboard == KeyboardType['CONTACTS']:
                ContactsSearchText += btnKeyboardMap[index]
                MySkype.SetFilterSearchText(ContactsSearchText)
            
            elif CurrentKeyboard == KeyboardType['MEETINGS']:
                MeetingsJoinText += btnKeyboardMap[index]
                btnMeetingCreateJoinID.SetText(MeetingsJoinText)

#-------------------------------------------------------------------------------
# Contacts
#-------------------------------------------------------------------------------

@event(btnContactsListTop, BUTTON_STATES)
def ContactsListTopSelect(button, state):
    if state == 'Pressed':
        SetContactList(1)


@event(btnContactsListUp, BUTTON_STATES)
def ContactsListUpSelect(button, state):
    if state == 'Pressed':
        global ContactsCurrentPage
        SetContactList(ContactsCurrentPage - 1)


@event(btnContactsListDown, BUTTON_STATES)
def ContactsListDownSelect(button, state):
    if state == 'Pressed':
        global ContactsCurrentPage
        SetContactList(ContactsCurrentPage + 1)


@event(btnContactsListBottom, BUTTON_STATES)
def ContactsListBottomSelect(button, state):
    if state == 'Pressed':
        global ContactsTotalPages
        SetContactList(ContactsTotalPages)


@event(btnContactsListHome, BUTTON_STATES)
def ContactsListHomeSelect(button, state):
    if state == 'Pressed':
        MySkype.ResetList()


@event(btnContactsList, BUTTON_STATES)
def ContactsListItemSelect(button, state):
    if state == 'Pressed':
        global ContactsCurrentPage
        
        index = btnContactsList.index(button)
        
        if index < CONTACTS_LIST_SIZE:
            contact = (CONTACTS_LIST_SIZE * (ContactsCurrentPage - 1)) + index + 1
            
            MySkype.SelectListItem(contact)


@event(btnContactsSearch, BUTTON_STATES)
def ContactsListSearchSelect(button, state):
    if state == 'Pressed':
        MySkype.PerformSearch()


@event(btnContactsGroupCreateSession, BUTTON_STATES)
def ContactsGroupCreateSessionSelect(button, state):
    if state == 'Pressed':
        MySkype.CreateSessionUsingSelection()
        
 
@event(btnContactsGroupDeleteGroup, BUTTON_STATES)
def ContactsGroupDeleteGroupSelect(button, state):
    if state == 'Pressed':
        MySkype.DeleteGroup()


@event(btnContactsContactDialContact, BUTTON_STATES)
def ContactsContactDialContactSelect(button, state):
    if state == 'Pressed':
        MySkype.DialContact()
        
        
@event(btnContactsContactAddToGroup, BUTTON_STATES)
def ContactsContactAddToGroupSelect(button, state):
    if state == 'Pressed':
        MySkype.AddContactToGroup()
        
        
@event(btnContactsContactRemoveFromGroup, BUTTON_STATES)
def ContactsContactRemoveFromGroupSelect(button, state):
    if state == 'Pressed':
        MySkype.RemoveContactFromGroup()
        
        
@event(btnContactsContactRemoveFromAllGroups, BUTTON_STATES)
def ContactsContactRemoveFromAllGroupsSelect(button, state):
    if state == 'Pressed':
        MySkype.RemoveContactFromAllGroups()
        
        
@event(btnContactsContactAddToActiveSession, BUTTON_STATES)
def ContactsContactAddToActiveSessionSelect(button, state):
    if state == 'Pressed':
        MySkype.AddContactActiveSession()
        
        
@event(btnContactsContactCreateSession, BUTTON_STATES)
def ContactsContactCreateSessionSelect(button, state):
    if state == 'Pressed':
        MySkype.CreateSessionUsingSelection()


@event(btnContactsString, BUTTON_STATES)
def ContactsListStringSelect(button, state):
    if state == 'Pressed':
        global CurrentKeyboard
        
        CurrentKeyboard = KeyboardType['CONTACTS']
        TLPPro1020T.ShowPopup(puKeyboard)


@event(MySkype, 'OnContactListUpdate')
def ContactListUpdateChangeEvent(interface, info):
    global ContactsFull, ContactsTotalPages
    
    if isinstance(info, MyClearOne.S4BContactListResponse) and isinstance(info.List, list):
        ContactsFull = info.List.copy()

        if len(ContactsFull) > 0:  
            ContactsTotalPages = (len(ContactsFull) + (CONTACTS_LIST_SIZE - 1)) // CONTACTS_LIST_SIZE
        else:
            ContactsTotalPages = 0
        
        SetContactList(1)


@event(MySkype, 'OnSelectedItemChange')
def ContactListSelectionChangeEvent(interface, info):
    if isinstance(info, MyClearOne.S4BSelectedItem):
        btnContactsGroupName.SetText(info.DisplayName)
        
        btnContactsContactName.SetText(info.DisplayName)
        btnContactsContactGroup.SetText(info.Group)
        
        if info.Type == MyClearOne.S4BListType['GROUPS']:
            SetContactItemVisibility(False)
            SetGroupItemVisibility(True)            
        
        elif info.Type == MyClearOne.S4BListType['CONTACTS']:
            SetGroupItemVisibility(False)
            SetContactItemVisibility(True)
        
        else:
            SetGroupItemVisibility(False)
            SetContactItemVisibility(False)


@event(MySkype, 'OnFilterSearchTextChange')
def FilterSearchTextChangeEvent(interface, value):
    global ContactsSearchText
    
    ContactsSearchText = value
    btnContactsString.SetText(value)
    
#-------------------------------------------------------------------------------
# Meetings
#-------------------------------------------------------------------------------

@event(btnMeetingListTop, BUTTON_STATES)
def MeetingListTopSelect(button, state):
    if state == 'Pressed':
        SetMeetingsList(1)


@event(btnMeetingListUp, BUTTON_STATES)
def MeetingListUpSelect(button, state):
    if state == 'Pressed':
        global MeetingsCurrentPage
        SetMeetingsList(MeetingsCurrentPage - 1)


@event(btnMeetingListDown, BUTTON_STATES)
def MeetingListDownSelect(button, state):
    if state == 'Pressed':
        global MeetingsCurrentPage
        SetMeetingsList(MeetingsCurrentPage + 1)


@event(btnMeetingListBottom, BUTTON_STATES)
def MeetingListBottomSelect(button, state):
    if state == 'Pressed':
        global MeetingsTotalPages
        SetMeetingsList(MeetingsTotalPages)


@event(btnMeetingList, BUTTON_STATES)
def MeetingsListItemSelect(button, state):
    if state == 'Pressed':
        global MeetingsCurrentPage
        
        index = btnMeetingList.index(button)
        
        if index < MEETINGS_LIST_SIZE:
            session = (MEETINGS_LIST_SIZE * (MeetingsCurrentPage - 1)) + index + 1
            
            MySkype.SelectSession(session)


@event(btnMeetingActiveAccept, BUTTON_STATES)
def MeetingActiveAcceptSelect(button, state):
    if state == 'Pressed':
        MySkype.Accept()


@event(btnMeetingActiveLeave, BUTTON_STATES)
def MeetingActiveLeaveSelect(button, state):
    if state == 'Pressed':
        MySkype.Leave()
        
        
@event(btnMeetingActiveResume, BUTTON_STATES)
def MeetingActiveResumeSelect(button, state):
    if state == 'Pressed':
        MySkype.Resume()


@event(btnMeetingActiveHold, BUTTON_STATES)
def MeetingActiveHoldSelect(button, state):
    if state == 'Pressed':
        MySkype.Hold()


@event(btnMeetingCreateJoinID, BUTTON_STATES)
def MeetingCreateJoinIDSelect(button, state):
    if state == 'Pressed':
        global CurrentKeyboard
        
        CurrentKeyboard = KeyboardType['MEETINGS']
        TLPPro1020T.ShowPopup(puKeyboard)


@event(btnMeetingCreateJoinMeeting, BUTTON_STATES)
def MeetingCreateJoinMeetingSelect(button, state):
    if state == 'Pressed':
        global MeetingsJoinText
        
        if isinstance(MeetingsJoinText, str) and len(MeetingsJoinText) > 0:
            MySkype.JoinMeeting(MeetingsJoinText)
            
            MeetingsJoinText = ''
            btnMeetingCreateJoinID.SetText('')
            

@event(btnMeetingCreateJoinConference, BUTTON_STATES)
def MeetingCreateJoinConferenceSelect(button, state):
    if state == 'Pressed':
        global MeetingsJoinText
        
        if isinstance(MeetingsJoinText, str) and len(MeetingsJoinText) > 0:
            MySkype.JoinConference(MeetingsJoinText)
            
            MeetingsJoinText = ''
            btnMeetingCreateJoinID.SetText('')
        

@event(btnMeetingCreateJoinNow, BUTTON_STATES)
def MeetingCreateJoinNowSelect(button, state):
    if state == 'Pressed':
        MySkype.MeetNow()


@event(MySkype, 'OnSessionListUpdate')
def SessionListUpdateChangeEvent(interface, info):
    global MeetingsFull, MeetingsTotalPages
    
    if isinstance(info, MyClearOne.S4BSessionListResponse) and isinstance(info.List, list):
        MeetingsFull = info.List.copy()

        if len(MeetingsFull) > 0:  
            MeetingsTotalPages = (len(MeetingsFull) + (MEETINGS_LIST_SIZE - 1)) // MEETINGS_LIST_SIZE
        else:
            MeetingsTotalPages = 0
        
        SetMeetingsList(1) 

#-------------------------------------------------------------------------------
# Program Initialize
#-------------------------------------------------------------------------------

def Initialize():
    global ContactsSearchText, MeetingsJoinText
    
    ContactsSearchText = ''
    btnContactsString.SetText('')
    
    MeetingsJoinText = ''
    btnMeetingCreateJoinID.SetText('')
    
    CurrentKeyboard = KeyboardType['NONE']

    TLPPro1020T.HideAllPopups()
    btnMenuListME.SetCurrent(0)
    TLPPro1020T.ShowPopup(puMenuNames[0])
    
    MyModule.SetDebug(False)
    MyModule.SetPoll(False)
    MyModule.Reinit()

################################################################################
## Main
################################################################################

Initialize()